/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "stdafx.h"
#include "DVDSubEdit.h"
#include "DVDSubEditDlg.h"
#include "Cell.h"
#include "Utils.h"
#include "SubPic.h"
#include "MiscDialogs.h"
#include <math.h>
#include <wingdi.h>

#ifndef AC_SRC_ALPHA
#define AC_SRC_ALPHA                0x01
#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define MAX_V_MOVE 600
#define MAX_H_MOVE 500
#define MAX_BORDER 100
#define MAX_DURATION 10000
#define GRAY RGB(100,100,100)
#if 0
#define BLACK RGB(0,0,0)
#define RED RGB(255,0,0)
#define GREEN RGB(0,134,37)
#define BLUE RGB(0,0,255)
#else
#define BLACK 0
#define RED 1
#define GREEN 2
#define BLUE 3
#endif


#define DispThis1(ID,format) {SetDlgItemText(ID,format);}
#define DispThis2(ID,format,val) {char Temp[512]; sprintf(Temp,format,val);SetDlgItemText(ID,Temp);}
#define DispThis3(ID,format,val1,val2) {char Temp[512]; sprintf(Temp,format,val1,val2);SetDlgItemText(ID,Temp);}

#define ENABLE(ID,En){CButton* Temp = (CButton*)GetDlgItem(ID);Temp->EnableWindow(En);}

void OutputError(char* AMessage, int Langage)
{
	AfxMessageBox(AMessage,MB_ICONSTOP);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}

FILE*	OutLogFd;

void OutputLog(char* AMessage, int Langage)
{
	CTime CurrentTime = CTime::GetCurrentTime();
	CString Time = CurrentTime.Format("%H:%M:%S");
	if(OutLogFd) {fprintf(OutLogFd,"%s -- %s\n",Time,AMessage);fflush(OutLogFd);}
	strcpy(Message,AMessage);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}

enum
{
	BORDER_CUSTOM = 0,
	BORDER_SAFE,
	BORDER_4_3_L,
	BORDER_4_3_LS,
};

extern "C" void RunOCR(unsigned char* Bitmap,int nx, int ny, char (*Line)[128], int *NumLines, int spc);

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

	CDVDSubEditDlg	*m_Edit;

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	virtual BOOL OnInitDialog();
	afx_msg void OnShowhelp();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BOOL CAboutDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	SetDlgItemText(IDC_VERSIONSTRING,m_Edit->m_Version);
	return TRUE; 
}

void CAboutDlg::OnShowhelp() 
{
	OpenRSCFile(IDR_HELPFILE, ".pdf");
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
	ON_BN_CLICKED(IDC_SHOWHELP, OnShowhelp)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
class CScanDlg : public CDialog
{
public:
	CScanDlg();

	CSubPicEngine*	m_Engine;
	void			DisplayInfo(int Progress, char* Info, char* Message);

// Dialog Data
	//{{AFX_DATA(CScanDlg)
	enum { IDD = IDD_SCANDIALOG };
	CProgressCtrl	m_Progress;
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CScanDlg)
	public:
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CScanDlg)
	afx_msg void StartScan();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CScanDlg::CScanDlg() : CDialog(CScanDlg::IDD)
{
	//{{AFX_DATA_INIT(CScanDlg)
	//}}AFX_DATA_INIT
	m_Engine = 0;
}

void CScanDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CScanDlg)
	DDX_Control(pDX, IDC_PROGRESS, m_Progress);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CScanDlg, CDialog)
	//{{AFX_MSG_MAP(CScanDlg)
	ON_BN_CLICKED(IDC_RUN, StartScan)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


void CScanDlg::StartScan() 
{
	m_Engine->m_AbortNow = true;
}

void CScanDlg::DisplayInfo(int Progress, char* Info, char* AMessage)
{
	m_Progress.SetPos(Progress);
	SetWindowText(Info);
	SetDlgItemText(IDC_INFO,Info);
	SetDlgItemText(IDC_MESSAGE,AMessage);
}

BOOL CScanDlg::PreTranslateMessage(MSG* pMsg) 
{
	if(pMsg->message == WM_KEYDOWN)
	{
		switch(pMsg->wParam)
		{
			// Must be all CAPS...
		case VK_ESCAPE: m_Engine->m_AbortNow = true;  break;
		default: return GetParent()->PreTranslateMessage(pMsg);  break;
		}
		return TRUE;
	}
	else return CDialog::PreTranslateMessage(pMsg);
}

/////////////////////////////////////////////////////////////////////////////
void Modif::Reset()
{
	m_VMove = m_HMove = m_Transparency = m_StartCommand = m_Duration = m_Crop = -1;
	// These are special. I'm resetting them to 0, because they will be or'ed when modified.
	m_ColorPixel = m_Color = 0;
	m_Border = 0;
}

void Modif::RecordChange(CSubPic* TheSubpic, Change Which, int Value1, int Value2)
{
	// If this modification is applied to a "new" subpic, let's reset all the previously recorded changes.
	if(TheSubpic->m_SPUNumber != m_WhichSPU) Reset();
	// Register the changes.
	switch(Which)
	{
	case ChangeVMove : m_VMove = Value1; break;
	case ChangeHMove : m_HMove = Value1; break;
	case ChangeBorder : m_Border = Value1; break;
	case ChangeTransp : m_Transparency = Value1; m_TranspPixel = Value2; break;
	// Color is special: m_ColorPixel has bit 0 set for a change to pixel 0 etc. m_Color has 1 byte
	// storing the color per pixel.
	case ChangeColor : m_Color |= Value1; m_ColorPixel |= Value2; break;
	case ChangeStart : m_StartCommand = Value1; break;
	case ChangeDur : m_Duration = Value1; break;
	case ChangeCrop : m_Crop = Value1; break;
	}
	// Remember the CLUT, for changes to color, and the SPUNumber the changes came from.
	m_CLUT = TheSubpic->m_CLUT;
	m_WhichSPU = TheSubpic->m_SPUNumber;
}


/////////////////////////////////////////////////////////////////////////////
// CDVDSubEditDlg dialog

CDVDSubEditDlg::CDVDSubEditDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CDVDSubEditDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDVDSubEditDlg)
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	pErrorFunction = OutputError;
	pMessageFunction = OutputLog;
	m_OCRBitmap = 0;
	m_OCRLines = 0;
	m_InitWindowWidth = m_InitWindowHeight = -1;
}

void CDVDSubEditDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDVDSubEditDlg)
	DDX_Control(pDX, IDC_AUTOCLUT, m_AutoCLUT);
	DDX_Control(pDX, IDC_FLUSH, m_FlushBorders);
	DDX_Control(pDX, IDC_MOVELR, m_MoveHSlider);
	DDX_Control(pDX, IDC_SHOWSEARCHED, m_ShowSearched);
	DDX_Control(pDX, IDC_OCROUT, m_OCROut);
	DDX_Control(pDX, IDC_DCSQTSELECT, m_DCSQTSelect);
	DDX_Control(pDX, IDC_DURATIONS, m_DurationSlider);
	DDX_Control(pDX, IDC_SHOWMOD, m_ShowModified);
	DDX_Control(pDX, IDC_BORDERSELECT, m_BorderSelect);
	DDX_Control(pDX, IDC_BORDERS, m_BorderSlider);
	DDX_Control(pDX, IDC_START, m_StartCommand);
	DDX_Control(pDX, IDC_PIXSELECT, m_PixelSelect);
	DDX_Control(pDX, IDC_TRANSPARENCYS, m_TransSlider);
	DDX_Control(pDX, IDC_SHOWVIDEO, m_ShowVideo);
	DDX_Control(pDX, IDC_SAVEALL, m_SaveAll);
	DDX_Control(pDX, IDC_NOCLUT, m_IgnoreCLUT);
	DDX_Control(pDX, IDC_REVERTSAVED, m_Revert);
	DDX_Control(pDX, IDC_MOVEUDS, m_MoveVSlider);
	DDX_Control(pDX, IDC_VCIDSELECT, m_VCIDSelect);
	DDX_Control(pDX, IDC_SPUSLIDER, m_SPUSlider);
	DDX_Control(pDX, IDC_STREAMSELECT, m_StreamSelect);
	DDX_Control(pDX, IDC_COMMANDS, m_CommandDisplay);
	//}}AFX_DATA_MAP
}

static UINT FindDialogMessage = ::RegisterWindowMessage(FINDMSGSTRING);

BEGIN_MESSAGE_MAP(CDVDSubEditDlg, CDialog)
	//{{AFX_MSG_MAP(CDVDSubEditDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_DESTROY()
	ON_WM_TIMER()
	ON_WM_CTLCOLOR()
	ON_BN_CLICKED(IDC_SPUR, OnSpur)
	ON_WM_HSCROLL()
	ON_CBN_SELCHANGE(IDC_STREAMSELECT, OnSelchangeStreamselect)
	ON_BN_CLICKED(IDC_SPUL, OnSpul)
	ON_BN_CLICKED(IDC_REVERTSAVED, OnRevertsaved)
	ON_BN_CLICKED(IDC_PVCID, OnPvcid)
	ON_BN_CLICKED(IDC_NVCID, OnNvcid)
	ON_BN_CLICKED(IDC_PSTREAM, OnPstream)
	ON_BN_CLICKED(IDC_NSTREAM, OnNstream)
	ON_BN_CLICKED(IDC_APPLYTOALL, OnApplytoall)
	ON_BN_CLICKED(IDC_REVERTALL, OnRevertall)
	ON_WM_CLOSE()
	ON_BN_CLICKED(IDC_NOCLUT, OnNoclut)
	ON_BN_CLICKED(IDC_SAVEALL, OnSaveall)
	ON_BN_CLICKED(IDC_DEFBUTTON, OnDefbutton)
	ON_EN_KILLFOCUS(IDC_MOVEUDE, OnKillfocusMoveude)
	ON_EN_KILLFOCUS(IDC_MOVELRE, OnKillfocusMovelre)
	ON_BN_CLICKED(IDC_SHOWVIDEO, OnShowvideo)
	ON_CBN_SELCHANGE(IDC_PIXSELECT, OnSelchangePixselect)
	ON_CBN_SELCHANGE(IDC_START, OnSelchangeStart)
	ON_CBN_SELCHANGE(IDC_BORDERSELECT, OnSelchangeBorderselect)
	ON_BN_CLICKED(IDC_SHOWMOD, OnShowmod)
	ON_EN_KILLFOCUS(IDC_BORDERE, OnKillfocusBordere)
	ON_COMMAND(ID_HELP_ABOUT, OnHelpAbout)
	ON_COMMAND(ID_HELP_MANUAL, OnHelpManual)
	ON_COMMAND(ID_HELP_ONTHEWEB, OnHelpOntheweb)
	ON_COMMAND(ID_EDIT_REDO, OnEditRedo)
	ON_COMMAND(ID_FILE_EXIT, OnExit)
	ON_COMMAND(ID_EDIT_APPLYALL, OnEditApplyall)
	ON_NOTIFY(UDN_DELTAPOS, IDC_MOVEB, OnDeltaposMoveb)
	ON_NOTIFY(UDN_DELTAPOS, IDC_MOVELRB, OnDeltaposMovelrb)
	ON_NOTIFY(UDN_DELTAPOS, IDC_BORDERB, OnDeltaposBorderb)
	ON_COMMAND(ID_FILE_OPENVOB, OnOpenVob)
	ON_COMMAND(ID_FILE_OPENDOMAIN, OnOpenDomain)
	ON_CBN_SELCHANGE(IDC_DCSQTSELECT, OnSelchangeDcsqtselect)
	ON_NOTIFY(UDN_DELTAPOS, IDC_DSCQTB, OnDeltaposDscqtb)
	ON_COMMAND(ID_EDIT_HIDE, OnHidesubpic)
	ON_COMMAND(ID_HELP_VIEWLOG, OnHelpViewlog)
	ON_BN_CLICKED(IDC_FINDL, OnFindLeft)
	ON_BN_CLICKED(IDC_FINDR, OnFindRight)
	ON_COMMAND(ID_FILE_SAVEPPM, OnSavePPM)
	ON_BN_CLICKED(IDC_SHOWSEARCHED, OnShowsearched)
	ON_EN_KILLFOCUS(IDC_SEARCHSELECT, OnKillfocusSearchselect)
	ON_EN_KILLFOCUS(IDC_SEARCH, OnKillfocusSearch)
	ON_EN_KILLFOCUS(IDC_DURATIONE, OnKillfocusDuratione)
	ON_COMMAND(ID_FILE_SAVESUP, OnSaveSup)
	ON_BN_CLICKED(IDC_FLUSH, OnFlushBorder)
	ON_WM_SIZE()
	ON_EN_CHANGE(IDC_SEARCHSELECT, OnKillfocusSearchselect)
	ON_BN_CLICKED(IDC_HIDESUBPIC, OnHidesubpic)
	ON_CBN_SELCHANGE(IDC_VCIDSELECT, OnSelchangeStreamselect)
	//}}AFX_MSG_MAP
	ON_WM_INITMENU()
	ON_COMMAND(ID_EDIT_REVERTTHIS, OnRevertsaved)
	ON_COMMAND(ID_EDIT_REVERTALL, OnRevertall)
	ON_COMMAND(ID_FILE_SAVE, OnSaveall)
	ON_COMMAND(ID_FILE_OPENLAST, LoadLastFile)
	ON_COMMAND(ID_FILE_OPENSUP, OnOpenSup)
	ON_COMMAND(ID_FILE_OPENIFO, OnOpenIFO)
	ON_COMMAND(ID_OCR_ONE, OCRThisOne)
	ON_COMMAND(ID_OCR_REMOVEH, HideHearingImpaired)
	ON_WM_DROPFILES()
	ON_WM_GETMINMAXINFO()
	ON_NOTIFY_EX( TTN_NEEDTEXT, 0, OnToolTipNotify )
	ON_BN_CLICKED(IDC_RUNOCR, RunOCROnSelected)
	ON_BN_CLICKED(IDC_SAVEOCR, OnSaveOCR)
	ON_BN_CLICKED(IDC_RETIMEALL, OnTimingAll)
	ON_BN_CLICKED(IDC_RETIMEONE, OnTimingSingle)
	ON_BN_CLICKED(IDC_CROP, OnCropSubpic)
	ON_BN_CLICKED(IDC_AUTOCLUT, OnNoclut)
	ON_CONTROL_RANGE(BN_CLICKED, IDC_COLB, IDC_COLE2, OnClutClick )
END_MESSAGE_MAP()

BOOL CDVDSubEditDlg::OnToolTipNotify( UINT id, NMHDR * pNMHDR, LRESULT * pResult )
{
	// Multiline tooltips: http://www.thecodeproject.com/miscctrl/tooltipex.asp
    TOOLTIPTEXT *pTTT = (TOOLTIPTEXT *)pNMHDR;
    UINT nID =pNMHDR->idFrom;
    if (pTTT->uFlags & TTF_IDISHWND)
    {
        // idFrom is actually the HWND of the tool
        nID = ::GetDlgCtrlID((HWND)nID);
		switch(nID)
		{
		case IDC_AUTOCLUT : pTTT->lpszText = "Adjusts CLUT to automatically show character pixels in white, all others in black."; break; 
		case IDC_CROP : pTTT->lpszText = "Crop subpic background to allow adjustment of horizontal position in all cases."; break; 
		case IDC_STREAMSELECT : pTTT->lpszText = "Filter subpics based on stream ID"; break; 
		case IDC_SPUL : pTTT->lpszText = "Go to previous subpic"; break; 
		case IDC_SPUR : pTTT->lpszText = "Go to next subpic"; break; 
		case IDC_REVERTSAVED : pTTT->lpszText = "Reloads this subpic from disk"; break; 
		case IDC_VCIDSELECT : pTTT->lpszText = "Filter subpics based on V/CID"; break; 
		case IDC_PVCID : pTTT->lpszText = "Go to previous V/CID"; break; 
		case IDC_NVCID : pTTT->lpszText = "Go to next V/CID"; break; 
		case IDC_PSTREAM : pTTT->lpszText = "Go to previous subpic stream"; break; 
		case IDC_NSTREAM : pTTT->lpszText = "Go to next subpic stream"; break; 
		case IDC_APPLYTOALL : pTTT->lpszText = "Apply these modifications to all selected subpics"; break; 
		case IDC_REVERTALL : pTTT->lpszText = "Reloads all selected subpics from disk"; break; 
		case IDC_SPUSLIDER : pTTT->lpszText = "Picks a subpic in the current selection"; break; 
		case IDC_MOVEUDS : pTTT->lpszText = "Moves the subpic currently displayed up and down"; break; 
		case IDC_COLB : pTTT->lpszText = "Shows color used for 'b' (background) pixels"; break; 
		case IDC_COLP : pTTT->lpszText = "Shows color used for 'p' (pattern) pixels"; break; 
		case IDC_COLE1 : pTTT->lpszText = "Shows color used for 'e1' (emphasis 1) pixels"; break; 
		case IDC_COLE2 : pTTT->lpszText = "Shows color used for 'e2' (emphasis 2) pixels"; break; 
		case IDC_NOCLUT : pTTT->lpszText = "Don't use the PGC Color LookUp Table (CLUT). Use user-specified CLUT instead"; break; 
		case IDC_SAVEALL : pTTT->lpszText = "Saves all the changes back to file."; break; 
		case IDC_MOVEUDE : pTTT->lpszText = "Enter amount to move the subpic here."; break; 
		case IDC_PIXSELECT : pTTT->lpszText = "Selects of which pixel to change the transparency. All means all non-background pixels."; break; 
		case IDC_TRANSPARENCYS : pTTT->lpszText = "Adjusts transparency for selected pixels."; break; 
		case IDC_START : pTTT->lpszText = "Select the subpic start command. Forced means the subpic will be on if the stream is selected, even if subtitles are off."; break; 
		case IDC_BORDERS : pTTT->lpszText = "Adjusts the vertical margin between subtitles and image borders. Only available if there are no buttons"; break; 
		case IDC_BORDERSELECT : pTTT->lpszText = "Pick standard values for vertical borders. Only available if there are no buttons"; break; 
		case IDC_SHOWMOD : pTTT->lpszText = "Only show modified subpics"; break; 
		case IDC_DURATIONS : pTTT->lpszText = "Adjusts the subpic duration"; break; 
		case IDC_DCSQTSELECT : pTTT->lpszText = "Selects the DCSQT (set of commands) to visualize and modify"; break; 
		case IDC_HIDESUBPIC : pTTT->lpszText = "Completely hide this subpic (make it transparent)"; break; 

		case IDC_RUNOCR : pTTT->lpszText = "Run Optical Character Recognition on selected subpics"; break; 
		case IDC_SAVEOCR : pTTT->lpszText = "Save subpics as .srt file"; break; 
		case IDC_RETIMEALL : pTTT->lpszText = "Modify start time of all selected subpics"; break; 
		case IDC_RETIMEONE : pTTT->lpszText = "Modify start time of current subpic"; break; 
		case IDC_FINDR : pTTT->lpszText = "Find next occurence of search string in subpics"; break; 
		case IDC_FINDL : pTTT->lpszText = "Find previous occurence of search string in subpics"; break; 
		case IDC_SHOWSEARCHED : pTTT->lpszText = "Only show subpics containing one of the following characters"; break; 
		case IDC_SEARCHSELECT : pTTT->lpszText = "Type characters to be found in subpics"; break; 
		case IDC_SEARCH : pTTT->lpszText = "Type search string"; break; 
		case IDC_MOVELR : pTTT->lpszText = "Moves subpic left or right. If disabled, crop the background (button below)."; break; 
		case IDC_FLUSH : pTTT->lpszText = "Moves subpic flush with the nearest border, if borders are used"; break; 
		}
		pTTT->hinst = AfxGetResourceHandle();
		return(TRUE);
    }
    return(FALSE);
}

// This function will be called after the display of each video frame.
void PostVideoDisplayFunc(HDC hDC,void* Param)
{
	CDVDSubEditDlg* DVDSubEdit = (CDVDSubEditDlg*) Param;
	// We simply call our ShowBorders function;
	DVDSubEdit->ShowBorders();
}

extern void BitmapMixFunction(unsigned char* Bitmap, int Width, int Height, void* Param)
{
	CDVDSubEditDlg* DVDSubEdit = (CDVDSubEditDlg*) Param;
	// We simply call our DisplayBitmap function;
	DVDSubEdit->DisplayBitmap(Bitmap, Width, Height);
}



/////////////////////////////////////////////////////////////////////////////
// CDVDSubEditDlg message handlers

#define LOADCOLOR(Array,Col){*(Array)=GetRValue(Col);*(Array+1)=GetGValue(Col);*(Array+2)=GetBValue(Col);}

BOOL CDVDSubEditDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

    m_Menu.LoadMenu (IDR_MENU);
	SetMenu(&m_Menu);

	
	// TODO: Add extra initialization here
	DragAcceptFiles();
	EnableToolTips(true);

	// Make a note of initial window width and height.
	CRect TempRect;
	GetWindowRect(&TempRect);
	m_InitWindowWidth = TempRect.Width();
	m_InitWindowHeight = TempRect.Height();

	TheSettings.LoadSettings(m_hWnd,"MainWindowPos",0);
	InitializeCriticalSection(&m_csLock);
	
	m_QuitNow = 0;
	m_Running = 0;
	m_ThreadHandle = 0;
	strcpy(Message,"");
	m_ScanDialog = new CScanDlg;
	m_ScanDialog->CDialog::Create(IDD_SCANDIALOG);
	m_ScanDialog->m_Engine = this;
	TheSettings.LoadSettings(m_ScanDialog->m_hWnd,"ScanWindowPos",1);


	m_ApplyAllDialog = new CApplyToAllDlg;
	m_ApplyAllDialog->m_Engine = this;
	m_MoveDialog = new CMoveDlg;
	m_MoveDialog->m_Engine = this;
	TheSettings.LoadSettings(m_ApplyAllDialog->m_hWnd,"ApplyAllWindowPos",1);

	TheSettings.LoadSettings(&m_DisplayVideo,"ShowVideo",1);
	m_ShowVideo.SetCheck(m_DisplayVideo);

	DWORD Color = TheSettings.LoadSettings("DefClutB",(long)RGB(0,0,0));
	LOADCOLOR(m_DefaultColors,Color);
	Color = TheSettings.LoadSettings("DefClutP",(long)RGB(255,255,255));
	LOADCOLOR(m_DefaultColors+3,Color);
	Color = TheSettings.LoadSettings("DefClutE1",(long)RGB(0,255,0));
	LOADCOLOR(m_DefaultColors+6,Color);
	Color = TheSettings.LoadSettings("DefClutE2",(long)RGB(0,0,255));
	LOADCOLOR(m_DefaultColors+9,Color);

	CRect Rect;
	GetDlgItem(IDC_VIDEOBACK)->GetWindowRect(&Rect);
	m_DisplayHeight = Rect.Height();

	JustFoundVideoFormat();
	Reset();

	ZeroMemory(&m_BitMapHeader, sizeof(BITMAPINFOHEADER));
	m_BitMapHeader.biSize = sizeof(BITMAPINFOHEADER);
	m_BitMapHeader.biPlanes = 1;
	m_BitMapHeader.biBitCount = 32;
	m_BitMapHeader.biCompression = BI_RGB;

	m_Preview = new CDVDPreview;
	m_Preview->SetDrawButtons(true);
	// This is to conserve memory.
	m_Preview->SetMaxVideoScaling(1.5);
	m_PreviewHandle = m_Preview->InitPreview(m_hWnd,1,1,100,100,SW_HIDE,true,WS_CHILD,true);

	// Now setup the preview so our PostVideoDisplayFunc function is called after each drawn video
	// frame...
	m_Preview->SetBitmapMixFunction(BitmapMixFunction,this);
	m_Preview->SetPostDisplayFunc(PostVideoDisplayFunc,this);
	EnableUI(false);

	if(getenv("TEMP") != 0)
		sprintf(m_LogPath,"%s\\DVDSubEdit.log",getenv("TEMP"));
	OutLogFd = 0;

	return TRUE;  // return TRUE  unless you set the focus to a control
}

// IMPORTANT: Use a floor() not a round, otherwise, the sucpics might go beyond the video (which uses a 
// floor, not a round).
#define RESCALEX(x) ((int)((x)*m_ScalingX+0.0f))
#define RESCALEY(x) ((int)((x)*m_ScalingY+0.0f))

void CDVDSubEditDlg::JustFoundVideoFormat()
{
	GetDlgItem(IDC_VIDEOBACK)->GetWindowRect(&m_VideoRect);
	ScreenToClient(&m_VideoRect);
	int VideoHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
	switch(m_AspectRatio)
	{
	case 1: m_ScalingX = 1; break;
	case 2: m_ScalingX = 4.f/3/720*VideoHeight; break;
	case 3: m_ScalingX = 16.f/9/720*VideoHeight; break;
	case 4: m_ScalingX = 2.21f/720*VideoHeight; break;
	}
	// Tricky! We need to make sure both the width and the height fit our window!
	m_ScalingY = min((float)m_DisplayHeight / VideoHeight, (float)m_VideoRect.Width() / (m_ScalingX*720));
	m_ScalingX *= m_ScalingY;

	int temp = m_VideoRect.Width() - RESCALEX(720);
	m_VideoRect.left += temp/2;
	m_VideoRect.right -= temp/2;
	m_VideoRect.bottom = m_VideoRect.top + RESCALEY(VideoHeight);

	if(!m_NoVideo)
	{
		m_Preview->SetVideoScaling(m_ScalingX,m_ScalingY,0);
		m_Preview->SetVideoPosition(m_VideoRect.left,m_VideoRect.top,0);
	}
	// The flickering happens if you erase the background and there's no video. If you don't erase
	// no flicker if on video. Flickering IF video.
	// Question: I disabled the drawing in store.c but still when I resize, a rectangle is filled
	// with black when showvideo is on! HOW CAN THIS BE!!!! NOT in displaybitmap at least.
	// Not, this happens in ResizeWindow in gui.cpp. The MoveWindow() somehow must paint the window
	// background black, but I'm not sure how.
	InvalidateRect(0);
}

void CDVDSubEditDlg::ShowCol(HDC TheDC, uint ID, uchar* Colormap)
{
	CRect Temp;
	GetDlgItem(ID)->GetWindowRect(&Temp);
	ScreenToClient(&Temp);
	HBRUSH Color = (HBRUSH)CreateSolidBrush(RGB(Colormap[0],Colormap[1],Colormap[2]));
	FillRect(TheDC,Temp,Color);
	HBRUSH Black = (HBRUSH)CreateSolidBrush(BLACK);
	FrameRect(TheDC,&Temp,Black);
	// Don't forget this! If you don't release these resources, the app crashes after a while!
	DeleteObject(Color);
	DeleteObject(Black);
}


void CDVDSubEditDlg::ShowBorders()
{
	if(m_SelectedSubPic == NULL)
		return;
	CDC *TheDC = GetDC();
	HDC hDC = TheDC->m_hDC;

	// Now draw the borders if needed.
	if(m_SelectedSubPic->m_VerticalBorders != 0)
	{
		HPEN hpen;
		hpen = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));
		SelectObject(hDC, hpen);
		SelectObject(hDC, GetStockObject(HOLLOW_BRUSH)); 
		POINT Points[2];
		int VBorder = RESCALEY(m_SelectedSubPic->m_VerticalBorders);
		Points[0].x = m_VideoRect.left; Points[1].x = m_VideoRect.right-1; 
		Points[0].y = m_VideoRect.top+VBorder; 
		Points[1].y = m_VideoRect.top+VBorder;
		Polyline(hDC,Points,2);
		Points[0].y = m_VideoRect.bottom-VBorder-1; 
		Points[1].y = m_VideoRect.bottom-VBorder-1;
		Polyline(hDC,Points,2);
		DeleteObject(hpen);
	}
	ReleaseDC(TheDC);
}


void CDVDSubEditDlg::DisplayBitmap(unsigned char* Bitmap, long Width, long Height, bool Black)
{
	CDC *TheDC = GetDC();
	DWORD LastError = GetLastError();
	if(!TheDC)
		return;
	HDC hDC = TheDC->m_hDC;

	// I'm keeping a local copy here to avoid m_SelectedSubPic becoming 0 under me, which sometimes happens,
	// when the video display is on, and I've no idea why >:(
	CSubPic* SelectedSubPic = m_SelectedSubPic;
	if(Black || SelectedSubPic == 0)
	{
		CRect Temp = m_VideoRect;
		// If I paint all the way to the right, the video might not cover the black and that's ugly.
		Temp.right --;
		HBRUSH Black = (HBRUSH)CreateSolidBrush(BLACK);
		FillRect(hDC, &Temp, Black);
		DeleteObject(Black);
		ReleaseDC(TheDC);
		return;
	}

	// Find which VCID this guy belongs to:
	uchar* CLUT = 0;
	char Lang[32][16];
	CLUT = FindCLUT(SelectedSubPic->m_VOBID,SelectedSubPic->m_CELLID,Lang);
	if(CLUT)
	{
		// Set language strings...
		m_StreamSelect.ResetContent();
		m_StreamSelect.AddString("All Streams");
		int j;
		for(j=0;j<=m_TopStreamID;j++)
			m_StreamSelect.AddString(Lang[j]);
		m_StreamSelect.SetCurSel(m_SelectedStream);
	}
	else
		// It's possible that the subpic has a clut that was assigned from reading a separate IFO.
		CLUT = SelectedSubPic->m_CLUT;
	m_IgnoreCLUT.EnableWindow(CLUT != 0);
	m_IgnoreCLUT.SetWindowText(CLUT?"Ignore CLUT":"No CLUT");
	if(m_IgnoreCLUT.GetCheck() || CLUT == 0 || m_AutoCLUT.GetCheck()) 
		CLUT = 0;

	uchar* ColorMap;
	if(Bitmap != 0)
	{
		// Mix subpic into bitmap with transparency
		ColorMap = SelectedSubPic->MixIntoBitmap(Bitmap, Width, Height,CLUT,
			m_AutoCLUT.GetCheck() ? 0 : m_DefaultColors);
	}
	else
	{
		// Not mixing into an existing bitmap. We have to create one, and then display it.
		// FIX THIS: shouldn't allocate and deallocate!
		long NewWidth = 0;
		long NewHeight = 0;
		int ScaledWidth = 4+RESCALEX(SelectedSubPic->m_SubPicWidth) ;
		int ScaledHeight = 1+RESCALEY(SelectedSubPic->m_SubPicHeight);
		int ScaledStartX = m_VideoRect.left+RESCALEX(SelectedSubPic->m_StartX);
		int ScaledStartY = m_VideoRect.top+RESCALEY(SelectedSubPic->m_StartY);		
		uchar* TempBitmap = new uchar[ScaledWidth*ScaledHeight*3];

		ColorMap = SelectedSubPic->SaveToBitmap(TempBitmap, NewWidth, NewHeight, 1/m_ScalingX, 1/m_ScalingY, CLUT,
			m_AutoCLUT.GetCheck() ? 0 : m_DefaultColors);
		m_BitMapHeader.biBitCount = 24;
		m_BitMapHeader.biWidth = NewWidth;
		m_BitMapHeader.biHeight = NewHeight;
		m_BitMapHeader.biSizeImage = m_BitMapHeader.biWidth * m_BitMapHeader.biHeight * m_BitMapHeader.biBitCount / 8;

		HBRUSH Black = (HBRUSH)CreateSolidBrush(BLACK);
		FillRect(hDC, &m_VideoRect, Black);
		DeleteObject(Black);
		
		SetDIBitsToDevice(hDC, ScaledStartX, ScaledStartY, NewWidth, NewHeight,
			0, 0, 0, NewHeight, TempBitmap, (LPBITMAPINFO)&m_BitMapHeader, DIB_RGB_COLORS);
		delete[] TempBitmap;
		ShowBorders();
	}

	// Show current CLUT colors... THIS IS WHAT CRASHES THE APP!!!!
	ShowCol(hDC, IDC_COLB,ColorMap); ShowCol(hDC,IDC_COLP,ColorMap+3);
	ShowCol(hDC, IDC_COLE1,ColorMap+6); ShowCol(hDC,IDC_COLE2,ColorMap+9);

	ReleaseDC(TheDC);
}


void CDVDSubEditDlg::Reset()
{
	CSubPicEngine::Reset();
	SetWindowText(m_Version);
	m_SelectedStream = 0;
	m_PercentDone = 0;
	m_NumOCRLines = 0;
	strcpy(m_SearchStr,"");
	m_CommandDisplay.SetWindowText("");
	m_OCROut.SetWindowText("");
	m_StreamSelect.ResetContent();
	m_StreamSelect.AddString("All Streams");
	m_StreamSelect.SetCurSel(0);
	m_VCIDSelect.ResetContent();
	m_VCIDSelect.AddString("All VCID");
	m_VCIDSelect.SetCurSel(0);
	m_SPUSlider.SetRange(0,0);
	m_ScalingY = m_ScalingX = 0;
	m_SPUSlider.ShowWindow(SW_HIDE);
	m_MoveVSlider.SetRange(0,MAX_V_MOVE*2);
	m_MoveHSlider.SetRange(0,MAX_H_MOVE*2);
	m_TransSlider.SetRange(0,15);
	m_BorderSlider.SetRange(0,MAX_BORDER);
	m_DurationSlider.SetRange(0,MAX_DURATION);
	m_FlushBorders.SetCheck(0);
	SetBorderSelect(-1);
	SetMoveVSlider(MAX_V_MOVE);
	m_PixelSelect.SetCurSel(4);
	m_ApplyAllDialog->ResetAllCheck();
	DisplayBitmap(0,0,0);
	GetDlgItem(IDC_REVERTALL)->EnableWindow(false);
	GetDlgItem(IDC_REVERTSAVED)->EnableWindow(false);
	m_MenuEnabled = true;
	m_FixItalics = true;
	m_Preview->CloseFile();

	m_ChangeRecord.Reset();
}


// Found here: http://www.codeguru.com/Cpp/controls/richedit/article.php/c2413/
static DWORD CALLBACK EditStreamCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
	CString *pstr = (CString *)dwCookie;

	if( pstr->GetLength() < cb )
	{
		*pcb = pstr->GetLength();
		memcpy(pbBuff, (LPCSTR)*pstr, *pcb );
		pstr->Empty();
	}
	else
	{
		*pcb = cb;
		memcpy(pbBuff, (LPCSTR)*pstr, *pcb );
		*pstr = pstr->Right( pstr->GetLength() - cb );
	}
	return 0;
}

CString rtfString;
void CDVDSubEditDlg::DispInfo(char* Info, int Color)
{
	CString Temp; 
	if(Info[1] == 'C')
		Temp.Format("{\\ul\\cf%d %s}",Color,Info);
	else
		Temp.Format("\\cf%d %s",Color,Info);
	rtfString += Temp;
	rtfString += "\\par ";
}

void CDVDSubEditDlg::ViewSPU(int SPUIndex, bool RefreshVideo, bool RefreshCommands)
{
//	if(SPUIndex < m_PruneLength && m_SelectedSubPic!= 0)

	if(SPUIndex < m_PruneLength)
	{
		char Temp[1024*16];
		int SelectedPixel;
		m_SelectedSubPic = m_PrunedList[SPUIndex];
		// I'm keeping a local copy here to avoid m_SelectedSubPic becoming 0 under me, which sometimes happens,
		// when the video display is on, and I've no idea why >:(
		CSubPic* SelectedSubPic = m_SelectedSubPic;
		ASSERT(SelectedSubPic != 0);
		if(RefreshCommands)
		{
			// Now show commands!
			// http://netghost.narod.ru/gff/graphics/summary/micrtf.htm
			CString rtfPrefix, rtfPostfix;
			rtfPrefix = "{\\rtf1\\ansi\\deff0\\deftab720{\\fonttbl{\\f0\\froman "
				"Times New Roman;}}\n{\\colortbl\\red0\\green0\\blue0;\\red255\\green0\\blue0;\\red0\\green134\\blue37;\\red0\\green0\\blue255;}\n"
				"\\deflang1033\\pard\\tx360\\tx720\\tx1080\\tx1440\\tx1800"
				"\\tx2160\\tx2520\\tx2880\\tx3240\\tx3600\\tx3960\\tx4320"
				"\\tx4680\\tx5040\\tx5400\\tx5760\\tx6120"
				"\\tx6480\\plain\\f3\\fs20 ";
			rtfString = rtfPrefix ;
			// Display various info.
			sprintf(Temp,"\\b\\ul Subpic# = %d\\b0\\ul0\n",SelectedSubPic->m_SPUNumber); DispInfo(Temp,BLUE);
			sprintf(Temp,"StreamID = 0x%x\n",SelectedSubPic->m_StreamID+0x20); DispInfo(Temp,BLUE);
			sprintf(Temp,"LBA = %d\n",SelectedSubPic->m_LBAFrom[0]); DispInfo(Temp,BLUE);
			sprintf(Temp,"V/CID = %d/%d\n",SelectedSubPic->m_VOBID,SelectedSubPic->m_CELLID); DispInfo(Temp,BLUE);
			sprintf(Temp,"PTS = %d (%.2fs)\n",SelectedSubPic->m_PTS,(float)SelectedSubPic->m_PTS/90000); DispInfo(Temp,BLUE);
			if(SelectedSubPic->m_DCSQT.size() == 1)
			{SetDurationSlider(0);sprintf(Temp,"Duration = infinite\n"); DispInfo(Temp,BLUE);}
			else
			{
				int DurationMs = SelectedSubPic->GetDCSQTDelay();
				SetDurationSlider(DurationMs);
				sprintf(Temp,"Duration = %.3fs\n",(float)DurationMs/1000); DispInfo(Temp,BLUE);
			}
			if(SelectedSubPic->m_Modified)
				DispInfo("Modified\n",GREEN);
			while(RetreiveNextCommand(Temp) == 0)
				DispInfo(Temp,(Temp[0] != ' ')?RED:BLACK);
			rtfString += "\n\\par }";
			EDITSTREAM es = {(DWORD)&rtfString, 0, EditStreamCallBack};
			m_CommandDisplay.StreamIn(SF_RTF, es);
		}
		SelectedSubPic->DecodeSubpic();
		SetMoveVSlider(SelectedSubPic->m_LastVMove+MAX_V_MOVE);
		SetMoveHSlider(SelectedSubPic->m_LastHMove+MAX_H_MOVE);
		SelectedPixel = m_PixelSelect.GetCurSel();
		// The alphas are between 0 and 255;
		if(SelectedPixel < 4)
			SetTransSlider(SelectedSubPic->m_AlphaPtr[SelectedPixel]);
		else
			SetTransSlider(SelectedSubPic->GetMeanTransparency());
		// Start command.
		m_StartCommand.SetCurSel(SelectedSubPic->FindCommand(STA_DSP) != -1 ? 0 : 1);

		// Borders.
		m_BorderSlider.EnableWindow(!SelectedSubPic->m_HasButtons);
		m_BorderSelect.EnableWindow(!SelectedSubPic->m_HasButtons);
		GetDlgItem(IDC_BORDERB)->EnableWindow(!SelectedSubPic->m_HasButtons);

		SetBorderSlider(SelectedSubPic->m_VerticalBorders);
		if(SelectedSubPic->m_VerticalBorders == (FORMAT_NTSC ? 65 : 78))
			SetBorderSelect(BORDER_4_3_LS);
		else if(SelectedSubPic->m_VerticalBorders == (FORMAT_NTSC ? 60 : 72))
			SetBorderSelect(BORDER_4_3_L);
		else if(SelectedSubPic->m_VerticalBorders == (FORMAT_NTSC ? 38 : 46))
			SetBorderSelect(BORDER_SAFE);
		else if(SelectedSubPic->m_VerticalBorders)
			SetBorderSelect(BORDER_CUSTOM);
		else 
			SetBorderSelect(-1);

		// DCSQT:
		m_DCSQTSelect.ResetContent();
		for(int i=0;i<SelectedSubPic->m_DCSQT.size()-1;i++)
		{
			sprintf(Temp,"DCSQT %d",i);
			m_DCSQTSelect.AddString(Temp);
		}
		sprintf(Temp,"DCSQT %d",i);
		if(SelectedSubPic->m_DCSQT.back().m_Commands.size() > 2)
			m_DCSQTSelect.AddString(Temp);
		m_DCSQTSelect.SetCurSel(SelectedSubPic->m_SelectedDCSQT);

		m_Revert.EnableWindow(SelectedSubPic->m_Modified && !m_Running);

		bool CanHMove = SelectedSubPic->CanHMove();
		m_MoveHSlider.EnableWindow(CanHMove);
		GetDlgItem(IDC_MOVELRE)->EnableWindow(CanHMove);
		GetDlgItem(IDC_MOVELRB)->EnableWindow(CanHMove);

		if(!RefreshVideo)
			return;
		
		m_OCROut.SetWindowText("");
		for(i=0;;i++)
		{
			if(SelectedSubPic->m_OCRLines[i][0] == '\0') break;
			m_OCROut.SetSel(1000,1000);
			if(SelectedSubPic->m_OCRLines[i+1][0] != '\0')
				m_OCROut.ReplaceSel(CatStr((char*)SelectedSubPic->m_OCRLines[i],"\n"));
			else
				m_OCROut.ReplaceSel((char*)SelectedSubPic->m_OCRLines[i]);
		}
#if 0
		if(i)
		{
			CHARFORMAT cf;
			cf.dwMask = CFM_FACE;
			strcpy(cf.szFaceName,"Times New Roman");
//			strcpy(cf.szFaceName,"Courier New");
//			strcpy(cf.szFaceName,"Arial");
			m_OCROut.SetSel(0,-1);
			m_OCROut.SetSelectionCharFormat(cf);
		}
#endif

		// Now's the time to display the video, if needed, then the subpics.
		if(!m_DisplayVideo || m_NoVideo)
		{
			// No background video display. SImply display the bitmap.
			::ShowWindow(m_PreviewHandle,SW_HIDE);
			DisplayBitmap(0,0,0);
		}
		else
		{
			static int LastLBA = -1;
			// If we're displaying the same image, there's no need to seek to it and re-decode it.
			if(LastLBA != SelectedSubPic->m_NavPackLBA)
			{
				// When showing the video, the preview is responsible for drawing
				// everything. ShowBorders() will be called after the frame is drawn.
				TRACE("Showing video %d\n",SelectedSubPic->m_NavPackLBA);
				m_Preview->GoToSector(SelectedSubPic->m_NavPackLBA);
				m_Preview->WaitForThread();
				::ShowWindow(m_PreviewHandle,SW_SHOW);
				LastLBA = SelectedSubPic->m_NavPackLBA;
			}
			else
			{
				// Same image, just repaint the frame! This will display our subpic.
				m_Preview->RePaintVideo();
				::ShowWindow(m_PreviewHandle,SW_SHOW);
			}
		}
	}
	else
	{
		m_SelectedSubPic = 0;
		// Just to erase the background...
		DisplayBitmap(0,0,0);
	}
}

void CDVDSubEditDlg::OnDestroy() 
{
	KillTimer(m_Timer);
	m_AbortNow = m_QuitNow = true;
	WaitForSingleObject(m_ThreadHandle,1000);

	// Save info...
	char Temp[1024];
	sprintf(Temp,"%s/DVDSubEditLastFile.txt",getenv("APPDATA"));
	SaveSubpicInfo(Temp);

	m_ScanDialog->DestroyWindow();
	delete m_ScanDialog;
	m_ApplyAllDialog->DestroyWindow();
	delete m_ApplyAllDialog;
	m_MoveDialog->DestroyWindow();
	delete m_MoveDialog;
	m_Preview->CloseFile();
	delete m_Preview;
	m_Menu.DestroyMenu();
	if(OutLogFd) fclose(OutLogFd);
	delete[] m_OCRBitmap;
	delete[] m_OCRLines;
	CDialog::OnDestroy();
}

void BeforeFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	Dialog->m_Timer = Dialog->SetTimer(0,200,NULL);
	Dialog->DragAcceptFiles(FALSE);
	Dialog->m_Running = 1;
	Dialog->m_ScanDialog->ShowWindow(SW_SHOW);
	Dialog->m_AbortNow = 0;
	Dialog->EnableUI(false);
	Dialog->m_MenuEnabled = false;
}

void AfterFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	if(!Dialog->m_QuitNow)
	{
		Dialog->KillTimer(Dialog->m_Timer);
		Dialog->DragAcceptFiles(TRUE);
		Dialog->EnableUI(true);
		Dialog->m_ScanDialog->ShowWindow(SW_HIDE);
		Dialog->m_ScanDialog->DisplayInfo(0,"","");
		Dialog->m_Phase = Processing_Idle;
		Dialog->SetMoveVSlider(MAX_V_MOVE);
		Dialog->ViewSPU(Dialog->m_SPUSlider.GetPos());
		Dialog->m_MenuEnabled = true;
		Dialog->m_PercentDone = 0;
	}
	Dialog->m_Running = 0;
	Dialog->m_ThreadHandle = 0;
}

UINT DVDSubEditLoadLastFileFunc( LPVOID pParam )
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);
	Dialog->LoadSubpicInfo(Dialog->m_LastScanFile);
	AfterFunc(pParam);
	if(!Dialog->m_QuitNow)
	{
		char Temp[512];
		sprintf(Temp,"%s ... -- %d Subpics found",Dialog->m_FileInfos[0].m_FileName,Dialog->m_TotNumSPU);
		Dialog->SetWindowText(Temp);
		Dialog->OnSelchangeStreamselect();
		Dialog->EnableUI(true);
		Dialog->SetForegroundWindow();
	}
	return 0;
}

UINT DVDSubEditScanFunc( LPVOID pParam )
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->ScanFilesForSPUs();

	AfterFunc(pParam);

	if(!Dialog->m_QuitNow)
	{
		char Temp[512];
		sprintf(Temp,"%s ... -- %d Subpics found",Dialog->m_FileInfos[0].m_FileName,Dialog->m_TotNumSPU);
		Dialog->SetWindowText(Temp);
		Dialog->OnSelchangeStreamselect();
		Dialog->SetForegroundWindow();
		Sleep(1500);
	}
	return 0;
}

UINT ApplyToAllFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->m_Phase = Processing_Applying;
	int VMoveby = Dialog->m_ChangeRecord.m_VMove;
	int HMoveby = Dialog->m_ChangeRecord.m_HMove;
	int Border = Dialog->m_ChangeRecord.m_Border;
	int FlushIt = Dialog->m_FlushBorders.GetCheck();
	int Transparency = Dialog->m_ChangeRecord.m_Transparency;
	int PixSelect = Dialog->m_ChangeRecord.m_TranspPixel;
	int Color = Dialog->m_ChangeRecord.m_Color;
	int PixSelectC = Dialog->m_ChangeRecord.m_ColorPixel;
	uchar* CLUT = Dialog->m_ChangeRecord.m_CLUT;
	int Duration = Dialog->m_ChangeRecord.m_Duration;
	int DurMinMax = Dialog->m_ApplyAllDialog->m_MinMax;
	int StartSelect = Dialog->m_ChangeRecord.m_StartCommand;
	int i;
	for(i=0;i<Dialog->m_PruneLength;i++)
	{
		if(Dialog->m_AbortNow)
			break;
		// Check which of the modifications should be applied to all..
		if(Dialog->m_ApplyAllDialog->m_ApplyBorders)
			Dialog->m_PrunedList[i]->VMoveSubPicInFrame(Border,FlushIt);
		if(Dialog->m_ApplyAllDialog->m_ApplyVMove)
			Dialog->m_PrunedList[i]->VMoveSubPic(VMoveby);
		// Make sure to crop the background BEFORE you try to move horizontally.
		if(Dialog->m_ApplyAllDialog->m_ApplyCrop)
			Dialog->m_PrunedList[i]->CropBackground();
		if(Dialog->m_ApplyAllDialog->m_ApplyHMove)
			Dialog->m_PrunedList[i]->HMoveSubPic(HMoveby);
		if(Dialog->m_ApplyAllDialog->m_ApplyTransp)
			Dialog->m_PrunedList[i]->ChangeTransparency(Transparency,PixSelect);
		if(Dialog->m_ApplyAllDialog->m_ApplyColor && CLUT == Dialog->m_PrunedList[i]->m_CLUT)
			Dialog->ChangeColors(Color,PixSelectC,Dialog->m_PrunedList[i]);
		if(Dialog->m_ApplyAllDialog->m_ApplyDuration)
		{
			if(DurMinMax == 0) // Set exact duration.
				Dialog->m_PrunedList[i]->ChangeDCSQTDelay(Duration);
			else if(DurMinMax == 1 && Dialog->m_PrunedList[i]->GetDCSQTDelay() < Duration) // At least.
				Dialog->m_PrunedList[i]->ChangeDCSQTDelay(Duration);
			else if(DurMinMax == 2 && Dialog->m_PrunedList[i]->GetDCSQTDelay() > Duration) // At most.
				Dialog->m_PrunedList[i]->ChangeDCSQTDelay(Duration);
		}
		if(Dialog->m_ApplyAllDialog->m_ApplyStart)
		{
			if(StartSelect == 0)
				Dialog->m_PrunedList[i]->ChangeCommand(FSTA_DSP,STA_DSP);
			else
				Dialog->m_PrunedList[i]->ChangeCommand(STA_DSP,FSTA_DSP);
		}
		Dialog->m_PercentDone = i*100/Dialog->m_PruneLength;
	}
	Dialog->m_ChangeRecord.Reset();
	AfterFunc(pParam);
	return 0;
}

UINT RevertAllFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->m_Phase = Processing_Reverting;
	Dialog->m_ScanDialog->SetWindowText("Reverting changes");
	sprintf(Message,"Reverting changes");
	int i;
	for(i=0;i<Dialog->m_PruneLength;i++)
	{
		Dialog->m_PrunedList[i]->ReloadSPU();
		Dialog->m_PercentDone = i*100/Dialog->m_PruneLength;
	}
	Dialog->UpdateSPUList();
	AfterFunc(pParam);
	return 0;
}

UINT SaveAllFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->m_Phase = Processing_Saving;
	Dialog->m_ScanDialog->SetWindowText("Saving changes");
	sprintf(Message,"Saving changes");
	int i;
	for(i=0;i<Dialog->m_PruneLength;i++)
	{
		if(Dialog->m_PrunedList[i]->m_Modified)
			Dialog->m_PrunedList[i]->SaveSPU();
		Dialog->m_PercentDone = i*100/Dialog->m_PruneLength;
	}
	AfterFunc(pParam);
	return 0;
}

UINT RetimeAllFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->m_Phase = Processing_ReTiming;
	Dialog->m_ScanDialog->SetWindowText("ReTiming subpics");
	sprintf(Message,"ReTiming subpics");
	Dialog->MoveSelection(Dialog->m_MoveDialog->m_ChangeTimingBy);
	Dialog->UpdateSPUList();
	AfterFunc(pParam);
	return 0;
}

UINT RunOCRFunc(LPVOID pParam)
{
	CDVDSubEditDlg* Dialog = (CDVDSubEditDlg*) pParam;
	BeforeFunc(pParam);

	Dialog->m_ScanDialog->SetWindowText("Running OCR");
	int i;
	for(i=0;i<Dialog->m_PruneLength;i++)
	{
		Dialog->m_PercentDone = 100*i/Dialog->m_PruneLength;
		if(Dialog->m_AbortNow) break;
		Dialog->m_PrunedList[i]->DecodeSubpic();
		Dialog->DoRunOCR(Dialog->m_PrunedList[i]);
	}
	// This to update the OCR status...
	Dialog->UpdateSPUList();
	AfterFunc(pParam);
	return 0;
}

void CDVDSubEditDlg::StartScan() 
{
	if(m_Running == 0)
	{
		m_AbortNow = 0;
		EnableUI(false);
		OutLogFd = fopen(m_LogPath,"w");
		CWinThread *Temp = AfxBeginThread( DVDSubEditScanFunc, this);
		m_ThreadHandle = Temp->m_hThread;
		m_Running = 1;
	}
	else
		m_AbortNow = 1;
}

void CDVDSubEditDlg::OCRThisOne()
{
	if(!m_SelectedSubPic)
		return;
	m_SelectedSubPic->DecodeSubpic();
	DoRunOCR(m_SelectedSubPic); 
	m_OCROut.SetWindowText("");
	for(int i=0;;i++)
	{
		if(m_SelectedSubPic->m_OCRLines[i][0] == '\0') break;
		m_OCROut.SetSel(1000,1000);
		if(m_SelectedSubPic->m_OCRLines[i+1][0] != '\0')
			m_OCROut.ReplaceSel(CatStr((char*)m_SelectedSubPic->m_OCRLines[i],"\n"));
		else
			m_OCROut.ReplaceSel((char*)m_SelectedSubPic->m_OCRLines[i]);
	}
}

void CDVDSubEditDlg::RunOCROnSelected()
{
	if(m_Running)
		return;
	EnableUI(false,true);
	m_MenuEnabled = false;
	m_Running = true;
	MESSAGE1("");
	m_Phase = Processing_RunningOCR;
	CWinThread *Temp = AfxBeginThread( RunOCRFunc, this);
	m_ThreadHandle = Temp->m_hThread;
}

void CDVDSubEditDlg::LoadLastFile()
{
	if(WarnIfModifs() == false)
		return;
	sprintf(m_LastScanFile,"%s/DVDSubEditLastFile.txt",getenv("APPDATA"));
	if(m_Running == 0)
	{
		m_AbortNow = 0;
		EnableUI(false);
		OutLogFd = fopen(m_LogPath,"w");
		CWinThread *Temp = AfxBeginThread( DVDSubEditLoadLastFileFunc, this);
		m_ThreadHandle = Temp->m_hThread;
		m_Running = 1;
	}
	else
		m_AbortNow = 1;
}

void CDVDSubEditDlg::OnOpenIFO()
{
	if(!m_SupFileMode)
		return;
	static CFileDialog FileDialog(TRUE, CString("*.IFO"), "", OFN_FILEMUSTEXIST, CString("IFO Files (*.IFO)|*.IFO||"));
	FileDialog.m_ofn.lpstrTitle = "Load IFO file";
	if (FileDialog.DoModal() == IDOK)
	{
		OpenIFO((LPCTSTR)FileDialog.GetPathName());
		m_AutoCLUT.SetCheck(false);
		m_IgnoreCLUT.SetCheck(false);
	}
}


void CDVDSubEditDlg::OnApplytoall() 
{
	if(m_Running) return;
	if(m_ApplyAllDialog->DoModal() != IDOK)
		return;
	CWinThread *Temp = AfxBeginThread( ApplyToAllFunc, this);
	m_ThreadHandle = Temp->m_hThread;
	m_Running = 1;
}

void CDVDSubEditDlg::OnSaveall() 
{
	if(m_Running || !m_CanSave) return;
	if(AfxMessageBox("This will save all your changes to file. No undo...\nProceed?",MB_OKCANCEL) != IDOK)
		return;
	CWinThread *Temp = AfxBeginThread( SaveAllFunc, this);
	m_ThreadHandle = Temp->m_hThread;
	m_Running = 1;
}


void CDVDSubEditDlg::OnRevertall() 
{
	if(m_Running || !m_CanReload) return;
	if(AfxMessageBox("This will undo all your changes... Proceed?",MB_OKCANCEL) != IDOK)
		return;
	CWinThread *Temp = AfxBeginThread( RevertAllFunc, this);
	m_ThreadHandle = Temp->m_hThread;
	m_ShowModified.SetCheck(false);
	m_ChangeRecord.Reset();
	m_Running = 1;
}

void CDVDSubEditDlg::OnTimingAll() 
{
	static bool First = true;
	m_MoveDialog->m_RetimeAll = true;
	if(!m_CanMove)
	{
		AfxMessageBox("Retiming subtitles is only possible in the title domain\nand if the entire domain is loaded.\
\nNot available for .sup files");
		return;
	}
	if(m_SelectedSubPic == 0 || m_MoveDialog->DoModal() != IDOK)
		return;
	if(First && AfxMessageBox("There is NO undo for this\nThe VOB file will be modified. Proceed?",MB_OKCANCEL) != IDOK)
		return;
	First = false;
	EnableUI(false,true);
	CWinThread *Temp = AfxBeginThread(RetimeAllFunc, this);
	m_ThreadHandle = Temp->m_hThread;
	m_Running = 1;
}

void CDVDSubEditDlg::OnFileLoaded(const char* FileName)
{
	m_Preview->LoadFile(FileName);
}

void CDVDSubEditDlg::HideHearingImpaired()
{
	int i;
	for(i=0;i<m_PruneLength;i++)
	{
		if(m_PrunedList[i]->IsTextBracketed())
			m_PrunedList[i]->ChangeTransparency(0,5);
	}
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::EnableUI(bool Enable, bool Full)
{
	GetDlgItem(IDC_APPLYTOALL)->EnableWindow(Enable);
	GetDlgItem(IDC_REVERTALL)->EnableWindow(Enable);
	GetDlgItem(IDC_REVERTSAVED)->EnableWindow(Enable);
	GetDlgItem(IDC_SAVEALL)->EnableWindow(Enable);
	GetDlgItem(IDC_RUNOCR)->EnableWindow(Enable);
	GetDlgItem(IDC_SAVEOCR)->EnableWindow(Enable && m_ListHasOCR);
	GetDlgItem(IDC_RETIMEONE)->EnableWindow(Enable);
	GetDlgItem(IDC_RETIMEALL)->EnableWindow(Enable);
	GetDlgItem(IDC_FINDL)->EnableWindow(Enable);
	GetDlgItem(IDC_FINDR)->EnableWindow(Enable);
	GetDlgItem(IDC_SHOWVIDEO)->EnableWindow(!m_NoVideo);
	if(Full || Enable)
	{
		GetDlgItem(IDC_STREAMSELECT)->EnableWindow(Enable);
		GetDlgItem(IDC_PSTREAM)->EnableWindow(Enable);
		GetDlgItem(IDC_NSTREAM)->EnableWindow(Enable);
		GetDlgItem(IDC_VCIDSELECT)->EnableWindow(Enable);
		GetDlgItem(IDC_PVCID)->EnableWindow(Enable);
		GetDlgItem(IDC_NVCID)->EnableWindow(Enable);
		GetDlgItem(IDC_SHOWMOD)->EnableWindow(Enable);
		GetDlgItem(IDC_SHOWSEARCHED)->EnableWindow(Enable);
		GetDlgItem(IDC_SEARCHSELECT)->EnableWindow(Enable);
		GetDlgItem(IDC_DURATIONS)->EnableWindow(Enable);
		GetDlgItem(IDC_DURATIONE)->EnableWindow(Enable);
	}
}

void CDVDSubEditDlg::OnTimer(UINT nIDEvent) 
{
	static int Count = 0;

	if(m_QuitNow == 1)
		return;

	// Update progress bar... TODO: FIX THIS MESS!
	char Temp[512];
	switch(m_Phase)
	{
	case Processing_Idle : sprintf(Temp,"");break;
	case Processing_Scanning_VOB : 	sprintf(Temp,"Scanning VOB file(s)"); break;
	case Processing_Done : sprintf(Temp,"Done...");break;
	case Processing_Aborted : sprintf(Temp,"Aborted");break;
	case Processing_Applying : sprintf(Temp,"Applying modifications");break;
	case Processing_Reverting : sprintf(Temp,"Reverting");break;
	case Processing_Saving : sprintf(Temp,"Saving modifs");break;
	case Processing_ReTiming : sprintf(Temp,"Re-timing");break;
	case Processing_UpdatingSRI : sprintf(Temp,"Updating trick-play pointers");break;
	case Processing_RunningOCR : sprintf(Temp,"Running OCR... (GOCR) by Joerg Schulenburg");break;
	case Processing_FastReloading : sprintf(Temp,"Reloading last file");break;
	case Processing_Cropping : sprintf(Temp,"Cropping subpic background"); break;
	}
	if(m_Phase != Processing_Idle)
		m_ScanDialog->DisplayInfo(m_PercentDone,Temp,Message);
	if(m_Running && Count++ % 5 == 0)
	{
		Count = 1;
		if(m_TotNumSPU < 2)
			OnSelchangeStreamselect();
		else
			UpdateSPUList();
	}

	CDialog::OnTimer(nIDEvent);
}

//////////////////////////////////////////////////////////////////////////////////////////////
void CDVDSubEditDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.m_Edit = this;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CDVDSubEditDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else if(!m_DisplayVideo || m_NoVideo)
	{
		CPaintDC dc(this);
		// We're not displaying any video, so we have to draw our bitmap ourself!
		// Only do it if it's needed!
		if(RectVisible(dc.m_hDC,&m_VideoRect))
		{
	TRACE("ONPAINT\n");
			CDialog::OnPaint();
			PAINTSTRUCT Ps;
			BeginPaint(&Ps);
			DisplayBitmap(0,0,0);
			EndPaint(&Ps);
		}
	}
	else
		CDialog::OnPaint();
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CDVDSubEditDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}


#define MAX_NUM_FILES 64
void CDVDSubEditDlg::OnDropFiles(HDROP hDropInfo) 
{
	if(WarnIfModifs() == false)
		return;
	SetForegroundWindow();      // activate us first !
	POINT ThePoint;
	DragQueryPoint(hDropInfo,&ThePoint);

	UINT nFiles = ::DragQueryFile(hDropInfo, (UINT)-1, NULL, 0);
	CWinApp* pApp = AfxGetApp();
	ASSERT(pApp != NULL);
	char FileNames[MAX_NUM_FILES][512];
	int i;
	for(i=0;i<MAX_NUM_FILES;i++)
		strcpy(FileNames[i],"");
	int VTSNum = 0;
	for (UINT iFile = 0; iFile < nFiles; iFile++)
	{
		TCHAR szFileName[_MAX_PATH];
		::DragQueryFile(hDropInfo, iFile, szFileName, _MAX_PATH);
		if(strcmpi(GetExt(szFileName),".sup") == 0)
		{
			// Special code for sup files.
			if(LoadSUPFile(szFileName) != 0)
				return;
			OnSelchangeStreamselect();
			EnableUI(true);
			SetWindowText(szFileName);
			::DragFinish(hDropInfo);
			m_AutoCLUT.SetCheck(true);
			return;
		}
		if(strcmpi(GetExt(szFileName),".vob") != 0)
		{
			AfxMessageBox("Error: I can only open VOB files");
			::DragFinish(hDropInfo);
			return;
		}
		int Index = -1;
		int VTS = -1;
		// Get VTS number and index.
		if(sscanf(GetFname(szFileName),"%*[^_]_%d_%d",&VTS,&Index) != 2)
		{
			// Probably not a "vts_xx_xx", let's just add to our list at the next index...
			Index = (int)iFile;
			VTSNum = 0;
		}
		if(VTSNum == 0) VTSNum = VTS;
		else if(VTSNum != VTS)
		{
			AfxMessageBox("All the VOB files must come from the same VTS");
			::DragFinish(hDropInfo);
			return;
		}
		ASSERT(Index < 10 && Index >= 0);
		strcpy(FileNames[Index],szFileName);
	}
	::DragFinish(hDropInfo);
	
	// Now put these file names in order...
	Reset();
	for(i=0;i<MAX_NUM_FILES;i++)
	{
		if(strcmp(FileNames[i],"") == 0)
			continue;
		if(AddFile(FileNames[i]) == -1)
		{
			char Temp[512];
			sprintf(Temp,"Error opening %s\nIs the file writeable?",FileNames[i]);
			AfxMessageBox(Temp);
			return;
		}
		else
		{
			SetWindowText(FileNames[i]);
		}
	}
	EnableUI(true);

	// Now scan input files...
	m_ScanDialog->DisplayInfo(0,"","");
	m_ScanDialog->ShowWindow(SW_SHOW);
	StartScan();
	m_Preview->LoadFile(m_FileInfos[0].m_FileName);
}


HBRUSH CDVDSubEditDlg::OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor) 
{
	HBRUSH hbr = CDialog::OnCtlColor(pDC, pWnd, nCtlColor);
	int id = pWnd->GetDlgCtrlID();
	//COLORREF bkcolor = pDC->GetBkColor();
	if (id == IDC_INFO || id == IDC_MESSAGE)
	{
		// Set the text color to black
		pDC->SetTextColor(RGB(0, 0, 0)); 
	}	
	// TODO: Return a different brush if the default is not desired
	return hbr;
}

void CDVDSubEditDlg::OnSpur() 
{
	m_SPUSlider.SetPos(m_SPUSlider.GetPos()+1);
	ViewSPU(m_SPUSlider.GetPos()); 
}

void CDVDSubEditDlg::OnSpul() 
{
	m_SPUSlider.SetPos(m_SPUSlider.GetPos()-1);
	ViewSPU(m_SPUSlider.GetPos()); 
}

void CDVDSubEditDlg::OnCropSubpic()
{
	if(!m_SelectedSubPic)
		return;
	m_SelectedSubPic->CropBackground();
	ViewSPU(m_SPUSlider.GetPos());
	m_ChangeRecord.RecordChange(m_SelectedSubPic,ChangeCrop,1);
}

void CDVDSubEditDlg::JustAddedSPU(int StreamIndex, int SPUIndex)
{
	if(m_QuitNow)
		return;
	if(m_TotNumSPU == 1)
		OnSelchangeStreamselect();
//m_MasterList[m_TotNumSPU-1]->DecodeSubpic();
//DoRunOCR(m_MasterList[m_TotNumSPU-1]);
}

void CDVDSubEditDlg::JustFoundStream(int StreamIndex)
{
	if(m_QuitNow)
		return;
	if(StreamIndex+1 < m_StreamSelect.GetCount())
		return;
	m_StreamSelect.ResetContent();
	m_StreamSelect.AddString("All Streams");
	int i;
	for(i=0;i<=m_TopStreamID;i++)
	{
		char Temp[16];
		sprintf(Temp,"0x%x",i+32);
		m_StreamSelect.AddString(Temp);
	}
	m_StreamSelect.SetCurSel(m_SupFileMode?1:m_SelectedStream);
}

void CDVDSubEditDlg::JustFoundVCID(uint VCID)
{
	if(m_QuitNow)
		return;
	char Temp[64];
	sprintf(Temp,"%d/%d",VCID/256,VCID%256);
	m_VCIDSelect.AddString(Temp);
}

void CDVDSubEditDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	switch(pScrollBar->GetDlgCtrlID())
	{
	case IDC_SPUSLIDER : {
		static int LastPos = 0;
		if(LastPos == m_SPUSlider.GetPos())
			break;
		LastPos = m_SPUSlider.GetPos();
		ViewSPU(LastPos); 
		break;}
	case IDC_MOVEUDS : 
		DispThis2(IDC_MOVEUDE,"%d",m_MoveVSlider.GetPos()-MAX_V_MOVE);
		if(m_SelectedSubPic)// && nSBCode == SB_ENDSCROLL) 
		{
			m_SelectedSubPic->VMoveSubPic(m_MoveVSlider.GetPos()-MAX_V_MOVE);
			// Be smart: if we move the move slider, we probably want to apply that!
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeVMove, m_MoveVSlider.GetPos()-MAX_V_MOVE);
			ViewSPU(m_SPUSlider.GetPos());
		}
		break;
	case IDC_MOVELR : 
		DispThis2(IDC_MOVELRE,"%d",m_MoveHSlider.GetPos()-MAX_H_MOVE);
		if(m_SelectedSubPic)// && nSBCode == SB_ENDSCROLL) 
		{
			m_SelectedSubPic->HMoveSubPic(m_MoveHSlider.GetPos()-MAX_H_MOVE);
			// Be smart: if we move the move slider, we probably want to apply that!
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeHMove, m_MoveHSlider.GetPos()-MAX_H_MOVE);
			ViewSPU(m_SPUSlider.GetPos());
		}
		break;
	case IDC_TRANSPARENCYS : 
		if(m_SelectedSubPic)// && nSBCode == SB_ENDSCROLL) 
		{
			int SelectedPixel = m_PixelSelect.GetCurSel();
			int CurTrans = m_TransSlider.GetPos();
			// Be smart: if we move the transparency slider, we probably want to apply that!
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeTransp, CurTrans,SelectedPixel);
			if(m_SelectedSubPic->ChangeTransparency(CurTrans,SelectedPixel) == 0)
				ViewSPU(m_SPUSlider.GetPos());
		}
		break;
	case IDC_BORDERS : 
		if(m_SelectedSubPic)
		{
			int TopBorder = m_BorderSlider.GetPos();
			DispThis2(IDC_BORDERE,"%d",TopBorder);
			m_SelectedSubPic->VMoveSubPicInFrame(TopBorder,m_FlushBorders.GetCheck());
			ViewSPU(m_SPUSlider.GetPos());
			if(TopBorder == 0)
				SetBorderSelect(-1);
			else 
				SetBorderSelect(BORDER_CUSTOM);
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, TopBorder);

		}
		break;
	case IDC_DURATIONS : 
		if(m_SelectedSubPic)
		{
			int Duration = m_DurationSlider.GetPos();
			m_SelectedSubPic->ChangeDCSQTDelay(Duration);
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeDur, m_SelectedSubPic->GetDCSQTDelay());
			ViewSPU(m_SPUSlider.GetPos(),false);
		}
		break;
	default: break;
	}
	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}

void CDVDSubEditDlg::SetMoveVSlider(int Pos)
{
	m_MoveVSlider.SetPos(Pos);
	DispThis2(IDC_MOVEUDE,"%d",Pos-MAX_V_MOVE);
}

void CDVDSubEditDlg::SetMoveHSlider(int Pos)
{
	m_MoveHSlider.SetPos(Pos);
	DispThis2(IDC_MOVELRE,"%d",Pos-MAX_H_MOVE);
}

void CDVDSubEditDlg::SetTransSlider(int Pos)
{
	m_TransSlider.SetPos(Pos);
	DispThis2(IDC_TRANSPARENCYE,"%d",Pos);
}

void CDVDSubEditDlg::SetBorderSlider(int Pos)
{
	m_BorderSlider.SetPos(Pos);
	DispThis2(IDC_BORDERE,"%d",Pos);
}

void CDVDSubEditDlg::SetDurationSlider(int Pos)
{
	if(m_SelectedSubPic && m_SelectedSubPic->m_DCSQT.size() > 1)
	{
		// Find timing of next to last DCSQT
		int MinDurMs, MaxDurMs;
		m_SelectedSubPic->GetMinMaxDCSQTDelay(MinDurMs, MaxDurMs);
		m_DurationSlider.SetRange(MinDurMs,MaxDurMs,true);
		m_DurationSlider.EnableWindow(true);
	}
	else
	{
		m_DurationSlider.EnableWindow(false);
	}
	m_DurationSlider.SetPos(Pos);
	DispThis2(IDC_DURATIONE,"%.3f",(float)Pos/1000);
}


void CDVDSubEditDlg::SetBorderSelect(int Selection)
{
	if(Selection == 0)
	{
		// Custom borders.
		m_BorderSelect.DeleteString(0);
		m_BorderSelect.InsertString(0,"Custom Vertical Borders");
		m_BorderSelect.SetCurSel(Selection);
	}
	else if(Selection > 0)
	{
		// Standard borders.
		m_BorderSelect.DeleteString(0);
		m_BorderSelect.InsertString(0,"No Vertical Borders");
		m_BorderSelect.SetCurSel(Selection);
	}
	else
	{
		// -1 means, no border. Change the string name!
		m_BorderSelect.DeleteString(0);
		m_BorderSelect.InsertString(0,"No Vertical Borders");
		m_BorderSelect.SetCurSel(0);
	}
}

void CDVDSubEditDlg::OnSelchangeStreamselect() 
{
	m_ShowSearched.SetCheck(false);
	UpdateSPUList();
	if(!m_PruneLength)
		m_CommandDisplay.SetWindowText("");
}

void CDVDSubEditDlg::OnShowmod() 
{
	UpdateSPUList();
}

void CDVDSubEditDlg::OnShowsearched() 
{
	UpdateSPUList();
}

void CDVDSubEditDlg::OnKillfocusSearchselect() 
{
	CString Temp;
	GetDlgItemText(IDC_SEARCHSELECT,Temp);
	bool Active =  Temp != "";
	m_ShowSearched.SetCheck(Active);
	UpdateSPUList();
}

void CDVDSubEditDlg::OnKillfocusSearch() 
{
	OnFindRight();	
}

void CDVDSubEditDlg::UpdateSPUList()
{
	// User changed the selected stream. Create the pruned list.
	int ThisPos = m_SPUSlider.GetPos();
	int SPUIndex = m_SelectedSubPic ? m_SelectedSubPic->m_SPUNumber : -1;
	int NextPos = 0;
	m_SelectedStream = m_StreamSelect.GetCurSel();
	int CurrentVCID = m_VCIDSelect.GetCurSel();
	char SearchStr[128];
	GetDlgItemText(IDC_SEARCHSELECT,SearchStr,128);
	if(CurrentVCID == 0)
		NextPos = PruneSPUList(ThisPos,m_SelectedStream-1,-1,-1,0,0,m_ShowModified.GetCheck(),
		m_ShowSearched.GetCheck()?SearchStr: 0);
	else
		NextPos = PruneSPUList(ThisPos,m_SelectedStream-1,-1,-1,m_VCIDList[CurrentVCID-1]/256,
			m_VCIDList[CurrentVCID-1]%256,m_ShowModified.GetCheck(),
			m_ShowSearched.GetCheck()?SearchStr : 0);

	m_SPUSlider.SetRange(0,m_PruneLength-1,true);
	m_SPUSlider.ShowWindow((m_PruneLength > 1)?SW_SHOW:SW_HIDE);
	if(NextPos != -1)
		m_SPUSlider.SetPos(NextPos);
	GetDlgItem(IDC_SPUL)->ShowWindow((m_PruneLength > 1)?SW_SHOW:SW_HIDE);
	GetDlgItem(IDC_SPUR)->ShowWindow((m_PruneLength > 1)?SW_SHOW:SW_HIDE);
	ENABLE(IDC_FINDL,m_ListHasOCR);
	ENABLE(IDC_FINDR,m_ListHasOCR);
	ENABLE(IDC_SHOWSEARCHED,m_ListHasOCR);
	ENABLE(IDC_SAVEOCR,m_ListHasOCR);
	char Temp[64];
	sprintf(Temp,"%d SPUs",m_PruneLength);
	SetDlgItemText(IDC_NUMPRUNED,Temp);
	if(NextPos == -1 || m_PrunedList[NextPos]->m_SPUNumber != SPUIndex)
		ViewSPU(m_SPUSlider.GetPos());
}

void RemoveItalics(unsigned char* Bitmap,int Length, int Height, unsigned char BodyType, float& AverageSpace);

#define OVERSIZE 30
void CDVDSubEditDlg::DoRunOCR(CSubPic* TheSubpic)
{
	// Don't run OCR on button highlights. This is not a 
	if(!TheSubpic || TheSubpic->ContainsHighlights()) return;
	// Allocate bitmap and lines of characters for results if needed.
	if(m_OCRBitmap == 0) m_OCRBitmap = new uchar[(720+OVERSIZE)*576];
	if(m_OCRLines == 0) m_OCRLines = new char[32][128];
	long NewWidth, NewHeight;
	static int busy = 0;
	while(busy) Sleep(10);
//	TheSubpic->SavePicAsPPM("/temp/t2.ppm", 0, 1, 1, m_DefaultColors);

	int i,j;
	// Get the body pixel from the subpic. Use the previous one if it couldn't be detected...
	static int BodyPixel = -1;
	if(TheSubpic->m_BodyPix != -1)
		BodyPixel = TheSubpic->m_BodyPix;
	if(BodyPixel == -1)
		return;
#if 0
	if(m_IgnoreCLUT.GetCheck() || TheSubpic->m_CLUT == 0)
	{
		int Mean[4]={0,0,0,0};
		int MaxMean=0;
		ushort *Cols = TheSubpic->m_ColorPtr;
		for(i=0;i<4;i++)
		{
			if(i==TheSubpic->m_BackgroundPix) continue;
			Mean[i]=(m_DefaultColors[3*i]+m_DefaultColors[3*i+1]+m_DefaultColors[3*i+2])/3;
			if(Mean[i]>MaxMean) BodyPixel=i,MaxMean = Mean[i];
		}
	}
#endif

	busy = 1;

#if 0
	float AverageSpace;
	if(m_FixItalics)
		RemoveItalics(TheSubpic->m_Pixels,TheSubpic->m_SubPicWidth,TheSubpic->m_SubPicHeight,
			BodyPixel, AverageSpace);
	TRACE("AverageSpace = %.2f\n",AverageSpace);
#else
#endif

	// Colors for each pixel types. Black for everything but the character body pixels.
	uchar Colors[4] ={255,255,255,255};
	Colors[BodyPixel]=0;
	// Now copy the pixel data into our m_OCRBitmap array.
#if 0
	NewWidth = TheSubpic->m_SubPicWidth;
	NewHeight = TheSubpic->m_SubPicHeight;
	uchar *Ptr = m_OCRBitmap;
	for(i=0;i<NewHeight;i++)
	{
		uchar* src = TheSubpic->m_Pixels + i * NewWidth;
		for(j=0;j<NewWidth;j++)
			*Ptr++ = Colors[src[j]] ;
		for(;j<NewWidth+OVERSIZE;j++)
			*Ptr++ = 255 ;
	}
	NewWidth += OVERSIZE;
#else
	NewWidth = TheSubpic->m_SubPicWidth ;
	NewHeight = TheSubpic->m_BottomMostPixel - TheSubpic->m_TopMostPixel;
	uchar *Ptr = m_OCRBitmap;
	for(i=0;i<NewHeight;i++)
	{
		uchar* src = TheSubpic->m_Pixels + (TheSubpic->m_TopMostPixel+i) * NewWidth;
		for(j=0;j<NewWidth;j++)
			*Ptr++ = Colors[src[j]] ;
		for(;j<NewWidth+OVERSIZE;j++)
			*Ptr++ = 255 ;
	}
	NewWidth += OVERSIZE;
#endif
	// Detect italics and straighten them, see if this improves detection.
	float AverageSpace;
	if(m_FixItalics)
		RemoveItalics(m_OCRBitmap,NewWidth,NewHeight,0, AverageSpace);
	TRACE("AverageSpace = %.2f\n",AverageSpace);

//	FILE* fd = fopen("/temp/t2.ppm","wb");
//	fprintf (fd,"P6\n%d %d\n255\n", NewWidth, NewHeight);
//	fwrite(m_OCRBitmap,sizeof(char),3*NewWidth*NewHeight,fd);
//	fclose(fd);

	// A bit of handwaving here... The last parameter controls the space expected between
	// words. Since my italics code estimates that, I can adjust as a function of it.
	RunOCR(m_OCRBitmap,NewWidth,NewHeight,m_OCRLines,&m_NumOCRLines, 
		(int)(.5*AverageSpace+7.5));

	// This is silly. I'm fixing very comment error: '?' mistaken for 7. and "!" for "!."
	for(i=0;i<m_NumOCRLines;i++)
		strcpy((char*)TheSubpic->m_OCRLines[i],m_OCRLines[i]);
	TheSubpic->m_OCRLines[i][0]='\0';
	m_OCROut.SetWindowText(m_OCRLines[0]);

	busy = 0;
}


BOOL CDVDSubEditDlg::PreTranslateMessage(MSG* pMsg) 
{
	unsigned short Mask = 0x8000;

	// This will capture all messages arriving at the app before they're translated and dispatched.
	// Don't process keydown messages coming from the search or searchselect forms.
	if(pMsg->message == WM_KEYDOWN && pMsg->hwnd != GetDlgItem(IDC_SEARCH)->m_hWnd && 
		pMsg->hwnd != GetDlgItem(IDC_SEARCHSELECT)->m_hWnd)
//	if(pMsg->message == WM_KEYDOWN)
	{
		BOOL ShiftOn = (GetAsyncKeyState(VK_SHIFT) & Mask);
		BOOL CtrlOn = (GetAsyncKeyState(VK_CONTROL) & Mask);
		switch(pMsg->wParam)
		{
			// Must be all CAPS...
		case VK_F1:	OnHelpManual();return true;
		case VK_F5:	InvalidateRect(0,0);return true;
		case VK_F3:	FindNext(ShiftOn?-1:1);return true;
		case VK_RIGHT:	OnSpur(); return true;
		case VK_LEFT:	OnSpul(); return true;
		case VK_ESCAPE: m_AbortNow = true;  return true;
		}
		if(!CtrlOn)
			return CDialog::PreTranslateMessage(pMsg); 

		switch(pMsg->wParam)
		{
			// Must be all CAPS...
		case 'V': OnEditRedo(); break;
		case 'O': LoadLastFile(); break;
		case 'E': OnCropSubpic(); break;
		case 'U': OnOpenSup(); break;
		case 'I': m_FixItalics = !m_FixItalics; break;
		case 'A': OnEditApplyall(); break;
		case 'T': OCRThisOne(); break;
		case 'Z': OnRevertsaved(); break;
		case 'J': OnOpenIFO(); break;
		case 'S': if(ShiftOn) OnSavePPM() ; else OnSaveall(); break;
		case 'R': HideHearingImpaired(); break;
#if 1
		case '0': 
		case '1': 
		case '2': 
		case '3': 
		case '4': 
		case '5': 
		case '6': 
		case '7': 
		case '8': 
		case '9': 
			m_DCSQTSelect.SetCurSel(pMsg->wParam-'0');
			OnSelchangeDcsqtselect();
			break;
#endif
		default: return CDialog::PreTranslateMessage(pMsg);
		}
		return TRUE;
	}
	else return CDialog::PreTranslateMessage(pMsg);
}

void CDVDSubEditDlg::OnRevertsaved() 
{
	if(!m_CanReload)
		return;
	if(m_SelectedSubPic != 0)
		m_SelectedSubPic->ReloadSPU();
	SetMoveVSlider(MAX_V_MOVE);
	UpdateSPUList();
	ViewSPU(m_SPUSlider.GetPos());	
	m_ChangeRecord.Reset();
}


void CDVDSubEditDlg::OnPvcid() 
{
	if(m_VCIDSelect.GetCurSel() < 1) return;
	m_VCIDSelect.SetCurSel(m_VCIDSelect.GetCurSel()-1);
	OnSelchangeStreamselect();
}

void CDVDSubEditDlg::OnNvcid() 
{
	if(m_VCIDSelect.GetCurSel() >= m_VCIDLength) return;
	m_VCIDSelect.SetCurSel(m_VCIDSelect.GetCurSel()+1);
	OnSelchangeStreamselect();
}

void CDVDSubEditDlg::OnPstream() 
{
	if(m_StreamSelect.GetCurSel() < 1) return;
	m_StreamSelect.SetCurSel(m_StreamSelect.GetCurSel()-1);
	OnSelchangeStreamselect();
}

void CDVDSubEditDlg::OnNstream() 
{
	if(m_StreamSelect.GetCurSel() >= m_TopStreamID+1) return;
	m_StreamSelect.SetCurSel(m_StreamSelect.GetCurSel()+1);
	OnSelchangeStreamselect();
}



void CDVDSubEditDlg::OnShowvideo() 
{
	m_DisplayVideo = m_ShowVideo.GetCheck();
	// Hide the preview window!
	::ShowWindow(m_PreviewHandle,m_DisplayVideo?SW_SHOW:SW_HIDE);
	// Show the current SPU.
	ViewSPU(m_SPUSlider.GetPos());	
}

void CDVDSubEditDlg::OnClose() 
{
	if(WarnIfModifs() == false)
		return;
	// Saving preferences here because ScanDialog is a child and will be destroyed before OnDestroy.
	TheSettings.SaveSettings();
	TheSettings.SaveSettings("DefClutB",(long)RGB(m_DefaultColors[0],m_DefaultColors[1],m_DefaultColors[2]));
	TheSettings.SaveSettings("DefClutP",(long)RGB(m_DefaultColors[3],m_DefaultColors[4],m_DefaultColors[5]));
	TheSettings.SaveSettings("DefClutE1",(long)RGB(m_DefaultColors[6],m_DefaultColors[7],m_DefaultColors[8]));
	TheSettings.SaveSettings("DefClutE2",(long)RGB(m_DefaultColors[9],m_DefaultColors[10],m_DefaultColors[11]));

	CDialog::OnClose();
}

bool CDVDSubEditDlg::WarnIfModifs()
{
	if(!m_CanSave && !m_SupFileMode)
		return true;
	int i;
	for(i=0;i<m_TotNumSPU;i++)
	{
		if(m_MasterList[i]->m_Modified)
			break;
	}
	if(i != m_TotNumSPU && 
		AfxMessageBox("You have modified some of the subpics... \nDo you want to close without saving?",MB_YESNO) != IDYES)
		return false;
	return true;
}

void CDVDSubEditDlg::OnNoclut() 
{
	if(m_AutoCLUT.GetCheck())
		m_IgnoreCLUT.SetCheck(true);
	ViewSPU(m_SPUSlider.GetPos());	
}

void CDVDSubEditDlg::OnDefbutton() 
{
	CWnd* Ptr = SetFocus();
	if(Ptr == GetDlgItem(IDC_SEARCH))
		GetDlgItem(IDC_FINDR)->SetFocus();
}

void CDVDSubEditDlg::OnKillfocusMoveude() 
{
	int move = GetDlgItemInt(IDC_MOVEUDE);
	move = min(MAX_V_MOVE,move);
	move = max(-MAX_V_MOVE,move);
	SetMoveVSlider(MAX_V_MOVE+move);
	if(!m_SelectedSubPic)
		return ;
	m_SelectedSubPic->VMoveSubPic(m_MoveVSlider.GetPos()-MAX_V_MOVE); 
	m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeVMove, m_MoveVSlider.GetPos()-MAX_V_MOVE);
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnKillfocusMovelre() 
{
	int move = GetDlgItemInt(IDC_MOVELRE);
	move = min(MAX_H_MOVE,move);
	move = max(-MAX_H_MOVE,move);
	SetMoveHSlider(MAX_H_MOVE+move);
	if(!m_SelectedSubPic)
		return ;
	m_SelectedSubPic->HMoveSubPic(m_MoveHSlider.GetPos()-MAX_H_MOVE); 
	m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeHMove, m_MoveHSlider.GetPos()-MAX_H_MOVE);
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnKillfocusBordere() 
{
	int border = GetDlgItemInt(IDC_BORDERE);
	border = min(border,MAX_BORDER);
	border = max(border,0);
	SetBorderSlider(border);
	if(!m_SelectedSubPic)
		return ;
	m_SelectedSubPic->VMoveSubPicInFrame(border,m_FlushBorders.GetCheck()); 
	m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, border);
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnKillfocusDuratione() 
{
	char Temp[1024];
	GetDlgItemText(IDC_DURATIONE,Temp,1024);
	float Duration;
	sscanf(Temp,"%f",&Duration);
	if(!m_SelectedSubPic)
		return ;
	m_SelectedSubPic->ChangeDCSQTDelay((int)(Duration*1000));
	m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeDur, m_SelectedSubPic->GetDCSQTDelay());
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnSelchangePixselect() 
{
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnSelchangeStart() 
{
	if(!m_SelectedSubPic)
		return ;
	// Change the start command according to the selector...
	if(m_StartCommand.GetCurSel() == 0)
		m_SelectedSubPic->ChangeCommand(FSTA_DSP,STA_DSP);
	else
		m_SelectedSubPic->ChangeCommand(STA_DSP,FSTA_DSP);
	// Be smart, if se change the start command, we might want to apply to all.
	m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeStart, m_StartCommand.GetCurSel());
	ViewSPU(m_SPUSlider.GetPos(),false);	
}

void CDVDSubEditDlg::OnSelchangeBorderselect() 
{
	if(!m_SelectedSubPic)
		return ;
	switch(m_BorderSelect.GetCurSel())
	{
	case BORDER_CUSTOM:
		m_SelectedSubPic->VMoveSubPicInFrame(0,m_FlushBorders.GetCheck());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, 0);
		ViewSPU(m_SPUSlider.GetPos());
		break;
	case BORDER_SAFE: 
		// I'm using 8% for the safe area.
		m_SelectedSubPic->VMoveSubPicInFrame((FORMAT_NTSC ? 38 : 46),m_FlushBorders.GetCheck());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, (FORMAT_NTSC ? 38 : 46));
		ViewSPU(m_SPUSlider.GetPos());
		break;
	case BORDER_4_3_L: 
		// 16:9 Letterboxed.
		m_SelectedSubPic->VMoveSubPicInFrame((FORMAT_NTSC ? 60 : 72),m_FlushBorders.GetCheck());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, (FORMAT_NTSC ? 60 : 72));
		ViewSPU(m_SPUSlider.GetPos());
		break;
	case BORDER_4_3_LS: 
		// 16:9 letterboxed safe.
		m_SelectedSubPic->VMoveSubPicInFrame((FORMAT_NTSC ? 65 : 78),m_FlushBorders.GetCheck());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, (FORMAT_NTSC ? 65 : 78));
		ViewSPU(m_SPUSlider.GetPos());
		break;
	}
}

void CDVDSubEditDlg::OnSelchangeDcsqtselect() 
{
	if(!m_SelectedSubPic)
		return ;
	if(m_SelectedSubPic->m_SelectedDCSQT != m_DCSQTSelect.GetCurSel())
	{
		m_SelectedSubPic->SelectDCSQT(m_DCSQTSelect.GetCurSel());
		ViewSPU(m_SPUSlider.GetPos(),true,false);
	}
	else
		m_SelectedSubPic->SelectDCSQT(m_DCSQTSelect.GetCurSel());
}

void CDVDSubEditDlg::OnDeltaposMoveb(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	SetMoveVSlider(m_MoveVSlider.GetPos()-2*pNMUpDown->iDelta);
	if(m_SelectedSubPic)
	{
		m_SelectedSubPic->VMoveSubPic(m_MoveVSlider.GetPos()-MAX_V_MOVE);
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeVMove, m_MoveVSlider.GetPos()-MAX_V_MOVE);
		ViewSPU(m_SPUSlider.GetPos());
	}
	*pResult = 0;
}

void CDVDSubEditDlg::OnDeltaposMovelrb(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	SetMoveHSlider(m_MoveHSlider.GetPos()-pNMUpDown->iDelta);
	if(m_SelectedSubPic)
	{
		m_SelectedSubPic->HMoveSubPic(m_MoveHSlider.GetPos()-MAX_H_MOVE);
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeHMove, m_MoveHSlider.GetPos()-MAX_H_MOVE);
		ViewSPU(m_SPUSlider.GetPos());
	}
	*pResult = 0;
}


void CDVDSubEditDlg::OnDeltaposBorderb(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	SetBorderSlider(m_BorderSlider.GetPos()-pNMUpDown->iDelta);
	if(m_SelectedSubPic)
	{
		m_SelectedSubPic->VMoveSubPicInFrame(m_BorderSlider.GetPos(),m_FlushBorders.GetCheck());
		ViewSPU(m_SPUSlider.GetPos());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, m_BorderSlider.GetPos());
	}
	*pResult = 0;
}


void CDVDSubEditDlg::OnFlushBorder() 
{
	if(m_SelectedSubPic && m_BorderSlider.GetPos() != 0 && m_FlushBorders.GetCheck())
	{
		m_SelectedSubPic->VMoveSubPicInFrame(m_BorderSlider.GetPos(),m_FlushBorders.GetCheck());
		ViewSPU(m_SPUSlider.GetPos());
		m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeBorder, m_BorderSlider.GetPos());
	}
}


void CDVDSubEditDlg::OnDeltaposDscqtb(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
	// TODO: Add your control notification handler code here
	m_DCSQTSelect.SetCurSel(max(0,m_DCSQTSelect.GetCurSel()+pNMUpDown->iDelta));
	OnSelchangeDcsqtselect();
	*pResult = 0;
}

void CDVDSubEditDlg::OnHidesubpic() 
{
	if(!m_SelectedSubPic)
		return;
	m_SelectedSubPic->ChangeTransparency(0,5);
	ViewSPU(m_SPUSlider.GetPos());
	m_ChangeRecord.RecordChange(m_SelectedSubPic,ChangeTransp,0,5);
}













// __________________________ MENU STUFF _______________________


void CDVDSubEditDlg::OnHelpAbout() 
{
	CAboutDlg dlgAbout;
	dlgAbout.m_Edit = this;
	dlgAbout.DoModal();
}

void CDVDSubEditDlg::OnHelpManual() 
{
	OpenRSCFile(IDR_HELPFILE, ".pdf");
}

void CDVDSubEditDlg::OnHelpOntheweb() 
{
	ShellExecute(NULL, "open", "http://www.videohelp.com/~DVDSubEdit", NULL, NULL, SW_SHOWNORMAL);
}

void CDVDSubEditDlg::OnEditRedo() 
{
	if(m_SelectedSubPic == 0)
		return;
	if(m_ChangeRecord.m_VMove != -1)
		m_SelectedSubPic->VMoveSubPic(m_ChangeRecord.m_VMove);
	if(m_ChangeRecord.m_HMove != -1)
		m_SelectedSubPic->HMoveSubPic(m_ChangeRecord.m_HMove);
	if(m_ChangeRecord.m_Border != 0)
		m_SelectedSubPic->VMoveSubPicInFrame(m_ChangeRecord.m_Border,m_FlushBorders.GetCheck());
	if(m_ChangeRecord.m_Transparency != -1)
		m_SelectedSubPic->ChangeTransparency(m_ChangeRecord.m_Transparency,m_ChangeRecord.m_TranspPixel);
	if(m_ChangeRecord.m_Color != -1)
		ChangeColors(m_ChangeRecord.m_Color,m_ChangeRecord.m_ColorPixel,m_SelectedSubPic);
	if(m_ChangeRecord.m_Crop != -1)
		m_SelectedSubPic->CropBackground();
	if(m_ChangeRecord.m_StartCommand != -1)
	{
		if(m_ChangeRecord.m_StartCommand == 0)
			m_SelectedSubPic->ChangeCommand(FSTA_DSP,STA_DSP);
		else
			m_SelectedSubPic->ChangeCommand(STA_DSP,FSTA_DSP);
	}
	if(m_ChangeRecord.m_Duration != -1)
		m_SelectedSubPic->ChangeDCSQTDelay(m_ChangeRecord.m_Duration);
	ViewSPU(m_SPUSlider.GetPos());
}

void CDVDSubEditDlg::OnExit() 
{
    AfxGetMainWnd()->SendMessage(WM_CLOSE);
}

void CDVDSubEditDlg::OnHelpViewlog() 
{
	ShellExecute(NULL, "open", m_LogPath, NULL, NULL, SW_SHOWNORMAL);
}


void CDVDSubEditDlg::OnEditApplyall() 
{
	if(m_SelectedSubPic == 0)
		return;
	m_ApplyAllDialog->m_ApplyVMove = m_ChangeRecord.m_VMove != -1;
	m_ApplyAllDialog->m_ApplyHMove = m_ChangeRecord.m_HMove != -1;
	m_ApplyAllDialog->m_ApplyBorders = m_ChangeRecord.m_Border != 0;
	m_ApplyAllDialog->m_ApplyStart = m_ChangeRecord.m_StartCommand != -1;
	m_ApplyAllDialog->m_ApplyTransp = m_ChangeRecord.m_Transparency != -1;
	m_ApplyAllDialog->m_ApplyColor = m_ChangeRecord.m_ColorPixel != 0;
	m_ApplyAllDialog->m_ApplyDuration = m_ChangeRecord.m_Duration != -1;
	m_ApplyAllDialog->m_ApplyCrop = m_ChangeRecord.m_Crop != -1;
	CString Temp;
	Temp.Format("%.3f s",(float)max(0,m_ChangeRecord.m_Duration)/1000);
	m_ApplyAllDialog->m_Duration = Temp;
	OnApplytoall();
}


void CDVDSubEditDlg::OnOpenVob() 
{
	if(WarnIfModifs() == false)
		return;
	static CFileDialog FileDialog(TRUE, CString("*.vob"), "", OFN_FILEMUSTEXIST, CString("VOB Files (*.vob)|*.vob||"));
	FileDialog.m_ofn.lpstrTitle = "Open VOB file";
	if (FileDialog.DoModal() == IDOK)
	{
		Reset();
		AddFile((char*)(LPCTSTR)FileDialog.GetPathName());
		EnableUI(true);
		// Now scan input file...
		m_ScanDialog->DisplayInfo(0,"","");
		m_ScanDialog->ShowWindow(SW_SHOW);
		StartScan();
		m_Preview->LoadFile(m_FileInfos[0].m_FileName);
	}
}

void CDVDSubEditDlg::OnOpenDomain() 
{
	if(WarnIfModifs() == false)
		return;
	static CFileDialog FileDialog(TRUE, CString("*.vob"), "", OFN_FILEMUSTEXIST, CString("VOB Files (*.vob)|*.vob||"));
	FileDialog.m_ofn.lpstrTitle = "Open Full Domain";
	if (FileDialog.DoModal() == IDOK)
	{
		Reset();
		int Index = -1;
		int VTS = -1;
		char FileName[1024];
		strcpy(FileName,(char*)(LPCTSTR)FileDialog.GetPathName());
		// Get VTS number and index.
		if(sscanf(GetFname(FileName),"%*[^_]_%d_%d",&VTS,&Index) == 2 && Index > 0)
		{
			// This is a vts_nn_x.vob
			for(int i=1;i<10;i++)
			{
				FileName[strlen(FileName)-5]='0'+i;
				if(PathExists(FileName))
					AddFile(FileName);
			}
		}
		else
			AddFile(FileName);
		EnableUI(true);
		// Now scan input file...
		m_ScanDialog->DisplayInfo(0,"","");
		m_ScanDialog->ShowWindow(SW_SHOW);
		m_CanMove = true;
		StartScan();
		m_Preview->LoadFile(m_FileInfos[0].m_FileName);	
	}
}

void CDVDSubEditDlg::OnSavePPM() 
{
	if(m_SelectedSubPic == 0)
		return;
	static CFileDialog FileDialog(FALSE, CString("*.ppm"), "", OFN_OVERWRITEPROMPT, 
		CString("ppm Files (*.ppm)|*.ppm||"));
	FileDialog.m_ofn.lpstrTitle = "Save subpic as PPM file";
	uchar *CLUT = m_IgnoreCLUT.GetCheck() ? 0 : m_SelectedSubPic->m_CLUT;
	if (FileDialog.DoModal() == IDOK)
		m_SelectedSubPic->SavePicAsPPM((LPCTSTR)FileDialog.GetPathName(),CLUT,
			1,1,m_DefaultColors);
}


void CDVDSubEditDlg::OnSaveSup() 
{
	if(m_StreamSelect.GetCurSel() == 0)
	{
		ERROR1("You did not select a specific subpicture stream.\nSelect one first before saving as sup file.");
		return;
	}
	static CFileDialog FileDialog(FALSE, CString("*.sup"), "", OFN_OVERWRITEPROMPT, 
		CString("sup Files (*.sup)|*.sup||"));
	FileDialog.m_ofn.lpstrTitle = "Save subpics as .sup file";
	if (FileDialog.DoModal() == IDOK)
		SaveSelectedAsSupFile((LPCTSTR)FileDialog.GetPathName());
}

void CDVDSubEditDlg::OnOpenSup() 
{
	if(WarnIfModifs() == false)
		return;
	static CFileDialog FileDialog(TRUE, CString("*.sup"), "", OFN_FILEMUSTEXIST, 
		CString("sup Files (*.sup)|*.sup||"));
	FileDialog.m_ofn.lpstrTitle = "Load .sup file";
	if (FileDialog.DoModal() == IDOK)
	{
		if(LoadSUPFile((LPCTSTR)FileDialog.GetPathName()) != 0)
			return;
		m_ShowModified.SetCheck(false);
		OnSelchangeStreamselect();
		EnableUI(true);
		SetWindowText((LPCTSTR)FileDialog.GetPathName());
		m_Preview->SetDisplayMode(-1);
		m_AutoCLUT.SetCheck(true);
		OnOpenIFO();
	}
}

char* PrintPTS(int PTS, int durMs)
{
	static char Temp[128];
	float Seconds1 = (float)PTS/90000;
	float Seconds2 = (float)Seconds1+(float)durMs/1000;
	sprintf(Temp,"%02d:%02d:%02d,%d --> %02d:%02d:%02d,%d",
		(int) Seconds1 / 3600,
		((int) Seconds1 % 3600)/60,
		((int) Seconds1 % 3600)%60,
		(int)((Seconds1-(int)Seconds1)*1000),
		(int) Seconds2 / 3600,
		((int) Seconds2 % 3600)/60,
		((int) Seconds2 % 3600)%60,
		(int)((Seconds2-(int)Seconds2)*1000));
	return Temp;
}

void CDVDSubEditDlg::OnSaveOCR() 
{
	static CFileDialog FileDialog(FALSE, CString("*.txt"), "", OFN_OVERWRITEPROMPT, 
		CString("(*.srt)|*.srt||"));
	FileDialog.m_ofn.lpstrTitle = "Save OCR output as .srt file";
	if (FileDialog.DoModal() == IDOK)
	{
		FILE* fd = fopen((LPCTSTR)FileDialog.GetPathName(),"w");
		if(fd == 0)
		{
			ERROR1("Couldn't open file for writing!");
			return;
		}
		int i,j;
		for(i=0;i<m_PruneLength;i++)
		{
			CSubPic* Temp = m_PrunedList[i];
			if(Temp->m_OCRLines[0][0] == '\0' || Temp->GetMeanTransparency() == 0)
				continue;
			fprintf(fd,"%d\n%s\n",i,PrintPTS(Temp->m_PTS,Temp->GetDCSQTDelay()));
			for(j=0;j<32;j++)
			{
				if(Temp->m_OCRLines[j][0] == '\0')
					break;
				fprintf(fd,"%s\n",Temp->m_OCRLines[j]);
			}
			fprintf(fd,"\n");
		}
		fclose(fd);
	}
}

void CDVDSubEditDlg::OnTimingSingle() 
{
	static bool First = true;
	m_MoveDialog->m_RetimeAll = false;
	if(!m_CanMove)
	{
		AfxMessageBox("Retiming subtitles is only possible in the title domain\nand if the entire domain is loaded.\
\nNot available for .sup files");
		return;
	}
	if(m_SelectedSubPic == 0 || m_MoveDialog->DoModal() != IDOK)
		return;
	if(First && AfxMessageBox("There is NO undo for this\nThe VOB file will be modified. Proceed?",MB_OKCANCEL) != IDOK)
		return;
	First = false;
	if(m_MoveDialog->m_CurSel == 0)
		MoveThisOne(m_SelectedSubPic->m_SPUNumber, m_MoveDialog->m_ChangeTimingBy);
	else if(m_MoveDialog->m_CurSel == 1)
	{
		float DeltaSeconds = m_MoveDialog->m_ChangeTimingBy - PTSToSecs(m_SelectedSubPic->m_PTS);
		MoveThisOne(m_SelectedSubPic->m_SPUNumber, DeltaSeconds);
	}
	UpdateSPUList();
	ViewSPU(m_SPUSlider.GetPos());
}


void CDVDSubEditDlg::OnInitMenu( CMenu* pMenu )
{
	HMENU HMenu = pMenu->GetSafeHmenu( );

	
	EnableMenuItem(HMenu,ID_FILE_OPENIFO,(m_MenuEnabled && m_SupFileMode)?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_OPENVOB,m_MenuEnabled?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_EDIT_REVERTTHIS,(m_MenuEnabled && m_CanReload)?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_EDIT_REVERTALL,(m_MenuEnabled && m_CanReload)?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_EDIT_APPLYALL, m_MenuEnabled?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_OPENVOB,m_MenuEnabled?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_OPENDOMAIN,m_MenuEnabled?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_OPENSUP,m_MenuEnabled?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_SAVE,(m_MenuEnabled && m_CanSave) ?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_OPENLAST,(m_MenuEnabled && m_CanSave) ?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_FILE_SAVESUP,(m_MenuEnabled) ?MF_ENABLED:MF_GRAYED);
	EnableMenuItem(HMenu,ID_EDIT_CLUT,m_MenuEnabled?MF_ENABLED:MF_GRAYED);
}

void CDVDSubEditDlg::OnFindLeft() 
{
	FindNext(-1);
}

void CDVDSubEditDlg::OnFindRight() 
{
	FindNext(1);
}

void CDVDSubEditDlg::FindNext(int Inc)
{
	char Find[1024];
	GetDlgItemText(IDC_SEARCH,Find,1024);
	int i = m_SPUSlider.GetPos()+Inc;
	int k=0;
start:
	for(;i<m_PruneLength && i >= 0;i+=Inc)
	{
		// Ask SPU if it has that string.
		if(m_PrunedList[i]->FindString(Find))
		{
			// Yes!
			m_SPUSlider.SetPos(i);
			ViewSPU(i);
			return ;
		}
	}
	if(k++ == 0)
	{
		// We didn't find the search string, and this is the first time. Let's go back to start
		// or end, and search again.
		i=(Inc==1)?0:m_PruneLength-1;
		goto start;
	}
	MessageBeep(MB_ICONEXCLAMATION);
    return ;
}

// A 3-line class so I can set the title of the dialog the way I want it! GEEE!
class MyCColorDialog : public CColorDialog
{
public:
	char* m_Title;
	MyCColorDialog(COLORREF clrInit) : CColorDialog(clrInit) {};
	BOOL OnInitDialog() { CColorDialog::OnInitDialog(); SetWindowText(m_Title); return true;}
};

void CDVDSubEditDlg::OnClutClick(UINT nID) 
{
	// No color adjustment possible in auto clut.
	if(m_AutoCLUT.GetCheck())
		return;
	int i = nID-IDC_COLB;
	static CPickColorDlg PickCol;
	char Temp[512];
	char Foo[4][3]={"b","p","e1","e2"};
	if(m_IgnoreCLUT.GetCheck() || (m_SelectedSubPic != 0 && m_SelectedSubPic->m_CLUT == 0))
	{
		// Pop a color chosing dialog to pick the colors for the default CLUT
		sprintf(Temp,"Pick default color for pixel %s",Foo[i]);
		MyCColorDialog TheColDialog(RGB(m_DefaultColors[3*i],m_DefaultColors[3*i+1],m_DefaultColors[3*i+2])); 
		TheColDialog.m_Title = Temp;
		if(TheColDialog.DoModal() == IDOK)
		{
			COLORREF Chosen = TheColDialog.GetColor();
			m_DefaultColors[3*i]=GetRValue(Chosen);
			m_DefaultColors[3*i+1]=GetGValue(Chosen);
			m_DefaultColors[3*i+2]=GetBValue(Chosen);
			ViewSPU(m_SPUSlider.GetPos());
			return;
		}
	}
	else if(m_SelectedSubPic != 0)
	{
		// Here we're picking which tile in the CLUT to use for pixel i.
		sprintf(Temp,"Pick color for pix. %s",Foo[i]);
		PickCol.m_WindowTitle = Temp;
		PickCol.m_Engine = this;
		PickCol.m_WhichPixel = i;
		// Set the currently selected tile for this pixel.
		PickCol.m_SelectedCol = m_SelectedSubPic->m_ColorPtr[i];
		if(PickCol.DoModal() == IDOK)
		{
			// Change the color for this pixel
			m_SelectedSubPic->ChangeColor(PickCol.m_SelectedCol,i);
			ViewSPU(m_SPUSlider.GetPos());
			// And record the change. Set the right bit in Pixel and set the right Byte in Color.
			int Pixel = ((int)1)<<i;
			int Color = 0;
			uchar *Ptr = (uchar*) &Color;
			Ptr[i]=PickCol.m_SelectedCol; // The ith byte will hold the color for the ith pixel.
			m_ChangeRecord.RecordChange(m_SelectedSubPic, ChangeColor, Color, Pixel);
		}
	}
}

void CDVDSubEditDlg::ChangeColors(int Colors, int Pixels, CSubPic* TheSubpic)
{
	int i;
	uchar *Ptr = (uchar*) &Colors;
	for(i=0;i<4;i++)
	{
		// Iterate over all pixels and if the corresponding bit is set in Pixels, change the color.
		if(Pixels & ((int)1<<i))
			TheSubpic->ChangeColor((int)Ptr[i],i);
	}
}

#define MOVEWIN(A,DeltaX,DeltaY) 	{ CRect Tmp; CWnd* Wn = GetDlgItem(A); if(Wn != 0) {Wn->GetWindowRect(&Tmp); ScreenToClient(&Tmp); Wn->SetWindowPos(0,Tmp.left+DeltaX,Tmp.top+DeltaY,0,0,SWP_NOSIZE|SWP_NOZORDER|SWP_NOREDRAW);}}
#define MOVEWINS(A,DeltaY,DeltaW) 	{ CRect Tmp; CWnd* Wn = GetDlgItem(A); if(Wn != 0) {Wn->GetWindowRect(&Tmp); ScreenToClient(&Tmp); Wn->SetWindowPos(0,Tmp.left,Tmp.top+DeltaY,Tmp.Width()+DeltaW,Tmp.Height(),SWP_NOZORDER|SWP_NOREDRAW);}}
#define SIZEWIN(A,DeltaW,DeltaH) 	{ CRect Tmp; CWnd* Wn = GetDlgItem(A); if(Wn != 0) {Wn->GetWindowRect(&Tmp); ScreenToClient(&Tmp); Wn->SetWindowPos(0,Tmp.left,Tmp.top+DeltaY,Tmp.Width()+DeltaW,Tmp.Height()+DeltaH,SWP_NOMOVE|SWP_NOZORDER|SWP_NOREDRAW);}}

void CDVDSubEditDlg::OnSize(UINT nType, int cx, int cy) 
{
	// Avoid doing anything when minimizing!!!
	if(m_InitWindowWidth == -1 || nType == SIZE_MINIMIZED)
		return ;
	CRect NewRect;
	GetWindowRect(&NewRect);
	// Move whatever needs to be moved.
	static int LastWidth = m_InitWindowWidth;
	static int LastHeight = m_InitWindowHeight;
	int DeltaX = max(m_InitWindowWidth,NewRect.Width()) - LastWidth;
	int DeltaY = max(NewRect.Height(),m_InitWindowHeight) - LastHeight;
	LastWidth = max(m_InitWindowWidth,NewRect.Width());
	LastHeight = max(NewRect.Height(),m_InitWindowHeight);

	int i;
	for(i=IDC_UNREF;i<=IDC_AUTOCLUT;i++)
	{
		if(i == IDC_COMMANDS || i == IDC_FRAME7 || i == IDC_SPUSLIDER)
		{
			MOVEWINS(i,DeltaY,DeltaX);
		}
		else if (i != IDC_VIDEOBACK)
		{
			MOVEWIN(i,DeltaX,DeltaY);
		}
	}
	// We handle IDC_VIDEOBACK (the video fake window) differently.
	GetDlgItem(IDC_VIDEOBACK)->GetWindowRect(&m_VideoRect);
	ScreenToClient(&m_VideoRect);
	CRect BotRight;
	// We don't want IDC_VIDEOBACK to extend past IDC_DEFBUTTON
	GetDlgItem(IDC_DEFBUTTON)->GetWindowRect(&BotRight);
	int NewVidWidth = min(m_VideoRect.Width()+DeltaX,BotRight.left - m_VideoRect.left);
	int NewVidHeight = min(m_VideoRect.Height()+DeltaY,BotRight.top - m_VideoRect.top);
	// Resize that guy.
	GetDlgItem(IDC_VIDEOBACK)->SetWindowPos(0,0,0,NewVidWidth,NewVidHeight,
		SWP_NOMOVE|SWP_NOZORDER|SWP_NOREDRAW);
	GetDlgItem(IDC_VIDEOBACK)->GetWindowRect(&m_VideoRect);
	ScreenToClient(&m_VideoRect);
	m_DisplayHeight = m_VideoRect.Height();
	JustFoundVideoFormat();
	CDialog::OnSize(nType, cx, cy);
}

void CDVDSubEditDlg::OnGetMinMaxInfo( MINMAXINFO FAR* lpMMI )
{
	// This is called when 
	POINT Pt={m_InitWindowWidth,m_InitWindowHeight};
	lpMMI->ptMinTrackSize = Pt;
}
