// DVDSubEditDlg.h : header file
//

#if !defined(AFX_DVDSubEditDLG_H__8849CA41_8885_4178_8047_BB28B35DE116__INCLUDED_)
#define AFX_DVDSubEditDLG_H__8849CA41_8885_4178_8047_BB28B35DE116__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Utils.h"
#include "SubPicEngine.h"
#include "DVDPreview.h"

class CScanDlg;
class CApplyToAllDlg;
class CMoveDlg;

typedef enum
{
	ChangeVMove = 0,
	ChangeHMove,
	ChangeBorder,
	ChangeTransp,
	ChangeColor,
	ChangeStart,
	ChangeDur,
	ChangeCrop,
} Change;

class Modif
{
public:
	// Reset all params.
	void	Reset();
	// Record a change made to TheSubpic. Which indicates what was changed, and Value1 and Value2
	// are parameters describing the change. If TheSubpic->m_SPUNumber is not m_WhichSPU, all values
	// are reset before the change is recorded.
	void	RecordChange(CSubPic* TheSubpic, Change Which, int Value1, int Value2=0);
	
	int		m_VMove;
	int		m_HMove;
	int		m_Border;
	int		m_Transparency;
	int		m_TranspPixel;
	int		m_StartCommand;
	int		m_Duration;
	int		m_Color;
	int		m_ColorPixel;
	int		m_Crop;
	uchar*	m_CLUT;
	int		m_WhichSPU;			// Which SPU were these modifications made on.
};

/////////////////////////////////////////////////////////////////////////////
// CDVDSubEditDlg dialog

class CDVDSubEditDlg : public CDialog, public CSubPicEngine
{
// Construction
public:
	CDVDSubEditDlg(CWnd* pParent = NULL);	// standard constructor

// Dialog Data
	//{{AFX_DATA(CDVDSubEditDlg)
	enum { IDD = IDD_DVDSubEdit_DIALOG };
	CButton	m_AutoCLUT;
	CButton	m_FlushBorders;
	CSliderCtrl	m_MoveHSlider;
	CButton	m_ShowSearched;
	CRichEditCtrl	m_OCROut;
	CComboBox	m_DCSQTSelect;
	CSliderCtrl	m_DurationSlider;
	CButton	m_ShowModified;
	CComboBox	m_BorderSelect;
	CSliderCtrl	m_BorderSlider;
	CComboBox	m_StartCommand;
	CComboBox	m_PixelSelect;
	CSliderCtrl	m_TransSlider;
	CButton	m_ShowVideo;
	CButton	m_SaveAll;
	CButton	m_IgnoreCLUT;
	CButton	m_Revert;
	CSliderCtrl	m_MoveVSlider;
	CComboBox	m_VCIDSelect;
	CSliderCtrl	m_SPUSlider;
	CComboBox	m_StreamSelect;
	CRichEditCtrl	m_CommandDisplay;
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CDVDSubEditDlg)
	public:
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);	// DDX/DDV support
	//}}AFX_VIRTUAL
	BOOL OnToolTipNotify( UINT id, NMHDR * pNMHDR, LRESULT * pResult );

public:
	CSettings			TheSettings;
	CScanDlg			*m_ScanDialog;
	CApplyToAllDlg		*m_ApplyAllDialog;
	CMoveDlg			*m_MoveDialog;
	CFindReplaceDialog	*m_FindDlg;
	CDVDPreview			*m_Preview;
	HWND				m_PreviewHandle;

	char				m_LogPath[1024];

	CMenu				m_Menu;
	bool				m_MenuEnabled;
	
	Modif				m_ChangeRecord;

	int					m_DisplayVideo;

	void				OnDropFiles(HDROP hDropInfo) ;
	void				StartScan();
	virtual void		JustAddedSPU(int StreamIndex, int SPUIndex);
	virtual void		JustFoundStream(int StreamIndex);
	virtual void		JustFoundVCID(uint VCID);
	virtual void		Reset();
	virtual void		JustFoundVideoFormat();
//	virtual int			PruneSPUList(int CurrentSelected, int StreamID = -1, int StartLBA = -1, int EndLBA = -1,
//		ushort VOBID = 0, uchar CELLID = 0);
						
	void				EnableUI(bool Enable, bool Full = false);
	int					m_SelectedStream;
	void				ShowCol(HDC TheDC, uint ID, uchar* Colormap);
	void				ViewSPU(int SPUIndex, bool RefreshVideo = true, bool RefreshCommands = true);

	int					m_Running;
	bool				m_QuitNow;
	int					m_Timer;
	HANDLE				m_ThreadHandle;
	int					m_DisplayHeight;
	bool				m_FixItalics;

	int					m_InitWindowWidth;		// Initial size of entire window.
	int					m_InitWindowHeight;		// Initial size of entire window.

	HDC					m_HiddenDC;
	BITMAPINFOHEADER	m_BitMapHeader;
	uchar*				m_SubpicBitmap ;
	uchar*				m_OCRBitmap;
	float				m_ScalingX;
	float				m_ScalingY;
	CRect				m_VideoRect;
	char				m_LastScanFile[1024];

	uchar				m_DefaultColors[16];
	char				(*m_OCRLines)[128];		// Results of OCR;
	int					m_NumOCRLines;			// How many lines were found.
	char				m_SearchStr[64];

	// This is used to avoid stepping on my own feet.
	CRITICAL_SECTION	m_csLock;

// Implementation
	HICON m_hIcon;

	void				DispInfo(char* Info, int Color);
	void				UpdateSPUList();
	void				ShowBorders();
	void				DisplayBitmap(unsigned char* Bitmap, long Width, long Height, bool Black = false);
	void				SetMoveVSlider(int Pos);
	void				SetMoveHSlider(int Pos);
	void				SetTransSlider(int Pos);
	void				SetBorderSlider(int Pos);
	void				SetBorderSelect(int Selection);
	void				SetDurationSlider(int Pos);
	bool				WarnIfModifs();
	void				DoRunOCR(CSubPic* TheSubpic);
	void				OCRThisOne();
	void				RunOCROnSelected();
	void				FindNext(int Inc);
	void				ChangeColors(int Colors, int Pixels, CSubPic* TheSubpic);
	void				LoadLastFile();
	void				OnFileLoaded(const char* FileName);
	void				OnCropSubpic();
	void				OnOpenIFO();
	void				HideHearingImpaired();

	// Generated message map functions
	//{{AFX_MSG(CDVDSubEditDlg)
	virtual BOOL OnInitDialog();
	afx_msg void OnSysCommand(UINT nID, LPARAM lParam);
	afx_msg void OnPaint();
	afx_msg HCURSOR OnQueryDragIcon();
	afx_msg void OnDestroy();
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg HBRUSH OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor);
	afx_msg void OnSpur();
	afx_msg void OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar);
	afx_msg void OnSelchangeStreamselect();
	afx_msg void OnSpul();
	afx_msg void OnRevertsaved();
	afx_msg void OnPvcid();
	afx_msg void OnNvcid();
	afx_msg void OnPstream();
	afx_msg void OnNstream();
	afx_msg void OnApplytoall();
	afx_msg void OnRevertall();
	afx_msg void OnClose();
	afx_msg void OnNoclut();
	afx_msg void OnSaveall();
	afx_msg void OnDefbutton();
	afx_msg void OnKillfocusMoveude();
	afx_msg void OnKillfocusMovelre();
	afx_msg void OnShowvideo();
	afx_msg void OnSelchangePixselect();
	afx_msg void OnSelchangeStart();
	afx_msg void OnSelchangeBorderselect();
	afx_msg void OnShowmod();
	afx_msg void OnKillfocusBordere();
	afx_msg void OnHelpAbout();
	afx_msg void OnHelpManual();
	afx_msg void OnHelpOntheweb();
	afx_msg void OnEditRedo();
	afx_msg void OnExit();
	afx_msg void OnEditApplyall();
	afx_msg void OnDeltaposMoveb(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDeltaposMovelrb(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnDeltaposBorderb(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnOpenVob();
	afx_msg void OnOpenDomain();
	afx_msg void OnSelchangeDcsqtselect();
	afx_msg void OnDeltaposDscqtb(NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg void OnHidesubpic();
	afx_msg void OnHelpViewlog();
	afx_msg void OnFindLeft();
	afx_msg void OnFindRight();
	afx_msg void OnSavePPM();
	afx_msg void OnShowsearched();
	afx_msg void OnKillfocusSearchselect();
	afx_msg void OnKillfocusSearch();
	afx_msg void OnKillfocusDuratione();
	afx_msg void OnSaveSup();
	afx_msg void OnFlushBorder();
	afx_msg void OnSize(UINT nType, int cx, int cy);
	//}}AFX_MSG
	void OnClutClick(UINT nID);
	afx_msg void OnSaveOCR();
	afx_msg void OnTimingSingle();
	afx_msg void OnTimingAll();
	afx_msg void OnOpenSup();
	afx_msg void OnGetMinMaxInfo( MINMAXINFO FAR* lpMMI );
	afx_msg void OnInitMenu( CMenu* pMenu );

	DECLARE_MESSAGE_MAP()
protected:
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_DVDSubEditDLG_H__8849CA41_8885_4178_8047_BB28B35DE116__INCLUDED_)
