#ifndef CPACK_H
#define CPACK_H
#include "includes.h"

// Here's a class for any kind of pack in a VOB file.
class CPack
{
public:
	CPack();
	~CPack();

	// There would be ways to minimize the size of this structure, as there might be cases where we have
	// to allocate quite a few.
	uint		m_InputLBA;		// LBA (sector) in input file.
	PackType	m_PackType;		// Nav, video, audio or subpic pack.
	uchar		m_StreamID;		// Stream ID for audio or subpics.
	FrameType	m_FrameType;	// Video frame type. I, P or B.
	IFrameType	m_IFrameType;	// Only for I-frames: Start of frame, middle, or last LBA of frame.
	char		m_IFrameTooFull;	// Only for I-frames: indicates we can't add anything to the I-Frame.
	short		m_IFrameSize;	// Size of I-Frame in sectors.
	uint		m_SRCL;			// System Reference Clock 4 lowest bytes.
	ushort		m_SRCH;			// System Reference Clock 2 highest bytes.
	uchar		m_NumButtons;	// Number of buttons for a navpack. Also used for subpic packs.


	// Given a buffer of data from a video pack, this will detect which type of video frame we have.
	void		FindPackType(uchar* Buffer, ushort &VOBID, uchar &CELLID);
	// Given a buffer of data, this will detect which kind of pack this is. If AspectRatio
	// is non zero, the aspect ratio code is returned. 1 for 1:1, 2 for 4:3, 3 for 16:9
	// and 4 for 2.21:1
	void		FindFrameType(uchar* Buffer, int* Offset= 0, uchar* AspectRatio = 0);

	// Finds the offset to the true end of an I-frame, in the mpeg data.
	int			FindEndFrameOffset(uchar* Buffer);

	int			FindSubpicLengthAndOffset(uchar* Buffer, ushort &Length, ushort &Offset, ushort &TrueOffset);

	// This trims the end of an I-Frame, putting an end_of_sequence header and adding a padding stream.
	void		TrimIFrame(uchar* Buf);

	// This adjusts the data in the GOP header of the I-frame. It resets the Closed and Broken GOP flags,
	// and it sets the temporal sequence number to 0.
	void		CloseGOP(uchar* Buffer);

	// Experimental: Removes a timed stop command in the last pack of a subpic stream.
	int			RemoveSubpicStop(uchar* Buffer, uchar* Buffer2, ushort TrueOffset, ushort Offset = 0);

	void		FindMaxButtonSize(uchar* Buffer, ushort &MaxSizeX, ushort &MaxSizeY, int &AllNops,
		int &ForcedAction, int &AnyNext);
};

// A small class to help with manipulating time in the DVD time format (eek!).
class CDVDTime
{
public:
	CDVDTime();
	~CDVDTime();

	CDVDTime(int PTSDuration);	// This initializes it with the duration in PTS ticks (90kHz).

	int		m_hours;
	int		m_minutes;
	int		m_seconds;
	int		m_frames;
	VFormat	m_format;

	// Read 4 bytes from Buffer and converts to a time. Uses the GetInt() function for
	// endian purposes.
	void	ReadFromBuffer(uchar* Buffer);
	// Writes the time back to Buffer (4 bytes), uses PutInt().
	void	WriteToBuffer(uchar* Buffer);

	// Time values are supposed to be positive here!
	CDVDTime	&operator += ( CDVDTime & );
	CDVDTime	&operator -= ( CDVDTime & );

	bool	IsZero();

};

// Utility functions...
CompositeSRC ParseSRC(uchar* NavPack);							// From uchar* -> CompositeSRC
void WriteSRC(uchar* NavPack, CompositeSRC SRC);				// CompositeSRC -> uchar*.
void FracToSrc(CompositeSRC CSRC, uint &SRCL, ushort &SRCH);	// From CompositeSRC -> SRCL / SRCH
void SrcToFrac(uint SRCL, ushort SRCH, CompositeSRC &CSRC);		// From SRCL / SRCH -> CompositeSRC
void AddSRC(CompositeSRC &SRC, CompositeSRC SRC2);				// Add SRC += SRC2;
void IncSRC(CompositeSRC &SRC);									// Increment SRC by 1 pack slot.
bool SRC1GESRC2(CompositeSRC SRC1, CompositeSRC SRC2);			// SRC1 >= SRC2
uint ParsePTS(uchar* PTSIn);									// From 5 byte PTS -> uint;
// For PTSPreset, use 
//   - 0x00 for no PTS, 
//   - 0x20 for PTS only, 
//   - 0x30 for PTS part of PTS+DTS,
//   - 0x10 for DTS part of PTS+DTS.
void BuildPTS(uchar* PTSOut, uint PTSIn, uchar PTSPreset);		// From uint -> 5Byte PTS.


#endif