/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "SubPic.h"
#include "Cell.h"

#ifndef _ASSERT
#include <crtdbg.h>
#endif


#define min(A,B)((A < B)?A:B)
#define max(A,B)((A > B)?A:B)

#define MAX_WIDTH 800
#define MAX_HEIGHT 600

// This is static, shared by all subpics.
static uchar PixelArray[MAX_WIDTH*MAX_HEIGHT];
uchar*	CSubPic::m_Pixels;


/////////////////////////////////////////////////////////////////
int CSubPic::m_DisplayHeight = 576;
int CSubPic::m_DisplayWidth = 720;

CSubPic::CSubPic()
{
	// Create OCR output storage.
	m_OCRLines = new uchar[32][128];
	Reset();
	m_StartOfLine = new ushort[MAX_HEIGHT];
	m_Data = 0;
	m_Pixels = PixelArray;
	m_NextSPUInStream = 0;
}

CSubPic::~CSubPic()
{
	delete[] m_Data;
	delete[] m_StartOfLine;
	delete[] m_OCRLines;
}

void CSubPic::Reset(bool ResetAll)
{
	// Reset a bunch of values.
	m_NavPackLBA = 0;
	m_CurrentBit = 0;		
	m_CurrentByte = 0;	
	m_PixelCount = 0;	
	m_CurrentPixelX = 0;	
	m_CurrentPixelY = 0;	
	m_SubPicWidth = 0;		
	m_SubPicHeight = 0;
	m_ModSubPicWidth = 0;
	m_SubPicSize = 0;
	m_DataBytesRead = 0;
	m_DataBytesWritten = 0;
	m_TopMostPixel = -1;
	m_BottomMostPixel = -1;
	m_TrueBottomLine = -1;
	m_BottomMostPixelT = -1;
	m_LastVMove = 0;
	m_LastHMove = 0;
	m_BackgroundPix = -1;
	m_BodyPix = -1;
	m_HasButtons = 0;
	m_VerticalBorders = 0;
	m_PTS = 0;
	m_CLUT = 0;
	// By default a '\0' indicates no more data in the OCR.
	int i;
	for(i=0;i<32;i++)
	{
		m_OCRLines[i][0]='\0';
	}

	m_StartX = m_StartXOrig = 0;
	m_StartY = m_StartYOrig = 0;
	m_EndX = m_EndXOrig = 0;
	m_EndY = m_EndYOrig = 0;
	m_TopFieldOffset = 0;
	m_BotFieldOffset = 0;
	m_EndOfField = 0;

	m_AlphaPtr = 0;
	m_ColorPtr = 0;
	m_SelectedDCSQT = 0;

	m_SPUReadOK = 0;
	m_SubPicDecoded = 0;
	m_RightMostPixel = 0;
	m_LeftMostPixel = 720;
	m_NumPackInSPU = 0;
	memset(m_FilesFrom,0,32*sizeof(FILE*));
	memset(m_LBAFrom,0,32*sizeof(int));
	memset(m_CLBAFrom,0,32*sizeof(int));
	memset(m_NumPix,0,4*sizeof(ushort));

	m_Modified = 0;
	m_VOBID = 0;
	m_CELLID = 0;
	m_StreamID = 0;
	m_SPUNumber = 0;

}


uchar CSubPic::pop2()
{
	_ASSERT(m_CurrentBit >= 1);
	uchar val = m_Data[m_CurrentByte];
	// Move val all the way to the right, then mask, keeping last 2 bits.
	val >>= (m_CurrentBit-1);
	val &= 0x03;
	// Decrementn m_CurrentBit, if needed go into next byte.
	m_CurrentBit -= 2;
	if(m_CurrentBit == -1)
	{
		m_CurrentBit = 7;
		m_CurrentByte ++;
	}
	_ASSERT(m_CurrentByte <= m_EndOfField);
	return val;
}

uchar CSubPic::pop4()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop2();
	return val*4+pop2();
}

uchar CSubPic::pop8()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop4();
	return val*16+pop4();
}

int CSubPic::Out(int n, int c)
{
	int i;
	static int LeftMostPixelForThisLine = -1;

	if(n == 0)
	{
		// Special case: output until end of line.
		n = m_SubPicWidth-m_CurrentPixelX;
	}

	// The background pixel type is the first one we see. This should work most of the time, but not always.
	if(!m_SubPicDecoded && m_PixelCount == 0)
		m_BackgroundPix = c;

	// If this isn't a background pix, remember if this is the first one we've seen.
	if(!m_SubPicDecoded && c != m_BackgroundPix && LeftMostPixelForThisLine == -1)
		LeftMostPixelForThisLine = m_CurrentPixelX;

	// Output pixel m_Data. Also count how many we see of each pixel.
	for(i=0;i<n;i++)
		m_Pixels[m_PixelCount++] = c;
	m_NumPix[c] += n;
	m_CurrentPixelX += n;

	// If this is not a background pixel, make a note if it's the rightmost one seen so far.
	if(!m_SubPicDecoded && c != m_BackgroundPix) 
	{
		m_RightMostPixel = max(m_RightMostPixel,m_CurrentPixelX-1);
		if(m_TopMostPixel == -1) m_TopMostPixel = m_CurrentPixelY;
		m_BottomMostPixelT = m_CurrentPixelY;
	}

	_ASSERT(m_CurrentPixelX <= m_SubPicWidth);
	if(m_CurrentPixelX == m_SubPicWidth)
	{
		// End of line! Keep track of the leftmost pixel.
		if(!m_SubPicDecoded && LeftMostPixelForThisLine != -1)
			m_LeftMostPixel = min(LeftMostPixelForThisLine,m_LeftMostPixel);
		LeftMostPixelForThisLine = -1;
		m_CurrentPixelX = 0;
		m_CurrentPixelY += 2;
		m_PixelCount += m_SubPicWidth;

		// If at the end of a line, we're not at a byte boundary, 4 fill bits of 0 are added.
		_ASSERT(m_CurrentBit == 7 || m_CurrentBit == 3);
		if(m_CurrentBit == 3)
		{
			m_CurrentBit = 7;
			m_CurrentByte ++;
		}
		if(!m_SubPicDecoded)
			m_StartOfLine[m_CurrentPixelY] = m_CurrentByte;
	}
	if(m_CurrentPixelY >= m_SubPicHeight)
		return -1;
	else 
		return 0;
}

// RLE decoding... See http://www.mpucoder.com/DVD/spu.html
void CSubPic::RLEDecode(int Field)
{
	int n;
	int c;
	int ret = 0;
	m_CurrentBit = 7;
	if(Field == 0)
	{
		m_PixelCount = 0;
		m_CurrentPixelX = m_CurrentPixelY = 0;
		m_CurrentByte = m_TopFieldOffset;
		m_EndOfField = m_BotFieldOffset;
	}
	else
	{
		m_PixelCount = m_SubPicWidth;
		m_CurrentPixelX = 0;
		m_CurrentPixelY = 1;
		m_CurrentByte = m_BotFieldOffset;
		m_EndOfField = m_OffsetToCmd;
	}

	while(ret == 0)
	{
		n = pop2();
		if(n != 0)
		{
			// 4 bits: nncc
			c = pop2();
			ret = Out(n,c);
		}
		else
		{
			// 00...
			n=pop2();
			if(n != 0)
			{
				// 8 bits: 00nnnncc
				n = 4*n+pop2();
				c = pop2();
				ret = Out(n,c);
			}
			else
			{
				// 0000...
				n=pop2();
				if(n != 0)
				{
					// 12 bits: 0000nnnnnncc
					n = n*16+pop4();
					_ASSERT(n>0);
					c = pop2();
					ret = Out(n,c);
				}
				else
				{
					// 16 bits: 000000nnnnnnnncc
					n = pop8();
					c = pop2();
					ret = Out(n,c);
				}
			}
		}
	}
}

// Push functions to encode pixel data.
void CSubPic::push4(ushort Code)
{
	if(m_CurrentBit == 0)
		m_Data[m_CurrentByte] = Code << 4;
	else
		m_Data[m_CurrentByte] |= Code;
	
	m_CurrentBit += 4;
	if(m_CurrentBit == 8)
		m_CurrentBit = 0, m_CurrentByte++;
}

void CSubPic::push8(ushort Code)
{
	// Not optimal, but easy!
	push4(Code >> 4);
	push4(Code & 0x000F);
}

void CSubPic::push12(ushort Code)
{
	// Not optimal, but easy!
	push4(Code >> 8);
	push8(Code & 0x00FF);
}

void CSubPic::push16(ushort Code)
{
	// Not optimal, but easy!
	push8(Code >> 8);
	push8(Code & 0x00FF);
}

void CSubPic::push(int count, uchar c) // Encode character c n times.
{
	// Special case, n==0;
	ushort Code = (ushort) c;
	Code |= count << 2;
	if(count==0 || count >= 64)
	{
		push16(Code);
		return;
	}
	if(count >= 16)
	{
		push12(Code);
		return;
	}
	if(count >= 4)
		push8(Code);
	else
		push4(Code);
}

void CSubPic::RLEEncode(int Field, int StartCol, int EndCol, int StartLine, int EndLine)
{
	int RepeatCount;
	uchar PixelType;
	int Line = (Field == 0) ? StartLine : StartLine+1;
	EndLine = min(EndLine,m_SubPicHeight-1);
	EndCol = min(EndCol,m_SubPicWidth-1);
	int Col = 0;
	uchar* Ptr = m_Pixels + Line*m_SubPicWidth;

	m_CurrentBit = 0;
	if(Field == 0)
	{
		m_CurrentByte = 4;		// 4 bytes for the size of the data, and the offset to commands.
		m_TopFieldOffset = 4;
	}
	else m_CurrentByte = m_BotFieldOffset;

	int PixCount=0;
	for(;Line<=EndLine;Line+=2)
	{
		Ptr = m_Pixels + Line*m_SubPicWidth + StartCol;
		for(Col = StartCol;Col<=EndCol;)
		{
			PixelType = *Ptr;
			RepeatCount = 0;
			// Check whether this character is repeated until the end of line.
			int Col2;
			uchar* Ptr2 = Ptr;
			for(Col2=Col;Col2<=EndCol;Col2++,RepeatCount++,Ptr2++)
				if(*Ptr2 != PixelType) break;
			if(Col2 > EndCol)
			{
				// Same character until end of line. If it appears more than 64 times, it's more
				// advantageous to output a special 16bit code. Otherwise, go with the 12bit code.
				push(RepeatCount >= 64 ? 0 : RepeatCount, PixelType);
				Ptr = Ptr2;
				// We're done with this line.
				break;
			}
			// This pixel type does not span the rest of the line. Update our pointers etc.
			RepeatCount = min(RepeatCount,255);
			Ptr += RepeatCount;
			Col += RepeatCount;
			_ASSERT(Col <= EndCol);
			// Here, either RepeatCount == 255 or we have a new char.
			// In any case, output data: We saw PixelType RepeatCount times.
			push(RepeatCount,PixelType);
			// And continue until we're done with this line.
		}
		// End of line. If we're not at a byte boundary, output 4 "0" bytes.
		if(m_CurrentBit != 0)
			push4(0);
	}
	// Set the pointer for the other field, if Field == 0;
	if(Field == 0)
		m_BotFieldOffset = m_CurrentByte;
}

void CSubPic::CropBackground()
{
	if(ContainsHighlights())
		return;
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return ;
	}
	// First of, undo any move. I shouldn't have to do that, but I'm taking the easy road.
	m_TopFieldOffset = m_StartOfLine[0];
	m_BotFieldOffset = m_StartOfLine[1];
	m_StartY = m_StartYOrig;
	m_EndY = m_EndYOrig;
	m_StartX = m_StartXOrig;
	m_EndX = m_EndXOrig;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	m_SubPicWidth = m_EndX-m_StartX+1;
	// This is needed so the bitmap corresponds to the original, unmodified subpic.
	DecodeSubpic();

	// Start colomns and lines.
	int StartCol = max(0,m_LeftMostPixel-1);
	StartCol = max(StartCol,0);
	int EndCol = min(m_RightMostPixel+1,m_SubPicWidth-1);
	
	int StartLine = max(0,m_TopMostPixel-1);
	StartLine = 2*(StartLine/2);
	int EndLine = min(m_SubPicHeight-1,m_BottomMostPixel+2);

	// RLE-Encode the two fields.
	RLEEncode(0,StartCol,EndCol,StartLine,EndLine);
	RLEEncode(1,StartCol,EndCol,StartLine,EndLine);

	// Now this is needed to place the subpic exactly where it was before cropping.
	m_StartXOrig += StartCol;
	m_StartX = m_StartXOrig;
	m_SubPicWidth = EndCol-StartCol+1;
	m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
	m_StartYOrig += StartLine;
	m_StartY = m_StartYOrig;
	m_SubPicHeight = EndLine-StartLine+1;
	m_EndX = m_EndXOrig = m_StartXOrig + m_SubPicWidth -1;
	m_EndY = m_EndYOrig = m_StartYOrig + m_SubPicHeight -1;
	WriteDAREA(m_StartXOrig, m_StartYOrig, m_EndXOrig, m_EndYOrig);
	int Offset = FindCommand(SET_DSPXA);
	PutShort(m_Data,Offset+1,m_TopFieldOffset);
	PutShort(m_Data,Offset+3,m_BotFieldOffset);
	m_Modified = true;
	// This is very important: it will ensure that the m_StartOfLine array will be updated.
	m_SubPicDecoded = 0;
	m_BottomMostPixelT = m_BottomMostPixel = m_TopMostPixel = -1;
	m_RightMostPixel = 0; m_LeftMostPixel = 720;
//	m_LastVMove = 0;
//	m_LastHMove = 0;
	DecodeSubpic();
	// Reapply moves.
	VMoveSubPic(m_LastVMove);
	HMoveSubPic(m_LastHMove);
}

int CSubPic::ReadThisPack(uchar* Data, FILE* fd, int LBA, int NavPackLBA, ushort VOBID, 
						  uchar CELLID, uchar StreamID, int CLBA)
{
Start:
	_ASSERT(m_SPUReadOK == 0);
	m_FilesFrom[m_NumPackInSPU] = fd;
	m_LBAFrom[m_NumPackInSPU] = LBA;
	m_CLBAFrom[m_NumPackInSPU] = CLBA;

	if(m_DataBytesRead == 0)
	{
		// Find start of subpic data... It's at 0x18+len...I also make sure this has a PTS value!
		// That's the only reliable way (?) to make sure this is truly the start of a SPU.
		uint len = (uint) (Data[0x16]);
		_ASSERT(len > 0 && (Data[0x15] & 0x80));
		m_PTS = ParsePTS(Data+0x17);
		// Also, save the first 64 bytes of the subpic pack.
		memcpy(m_PackStartData,Data,64);
		m_SubPicSize = GetShort(Data,0x18+len);
		m_Data = new uchar[m_SubPicSize];

		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Data+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
		// Now that we have the beginning of the data buffer, we can get the length and the
		// offset to commands.
		m_SubPicSize = GetShort(m_Data,0);
		m_OffsetToCmd = GetShort(m_Data,2);
		_ASSERT(m_OffsetToCmd < m_SubPicSize);
		if(m_OffsetToCmd >= m_SubPicSize)
		{
			ERROR2("Inconsistent data in subpic at LBA %d. Offset to commands past end of SPU",m_LBAFrom[0]);
			m_OffsetToCmd = -1;
		}
		m_NavPackLBA = NavPackLBA;
		m_VOBID = VOBID;
		m_CELLID = CELLID;
		m_StreamID = StreamID;
	}
	else
	{
		_ASSERT(m_SubPicSize != 0 && m_DataBytesRead != 0);
		// Check that the PES header data length is 0 or that no PTS is present (that would indicate
		// we missed the first SPU pack!).
		uint len = (uint) (Data[0x16]);
//		_ASSERT(len == 0 || (Data[0x15] & 0x80)==0);
		if(len != 0 && (Data[0x15] & 0x80) != 0)
		{
			// This is the beginning of a SPU! We're missing a pack from the previous one!
			ERROR2("SPU starting at LBA %d is incomplete! Dropping it...\n",m_CLBAFrom[0]);
			// Best thing to do is forget the preceding one and start again with this one...
			Reset();
			goto Start;
		}
		// Read as much as you can, except if we have enough data.
		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Data+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
	}
	m_NumPackInSPU++;
	if(m_SubPicSize == m_DataBytesRead)
	{
		// We're done! Now, let's read a few things.
		m_SPUReadOK = true;
		ParseCommands();
		SelectDCSQT(m_SelectedDCSQT);
		return 1;
	}
	else return 0;
}

void CSubPic::SelectDCSQT(int DCSQT)
{
	if(DCSQT >= m_DCSQT.size())
		return;
	Commands *TheCommand;
	if(FindCommand(SET_COLOR,DCSQT,&TheCommand) != -1)
		m_ColorPtr = TheCommand->m_Params;
	if(FindCommand(SET_CONTR,DCSQT,&TheCommand) != -1)
		m_AlphaPtr = TheCommand->m_Params;
#if 0
	// Could use this to show scrolling effects...
	if(FindCommand(SET_DAREA,DCSQT,&TheCommand) != -1)
	{
	m_StartX = TheCommand->m_Params[3];
	m_EndX = TheCommand->m_Params[2]; 
	m_StartY = TheCommand->m_Params[1];
	m_EndY = TheCommand->m_Params[0]; 
	if(!m_SubPicDecoded)
	{ m_StartYOrig = m_StartY ; m_EndYOrig = m_EndY ; }
	_ASSERT(m_StartX < 1024 && m_EndX < 1024 && m_StartY < 1024 && m_EndY < 1024);
	m_SubPicWidth = m_EndX - m_StartX + 1;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
	}
#endif

	m_SelectedDCSQT = DCSQT;
}

bool CSubPic::FindString(const char* Search)
{
	int j;
	int k=0;
	char Str1[1024]; 
	strcpy(Str1,Search); 
	// Make search string lowercase.
	_strlwr(Str1);
	char Str2[1024];
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0')
			return false;
		// Copy OCR output and make it lowercase.
		strcpy(Str2,(char*)m_OCRLines[j]); _strlwr(Str2);
		// Then search.
		if(strstr(Str2,Str1) != 0)
			return true;
	}
}

bool CSubPic::FindChars(const char* Search)
{
	int j;
	int k=0;
	char Str1[1024]; 
	strcpy(Str1,Search); 
	// Again, make it lowercase.
	_strlwr(Str1);
	char Str2[1024];
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0')
			return false;
		// Copy OCR output and make it lowercase.
		strcpy(Str2,(char*)m_OCRLines[j]); _strlwr(Str2);
		// Look for characters.
		if(strpbrk(Str2,Str1) != 0)
			return true;
	}
}

bool CSubPic::IsTextBracketed()
{
	int i,j;
	bool Found = false;
	for(j=0;;j++)
	{
		if(m_OCRLines[j][0] == '\0') break;
		// Find first non-space character.
		for(i=0;;i++)
			if(m_OCRLines[j][i] != ' ' && m_OCRLines[j][i] != '-' || m_OCRLines[j][i] == '\0') break;
		if(m_OCRLines[j][i] != '[' && m_OCRLines[j][i] != '(')
			return false;
		// Now find the matching closing bracket.
		Found = false;
		while(1)
		{
			if(m_OCRLines[j][i] == ']' || m_OCRLines[j][i] == ')') break;
			if(m_OCRLines[j][i] == '\0' && i==0)
				return false;
			// You might have to go to the next line!
			if(m_OCRLines[j][i] == '\0')
				i=-1,j++;
			i++;
		}
		// The matching bracket was found.
		Found = true;
		// Now check that there's no non-space character after that.
		for(i++;;i++)
		{
			if(m_OCRLines[j][i] == '\0') break;
			if(m_OCRLines[j][i] != ' ') return false;
		}
		// Continue on to the next line if any...
	}
	return Found;
}


bool CSubPic::ContainsHighlights()
{
	// This is a simple, but not foolproof test. If the subpic's navpack defines buttons, and if
	// our start command is a forced start, chances are these are highlights. I could also test
	// transparency, but some highlights are not transparents.
	return (m_HasButtons && FindCommand(FSTA_DSP,0) != -1);
}

int CSubPic::SaveSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	uchar Buffer[SECTOR_SIZE];
	m_DataBytesWritten = 0;
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to save.
		FILE* fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 || fread(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then save back our subpic data...
		uint len = (uint) (Buffer[0x16]);
		int ToWrite = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesWritten);
		memcpy(Buffer+0x18+len,m_Data+m_DataBytesWritten,ToWrite);
		// Write back the beginning of the first pack, in case we modified it.
		_ASSERT(0x18+len <= 64);
		if(i==0)
			memcpy(Buffer,m_PackStartData,0x18+len);
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 ||
			fwrite(Buffer,sizeof(char),SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to write %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		m_DataBytesWritten += ToWrite;
	}
	m_Modified = 0;
	return 0;
}

int CSubPic::ReloadSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	uchar Buffer[SECTOR_SIZE];
	m_DataBytesRead = 0;
	m_SubPicDecoded = 0;
	m_RightMostPixel = 0;
	m_LeftMostPixel = 720;
	m_TopMostPixel = m_BottomMostPixel = -1;
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to reload.
		FILE* fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 || fread(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then copy the subpic data...
		uint len = (uint) (Buffer[0x16]);
		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Buffer+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
	}
	if(m_DataBytesRead != m_SubPicSize)
	{
		ERROR1("Failed to re-load enough data");
		return -1;
	}
	ParseCommands();
	SelectDCSQT(m_SelectedDCSQT);
	m_Modified = 0;
	m_LastVMove = 0;
	m_LastHMove = 0;
	m_VerticalBorders = 0;
	return 0;
}

int CSubPic::FindTrueBottomLine()
{
	if(m_SPUReadOK == false)
		return -1;
	
	int NumActivePixels[1024];
	uchar* Ptr = m_Pixels;
	int i,j;
	// Count how many active (non-background) pixels we have per line.
	for(i=0;i<m_SubPicHeight;i++)
	{
		NumActivePixels[i] = 0;
		for(j=0;j<m_SubPicWidth;j++,Ptr++)
		{
			if(*Ptr != m_BackgroundPix)
				NumActivePixels[i]++;
		}
	}
	// Find the maximum number of set pixels in the last subpic line.
	int MaxNum = 0;
	for(i=m_SubPicHeight-1;i>=0;i--)
	{
		// Stop at the first empty line, after we've found one.
		if(NumActivePixels[i] == 0 && MaxNum > 0)
			break;
		MaxNum = max(MaxNum,NumActivePixels[i]);
	}

	// The true bottom line is the first line that has at least half as many set pixels as the largest
	// line
	for(i=m_SubPicHeight-1;i>=0;i--)
	{
		if(2*NumActivePixels[i] > MaxNum)
			break;
	}
	m_TrueBottomLine = (i>=0) ? i : m_BottomMostPixel;
	return 0;
}

int CSubPic::DecodeSubpic()
{
	if(m_SPUReadOK == false)
		return -1;
	if(!m_SubPicDecoded)
	{
		// Only set these the first time around. This way we'll keep a memory of them.
		m_StartOfLine[0] = m_TopFieldOffset;
		m_StartOfLine[1] = m_BotFieldOffset;
	}
	memset(m_NumPix,0,4*sizeof(ushort));
	RLEDecode(0);
	RLEDecode(1);

	if(m_BottomMostPixel == -1)
	{
		m_BottomMostPixel = m_BottomMostPixelT;
		// Now find the true bottom line of the subpic.
		FindTrueBottomLine();
	}
	// It's possible that the subpic is entirely empty!
	if(m_TopMostPixel == -1) 
		m_TopMostPixel = 0;

	if(!m_SubPicDecoded)
		FindBodyPixel();

	m_SubPicDecoded = true;
	return 0;
}

void CSubPic::FindBodyPixel()
{
#if 0
	// Now that we have the background pixel, we can determine the body pixel.
	if(m_CLUT)
	{
		int Mean[4]={0,0,0,0};
		int i,MaxMean;
		int tie=-1;
		for(i=MaxMean=0;i<4;i++)
		{
			// Ignore background pix.
			if(i==m_BackgroundPix) continue;
			// Calculate mean of RGB, the largest value yields the lightest color.
			Mean[i]=(m_CLUT[3*m_ColorPtr[i]]+m_CLUT[3*m_ColorPtr[i]+1]+m_CLUT[3*m_ColorPtr[i]+2]);
			if(Mean[i]>MaxMean) m_BodyPix=i,MaxMean = Mean[i],tie=-1;
			else if(Mean[i] == MaxMean) tie = i;
		}
		// It can happen that 2 pixel types have exactly the same color.
		// I'm picking the one that has the most pixels.
		if(tie != -1 && m_NumPix[tie] > m_NumPix[m_BodyPix])
			m_BodyPix = tie;
	}
#endif

#if 0
	int i;
	int line = m_TopMostPixel+3;
	uchar FirstNonBack = -1;
	uchar SecondNonBack = -1;
	uchar ThirdNonBack = -1;
	m_BodyPix = -1;
	// Here's the idea: go along horizontal lines until you find a succession of b-x-y-z pixels (b=backgd,
	// x y and z other types). Once you find that, you can assume that y or z is the character body...
	for(line;line<m_SubPicHeight;line++)
	{
		FirstNonBack = -1;
		SecondNonBack = -1;
		ThirdNonBack = -1;
		uchar* Ptr = m_Pixels + line*m_SubPicWidth;
		for(i=0;i<m_SubPicWidth;)
		{
			// Find background pixel
			for(;i<m_SubPicWidth;i++)
				if(Ptr[i] == m_BackgroundPix) break;
			if(i == m_SubPicWidth) break;
			// Then find first non-background pixel (x type above)
			for(;i<m_SubPicWidth;i++)
				if(Ptr[i] != m_BackgroundPix) break;
			if(i == m_SubPicWidth) break;
			FirstNonBack = Ptr[i];
			// Now look for the next type (y above).
			for(;i<m_SubPicWidth;i++)
				if(Ptr[i] != FirstNonBack) break;
			if(i == m_SubPicWidth) break;
			SecondNonBack = Ptr[i];
			// Now make sure the y is not a background pixel. If it is, keep on looking!
			if(SecondNonBack == m_BackgroundPix) 
				continue;
			// And look for the next type (z above).
			for(;i<m_SubPicWidth;i++)
				if(Ptr[i] != SecondNonBack) break;
			if(i == m_SubPicWidth) break;
			ThirdNonBack = Ptr[i];
			// We found it! The body is either Second or Third (y or z).
			if(ThirdNonBack == FirstNonBack || ThirdNonBack == m_BackgroundPix)
				m_BodyPix = SecondNonBack;
			else
				m_BodyPix = ThirdNonBack;
			break;
		}
		// We found it, break.
		if(m_BodyPix != -1)
			break;
	}
	// For testing.
	assert(m_BodyPix == 1);
#else
	// This seems a bit more reliable. Exclude the background pixel, and the first anti-alias pixel found
	// Then pick between the two remaining ones (if there are two!)
	int i,j;
	int line = m_TopMostPixel+1;
	uchar* Ptr = m_Pixels + line*m_SubPicWidth;
	// Find first non-background pixel.
	for(i=0;i<m_SubPicWidth;i++)
		if(Ptr[i] != m_BackgroundPix) break;
	// This has GOT to be an antialiasing pixel.
	uchar AntiAlias1 = Ptr[i];
	// If only three pixel types are used, we're pretty much done!
	uchar Others[2] = {-1,-1};
	for(i=j=0;i<4;i++)
	{
		if(i != m_BackgroundPix && i != AntiAlias1 && m_NumPix[i] != 0) 
			Others[j++] = i;
	}
	if(j == 0)
	{
		// No other pixel type!
		m_BodyPix = AntiAlias1;
		return;
	}
	if(j == 1)
	{
		// Only 1 other pixel type, we're done.
		m_BodyPix = Others[0];
		return;
	}
	// More difficult case. There are 2 other pixel types, and we need to decide which one is used as
	// the character body. Go through bitmap, and look until you see o1-o2-o1. This will indicate 
	// that o2 is the body.
	uchar First = -1;
	uchar Second = -1;
	int N = (m_BottomMostPixel + 1)*m_SubPicWidth;
	m_BodyPix = -1;
	for(i=line*m_SubPicWidth;i<N;i++)
	{
		// Find first non-background, non-AntiAlias1 pixel.
		for(;i<N;i++)
			if(m_Pixels[i] != m_BackgroundPix && m_Pixels[i] != AntiAlias1) break;
		if(i == N) break;
		First = m_Pixels[i];
		// Now look for next pixel type;
		for(;i<N;i++)
			if(m_Pixels[i] != First) break;
		if(i == N) break;
		// Make sure we're seeing o2!
		if(m_Pixels[i] == m_BackgroundPix || m_Pixels[i] == AntiAlias1)
			continue;
		Second = m_Pixels[i];
		// Now look for o1 again.
		for(;i<N;i++)
			if(m_Pixels[i] != Second) break;
		if(i == N) break;
		// Make sure we're seeing o1!
		if(m_Pixels[i] != First)
			continue;
		// We found it!
		m_BodyPix = Second;
		break;
	}

	// For testing...
//	assert(m_BodyPix == 2);
#endif
}


int CSubPic::OutputPicToTxtFile(const char* FileName)
{
	// Output "bitmap?"
	if(m_SPUReadOK == false)
		return -1;
	FILE* foo = fopen("FileName","w");
	if(!foo)
		return -1;
	int i,j;
	uchar* Ptr = m_Pixels;
	for(i=0;i<m_SubPicHeight;i++)
	{
		for(j=0;j<m_SubPicWidth;j++)
			fprintf(foo,"%d ",*Ptr++);
		fprintf(foo,"\n");
	}
	fclose(foo);
	return 0;
}

int CSubPic::FindCommand(CommandID Command, int DCSQT, Commands** TheCommand)
{
	if(m_SPUReadOK == false)
		return -1;
	// Go to DCSQT
	_ASSERT(DCSQT < m_DCSQT.size());
	if(DCSQT >= m_DCSQT.size())
		return -1;

	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	int i;
	for(i=0;i<TheDCSQT->m_Commands.size();i++)
	{
		if(TheDCSQT->m_Commands[i].m_CommandID == Command)
		{
			if(TheCommand)
				*TheCommand = &TheDCSQT->m_Commands[i];
			return TheDCSQT->m_Commands[i].m_Offset;
		}
	}
	return -1;
}

int CSubPic::ChangeCommand(CommandID OldCommand, CommandID NewCommand, int DCSQT)
{
	int Offset = FindCommand(OldCommand,DCSQT);
	if(Offset == -1)
		return -1;
	// Now change the command!
	m_Data[Offset] = (uchar)NewCommand;
	ParseCommands();
	SelectDCSQT(m_SelectedDCSQT);
	if(NewCommand != OldCommand)
		m_Modified = true;
	return 0;
}

void CSubPic::GetMinMaxDCSQTDelay(int &MinDurMs, int &MaxDurMs, int DCSQT)
{
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	DCSQT--;
	if(DCSQT < 0 || DCSQT >= m_DCSQT.size())
	{
		MinDurMs = MaxDurMs = 0;
		return;
	}
	// A CDSQT must be at least 1 frame duration later than the previous one.
	MinDurMs = m_DCSQT[DCSQT].m_Delay * 1024/90;
	int FrameDur = (VideoFormat == FORMAT_PAL ? 40 : 33);
	MinDurMs += FrameDur;
	// There must be at least 1 frame duration between then last DCSQT and the next spu PTS.
	if(m_NextSPUInStream)
		MaxDurMs = m_NextSPUInStream->m_PTS / 90 - m_PTS / 90 - FrameDur;
	else
		// JEAN: FIX THIS!
		MaxDurMs = 10000;
}

int CSubPic::ChangeDCSQTDelay(int DurationMs, int DCSQT)
{
	// Get our boundaries, make sure the subpics don't step on one another.
	int MinDurMs, MaxDurMs;
	GetMinMaxDCSQTDelay(MinDurMs, MaxDurMs, DCSQT);
	DurationMs = max(DurationMs,MinDurMs);
	DurationMs = min(DurationMs,MaxDurMs);
	// Now adjust the delay of the DCSQT.
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	if(DCSQT < 0 || DCSQT >= m_DCSQT.size())
		return -1;
	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	TheDCSQT->m_Delay = DurationMs*90/1024;
	uchar* Ptr = m_Data + TheDCSQT->m_OffsetToThis;
	PutShort(Ptr,0,TheDCSQT->m_Delay);
	m_Modified = true;
	return 0;
}

int CSubPic::GetDCSQTDelay(int DCSQT)
{
	if(DCSQT == -1) DCSQT = m_DCSQT.size()-1;
	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	return TheDCSQT->m_Delay * 1024 / 90;
}


// This reads 4 color or alpha values and stores them into the Dest array.
void ReadColorAlpha(ushort* Dest, uchar* Src)
{
	Dest[3] = Src[0]>>4;
	Dest[2] = Src[0]&0x0F;
	Dest[1] = Src[1]>>4;
	Dest[0] = Src[1]&0x0F;
}

// This takes an array of 4 color or alpha values and stores them into the Dest array.
void WriteColorAlpha(uchar* Dest, ushort* Src)
{
	Dest[1] &= 0xF0; Dest[1] |= Src[0];		
	Dest[1] &= 0x0F; Dest[1] |= (Src[1] << 4);
	Dest[0] &= 0xF0; Dest[0] |= Src[2];		
	Dest[0] &= 0x0F; Dest[0] |= (Src[3] << 4);
}

int CSubPic::ParseCommands(int NumDCSQT, int OffsetToDCSQT)
{
	if(m_SPUReadOK == false)
		return -1;
	if(NumDCSQT == 0)
	{
		m_DCSQT.clear();
		m_CHCOL.clear();
	}
	SP_DCSQT Foo;
	m_DCSQT.push_back(Foo);
	SP_DCSQT* TheDCSQT = &m_DCSQT[NumDCSQT];
	// Get time info and offset info.
	TheDCSQT->m_OffsetToThis = (NumDCSQT == 0) ? m_OffsetToCmd : OffsetToDCSQT;
	uchar* Ptr = m_Data + TheDCSQT->m_OffsetToThis;
	TheDCSQT->m_Delay = GetShort(Ptr,0);
	TheDCSQT->m_OffsetToNext = GetShort(Ptr,2);
	TheDCSQT->m_Commands.clear();

	// Now parse commands...
	Ptr += 4;
	bool Done = 0;
	while(!Done)
	{
		// Get command:
		// Add a command.
		Commands Temp;
		TheDCSQT->m_Commands.push_back(Temp);
		Commands *TheCmd = &TheDCSQT->m_Commands.back();
		TheCmd->m_CommandID = (CommandID) *Ptr;
		TheCmd->m_Offset = Ptr-m_Data;
		TheCmd->m_NumParams = 0;
		switch(*Ptr)
		{
		case FSTA_DSP : Ptr ++;  break;
		case STA_DSP : Ptr ++; break;
		case STP_DSP : Ptr ++; break;
		case SET_COLOR : 
			TheCmd->m_NumParams = 4;
			ReadColorAlpha(TheCmd->m_Params,Ptr+1);
			Ptr += 3; break;
		case SET_CONTR : 
			TheCmd->m_NumParams = 4;
			ReadColorAlpha(TheCmd->m_Params,Ptr+1);
			Ptr += 3; break;
		case SET_DAREA : 
			TheCmd->m_NumParams = 4;
			TheCmd->m_Params[3] = ((Ptr[1])<<4)|((Ptr[2])>>4);
			TheCmd->m_Params[2] = ((Ptr[2] &0x0F) <<8)|(Ptr[3]);
			TheCmd->m_Params[1] = ((Ptr[4])<<4)|((Ptr[5])>>4);
			TheCmd->m_Params[0] = ((Ptr[5] &0x0F) <<8)|(Ptr[6]);
			if(NumDCSQT == 0)
			{
				m_StartX = TheCmd->m_Params[3];
				m_EndX = TheCmd->m_Params[2]; 
				m_StartY = TheCmd->m_Params[1];
				m_EndY = TheCmd->m_Params[0]; 

				if(!m_SubPicDecoded)
				{ 
					m_StartYOrig = m_StartY ; m_EndYOrig = m_EndY ; 
					m_StartXOrig = m_StartX; m_EndXOrig = m_EndX;
				}
				_ASSERT(m_StartX < 1024 && m_EndX < 1024 && m_StartY < 1024 && m_EndY < 1024);
				m_SubPicWidth = m_EndX - m_StartX + 1;
				m_SubPicHeight = m_EndY - m_StartY + 1;
				m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
			}
			Ptr += 7; break;
		case SET_DSPXA : 
			TheCmd->m_NumParams = 2;
			TheCmd->m_Params[0] = m_TopFieldOffset = GetShort(Ptr,1);
			TheCmd->m_Params[1] = m_BotFieldOffset = GetShort(Ptr,3);
			_ASSERT(m_TopFieldOffset < 2*m_SubPicSize / 3 && m_BotFieldOffset < 2*m_SubPicSize / 3);
			Ptr += 5; break;
		case CHG_COLCON :
			{
			Ptr++;
			CHG_COLCON_Command CHG_COLCONCommand;
			ParseCHG_COLCON(&Ptr,&CHG_COLCONCommand); 
			CHG_COLCONCommand.m_DCSQT = NumDCSQT;
			m_CHCOL.push_back(CHG_COLCONCommand);
			break;
			}
		case CMD_END : 	Done = 1; break;
		default : 
			// This should NOT happen! Something went wrong here!
			ERROR3("Inconsistent data in subpic at LBA %d. Unknown command %d",m_LBAFrom[0],*Ptr);
			// Setting this to ensure we're exiting!
			TheDCSQT->m_OffsetToNext = TheDCSQT->m_OffsetToThis;
			Done = 1; 
			break;
		}
	}
	if(TheDCSQT->m_OffsetToNext != TheDCSQT->m_OffsetToThis)
		return ParseCommands(m_DCSQT.size(), TheDCSQT->m_OffsetToNext);
	return 0;
}


int CSubPic::ParseCHG_COLCON(uchar** ThePtr, CHG_COLCON_Command* Command)
{
	int bands;
	int i;
	// First are two bytes indicating total size of parameter area, including size.
	uchar* Ptr = *ThePtr;
	ushort ParamAreaSize = GetShort(Ptr,0);

#if 0
	// Don't parse, return immediately!
	Ptr += ParamAreaSize;
	return 0;
#endif

	Ptr += 2;
	// http://www.mpucoder.com/DVD/spu.html#dcsq
	for(bands=0;bands<64;bands++)
	{
		LN_CTLI TheBand;
		// Look for termination code.
		if(GetInt(Ptr,0) == 0x0FFFFFFF)
			break;
		// 4 bytes. 0s ss nt tt where sss is m_CSLN n is m_NumPX_CTLI and ttt is m_CTLN
		TheBand.m_CSLN = GetShort(Ptr,0);				// 0x0sss
		Ptr += 2;
		TheBand.m_NumPX_CTLI = (Ptr[0]>>4);			// 0x0n
		TheBand.m_CTLN = GetShort(Ptr,0) & 0x0FFF;		// 0x0ttt
		Ptr += 2;
		
		_ASSERT(TheBand.m_NumPX_CTLI < 8);
		for(i=0;i<TheBand.m_NumPX_CTLI;i++)
		{
			// 6 bytes, 2 by 2. First the start column
			TheBand.m_Columns[i].m_StartColNum = GetShort(Ptr,0); 
			Ptr += 2;	// Then the new colors.
			ReadColorAlpha(TheBand.m_Columns[i].m_NewColors,Ptr);
			Ptr += 2;	// Then the alphas.
			ReadColorAlpha(TheBand.m_Columns[i].m_NewAlphas,Ptr);
			Ptr += 2;
		}
		Command->m_Bands.push_back(TheBand);
	}
	Ptr += 4;
	_ASSERT(Ptr - *ThePtr == ParamAreaSize);
	*ThePtr = Ptr;
	MESSAGE2("Found CHG_COLCON, %d bands defined",i);
	Command->m_NumBands = bands;

	return 0;
}


uchar Colors[3*4] = {0,0,0,255,255,255,0,255,0,0,0,255};

int CSubPic::SavePicAsPPM(const char* FileName, uchar *CLUT, float FactX, float FactY, uchar* DefColors)
{
	if(m_SPUReadOK == false)
		return -1;
	int NewHeight;
	int NewWidth;
	int TotalByteCount;
	int i,j;
	static uchar  ColorMap[12];

	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_ColorPtr[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_ColorPtr[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_ColorPtr[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_ColorPtr[3],3*sizeof(uchar));
	}
	else if(DefColors)
		memcpy(ColorMap,DefColors,12*sizeof(char));
	else
		memcpy(ColorMap,Colors,12*sizeof(char));

	FILE* fd = fopen(FileName,"wb");
	if(fd == 0)
		return -1;
	
	uchar* BitMap = new uchar[(int)(m_SubPicHeight/FactY+1)*(int)(m_SubPicWidth/ FactX+1) * 3];
	uchar* dest = BitMap;

	NewWidth = (int)(m_SubPicWidth/FactX);

	for(i=NewHeight=0,TotalByteCount=0;;i++)
	{
		// Find start of new line. PPM files are stored top to bottom.
		int yindex = (int)(i * FactY + .5f);
		if(yindex >= m_SubPicHeight)
			break;
		uchar* src = m_Pixels + yindex * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<NewWidth;j++)
		{
			// Find pixel location in line.
			int xindex = (int)(j * FactX + .5f);
			// Copy the three RGB components for that pixel. Note the order is different than for BMP files.
			uchar Pick = 3*src[xindex];
			_ASSERT(Pick < 12);
			*dest++ = ColorMap[Pick];
			*dest++ = ColorMap[Pick+1];
			*dest++ = ColorMap[Pick+2];
			TotalByteCount+=3;
		}
		NewHeight++;
	}

	fprintf (fd,"P6\n%d %d\n255\n", NewWidth, NewHeight);
	fwrite(BitMap,sizeof(char),TotalByteCount,fd);
	fclose(fd);
	delete[] BitMap;
	return 0;
}


uchar* CSubPic::SaveToBitmap(uchar* Bitmap, long &Width, long &Height, float FactX, float FactY, uchar *CLUT, uchar* DefColors)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};

	// Get Alpha and color tables...
	AlphaMap[0]=m_AlphaPtr[0]*16;
	AlphaMap[1]=m_AlphaPtr[1]*16;
	AlphaMap[2]=m_AlphaPtr[2]*16;
	AlphaMap[3]=m_AlphaPtr[3]*16;
	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_ColorPtr[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_ColorPtr[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_ColorPtr[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_ColorPtr[3],3*sizeof(uchar));
	}
	else if(DefColors)
		memcpy(ColorMap,DefColors,12*sizeof(char));
	else
	{
		// Using m_BodyPix to mark the pixels.
		for(int i=0;i<4;i++)
		{
			if(i==m_BodyPix)
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=255;
			else
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=0;
		}
	}

	int i,j;
	int NewHeight;
	int NewWidth = (int)(m_SubPicWidth/FactX);
	// I HAVE to use a floor so the subpic does not end up longer than it should.
	NewWidth = ((NewWidth)/4)*4;
	// But then, I might miss the few last pixels, and that might be bad. Modify the
	// FactX so we don't!
	FactX = (float)m_SubPicWidth/NewWidth;
	for(i=NewHeight=0;i<m_SubPicHeight;i++)
	{
		// Find start of new line. Bitmaps are stored bottom to top.
		int yindex = m_SubPicHeight - 1 - (int)(i* FactY + .5f);
		if(yindex < 0)
			break;
		uchar* src = m_Pixels + yindex * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<NewWidth;j++)
		{
			// Copy the three RGB components for that pixel. The order is BGR!
			int xindex = (int)(j * FactX + .5f);
			uchar Pick = 3*src[xindex];
			ushort Alpha = AlphaMap[src[xindex]];
			_ASSERT(Pick < 12);
			*dest++ = ColorMap[Pick+2]*Alpha/255;	// B
			*dest++ = ColorMap[Pick+1]*Alpha/255;	// G
			*dest++ = ColorMap[Pick]*Alpha/255;	// R
		}
		NewHeight++;
	}
	Width = NewWidth;
	Height = NewHeight;
	return ColorMap;
}

uchar* CSubPic::SaveToAlphaBitmap(uchar* Bitmap, uchar *CLUT,
							 uchar* DefColors)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};

	// Get alpha and color tables.
	AlphaMap[0]=m_AlphaPtr[0]*16;
	AlphaMap[1]=m_AlphaPtr[1]*16;
	AlphaMap[2]=m_AlphaPtr[2]*16;
	AlphaMap[3]=m_AlphaPtr[3]*16;
	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_ColorPtr[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_ColorPtr[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_ColorPtr[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_ColorPtr[3],3*sizeof(uchar));
	}
	else if(DefColors)
		memcpy(ColorMap,DefColors,12*sizeof(char));
	else
	{
		// Using m_BodyPix to mark the pixels.
		for(int i=0;i<4;i++)
		{
			if(i==m_BodyPix)
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=255;
			else
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=0;
		}
	}

	int i,j;
	for(i=m_SubPicHeight-1;i>=0;i--)
	{
		// Find start of new line. PPM files are stored top to bottom.
		uchar* src = m_Pixels + i * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<min(m_SubPicWidth,m_ModSubPicWidth);j++)
		{
			// Copy the three RGB components for that pixel. The order is BGR!
			uchar Pick = 3*src[j];
			_ASSERT(Pick < 12);
			ushort Alpha = AlphaMap[src[j]];
			*dest++ = ColorMap[Pick+2]*Alpha/255;	// B
			*dest++ = ColorMap[Pick+1]*Alpha/255;	// G
			*dest++ = ColorMap[Pick]*Alpha/255;		// R
			*dest++ = (uchar)Alpha;					// Alpha
		}
		// If the width isn't a multiple of 4, add transparent pixels.
		for(;j<m_ModSubPicWidth;j++)
		{
			*dest++ = 0;
			*dest++ = 0;
			*dest++ = 0;
			*dest++ = 0;
		}
	}
	return ColorMap;
}

uchar* CSubPic::MixIntoBitmap(uchar* Bitmap, long Width, long Height, uchar *CLUT, uchar* DefColors)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};

	// Get alpha and color tables.
	AlphaMap[0]=m_AlphaPtr[0]*16;
	AlphaMap[1]=m_AlphaPtr[1]*16;
	AlphaMap[2]=m_AlphaPtr[2]*16;
	AlphaMap[3]=m_AlphaPtr[3]*16;
	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_ColorPtr[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_ColorPtr[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_ColorPtr[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_ColorPtr[3],3*sizeof(uchar));
	}
	else if(DefColors)
		memcpy(ColorMap,DefColors,12*sizeof(char));
	else
	{
		// Using m_BodyPix to mark the pixels.
		for(int i=0;i<4;i++)
		{
			if(i==m_BodyPix)
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=255;
			else
				ColorMap[3*i]=ColorMap[3*i+1]=ColorMap[3*i+2]=0;
		}
	}

	int i,j;
	assert(m_SubPicWidth <= Width);
	assert(m_SubPicHeight <= Height);

	for(i=0;i<m_SubPicHeight;i++)
	{
		// Find start of new line. Bitmaps are stored bottom to top, but our pixel array is stored
		// from top to bottom. FIX: Use the startX and startY values here!!!
		uchar* src = m_Pixels + i * m_SubPicWidth;
		dest = Bitmap + 3*(Height-i-1-m_StartY)*Width + 3*m_StartX;
		// Mix this line.
		for(j=0;j<m_SubPicWidth;j++,dest+=3)
		{
			// Mix the three RGB components for that pixel. The order is BGR!
			uchar Pick = 3*src[j];
			_ASSERT(Pick < 12);
			ushort Alpha = AlphaMap[src[j]];
			// Don't do anything if our pixel is transparent!
			if(Alpha != 0)
			{
				// I could make this faster by doing Alpha++ and using 256 instead of 255..
				dest[0] = dest[0]*(255-Alpha)/255 + ColorMap[Pick+2]*Alpha/255;	// B
				dest[1] = dest[1]*(255-Alpha)/255 + ColorMap[Pick+1]*Alpha/255;	// G
				dest[2] = dest[2]*(255-Alpha)/255 + ColorMap[Pick]*Alpha/255;		// R
			}
		}
	}
	return ColorMap;
}



void CSubPic::WriteDAREA(ushort StartX, ushort StartY, ushort EndX, ushort EndY)
{
	Commands *TheCommand;
	int Offset = FindCommand(SET_DAREA,0,&TheCommand);
	if(Offset != -1)
	{
		m_Data[Offset+1] = (uchar)(StartX >> 4);
		m_Data[Offset+2] = (uchar)((StartX & 0x000F)<<4);
		m_Data[Offset+2] |= (uchar)(EndX >> 8);
		m_Data[Offset+3] = (uchar)(EndX & 0x00FF);
		m_Data[Offset+4] = (uchar)(StartY >> 4);
		m_Data[Offset+5] = (uchar)((StartY & 0x000F)<<4);
		m_Data[Offset+5] |= (uchar)(EndY >> 8);
		m_Data[Offset+6] = (uchar)(EndY & 0x00FF);

		TheCommand->m_Params[3] = StartX;
		TheCommand->m_Params[2] = EndX;
		TheCommand->m_Params[1] = StartY;
		TheCommand->m_Params[0] = EndY;
	}
}

void CSubPic::WriteDSPXA(ushort TopOffset, ushort BotOffset)
{
	int Offset = FindCommand(SET_DSPXA);
	if(Offset != -1)
	{
		PutShort(m_Data,Offset+1,TopOffset);
		PutShort(m_Data,Offset+3,BotOffset);
	}
}

int CSubPic::VMoveSubPic(int NumPixels)
{
	// Decode subpic if needed.
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	// First of, "reset" the subpic, i.e. undo any move that might have been done before...
	m_TopFieldOffset = m_StartOfLine[0];
	m_BotFieldOffset = m_StartOfLine[1];
	m_StartY = m_StartYOrig;
	m_EndY = m_EndYOrig;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	// Make sure NumPixels is even!
	NumPixels = min(NumPixels,m_StartY-m_VerticalBorders+m_TopMostPixel);
	// We want the true bottommost pixel to be above the vertical border, AND the bottom most
	// pixel to be within the frame.
	NumPixels = max(m_StartY+m_TrueBottomLine-m_DisplayHeight+m_VerticalBorders,NumPixels);
	NumPixels = max(m_StartY+m_BottomMostPixel-m_DisplayHeight,NumPixels);
	NumPixels = 2*(NumPixels/2);

	m_LastVMove = NumPixels;
//	if(NumPixels >= m_SubPicHeight || -NumPixels >= m_SubPicHeight)
//		return -1;

	if(NumPixels > 0)
	{
		// Moving subpic up.
		// This is tricky. If m_StartY isn't at the top of the picture (say 1, for example),
		// then we can move the subpic simply by adjusting m_StartY and m_EndY without changing
		// the pointers. We can move that way by NumPix1. To move further, we have to change
		// the pointers. We move that way by NumPix2.
//		int NumPix1 = 2*((m_StartY - 1)/2);	// Must be even!
		int NumPix1 = 2*((m_StartY)/2);	// Must be even!
		NumPix1 = min(NumPix1,NumPixels);
		int NumPix2 = NumPixels-NumPix1;
		NumPix2 = min(NumPix2,m_SubPicHeight-3);	// so m_EndY >= m_StartY+2!
		_ASSERT(NumPix2+1 < m_SubPicHeight);
		// Moving subpics up. Set the offset to the top and bottom fields to skip the top
		// lines, then adjust the endy value...
		// First, find offset to NumPixels line.
		_ASSERT(NumPix2 >= 0);
		m_TopFieldOffset = m_StartOfLine[NumPix2];
		m_BotFieldOffset = m_StartOfLine[NumPix2+1];
		// Adjust EndY.
		_ASSERT(m_EndY > NumPix2+NumPix1);
		m_EndY -= (NumPix2+NumPix1);
		m_StartY -= NumPix1;
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		m_SubPicHeight = m_EndY - m_StartY + 1;
	}
	else
	{
		// Moving subpics down. Simply adjust the starty value... 
		// Adjust StartY and EndY, the hope is that if EndY is too large, the subpic will simply
		// be cropped.
		m_StartY += (-NumPixels);
		m_EndY += (-NumPixels);
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		m_SubPicHeight = min(m_EndY,m_DisplayHeight) - m_StartY + 1;
	}
	m_Modified = 1;
	return 0;
}

int CSubPic::VMoveSubPicInFrame(int MinTopMostPos, int FlushIt)
{
	// Avoid moving subpics that are actually highlights.
	if(m_HasButtons)
		return 0;
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	int BottomMostPixel = m_TrueBottomLine;
	// Bottom margin is the same as the top one.
	int MaxBottomMostPos = m_DisplayHeight - MinTopMostPos;
	m_VerticalBorders = MinTopMostPos;

	// m_StartY does not always indicate the current amount by which the subpics were
	// moved (for example if they were moved up by modifying the pointers).
	// Instead, m_StartYOrig - m_LastVMove tells me the true start Y
	// position, after the last move.
	long CurrentStartY = m_StartYOrig - m_LastVMove;

	// Make sure the subpic is either above our bottom limit or below our top one. If it's violating
	// both limits we can't do anything!
	_ASSERT(CurrentStartY+m_TopMostPixel >= MinTopMostPos || 
		CurrentStartY+BottomMostPixel <= MaxBottomMostPos);

	int MoveBy = 0;
	// Check whether our subpic is below MinTopMostPos.
	if(CurrentStartY+m_TopMostPixel < MinTopMostPos)
	{
		// We need to move this guy down! MoveBy will be negative.
		MoveBy = CurrentStartY+m_TopMostPixel - MinTopMostPos;
		// Check that by moving this guy down, we're not going past the bottom limit.
		if(CurrentStartY+BottomMostPixel-MoveBy > MaxBottomMostPos)
		{
			ERROR2("Can't apply border to subpic %d because it is too high!",m_SPUNumber);
			_ASSERT(0);
			return 0;
		}
	}
	else if(CurrentStartY+BottomMostPixel > MaxBottomMostPos)
	{
		// We need to move this guys up! MoveBy will be positive
		MoveBy = CurrentStartY+BottomMostPixel - MaxBottomMostPos;
		// Check that by moving this guy up, we're not going past the toplimit.
		if(CurrentStartY+m_TopMostPixel-MoveBy < MinTopMostPos)
		{
			ERROR2("Can't apply border to subpic %d because it is too high!",m_SPUNumber);
			_ASSERT(0);
			return 0;
		}
	}
	else if(FlushIt)
	{
		// The subpic is somewhere in the middle but we want to flush it!
		if(CurrentStartY + m_TopMostPixel - MinTopMostPos < MaxBottomMostPos - (CurrentStartY+BottomMostPixel))
			// We're closer to the top, move up by a large amount!
			MoveBy = 800;
		else
			// We're closer to the bottom, move down by a large amount.
			MoveBy = -800;
	}
	VMoveSubPic(m_LastVMove + MoveBy);
	return m_LastVMove + MoveBy;
}

int CSubPic::HMoveSubPic(int NumPixels)
{
	// Reset previous move.
	m_StartX = m_StartXOrig;
	m_EndX = m_EndXOrig;

	NumPixels = max(NumPixels,-m_StartXOrig);
	NumPixels = min(NumPixels,m_DisplayWidth-m_EndXOrig-1);

	m_StartX += NumPixels;
	m_EndX += NumPixels;
	m_LastHMove = NumPixels;
	WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
	if(NumPixels != 0)
		m_Modified = 1;
	return 0;
}

bool CSubPic::CanHMove()
{
	return (m_SubPicWidth < m_DisplayWidth - 10);	
}


int CSubPic::ChangeTransparency(int Transparency, int PixelSelect)
{
	Commands *TheCommand;
	// Only change the transparency if this DCSQT has a transparency command.
	int Offset = FindCommand(SET_CONTR,m_SelectedDCSQT,&TheCommand);
	if(Offset == -1)
		return -1;
	uchar* Ptr = m_Data+Offset;

	if(PixelSelect < 4)
		m_AlphaPtr[PixelSelect] = Transparency;
	else if(PixelSelect == 4)
	{
		// Decode subpic if needed. We need that for m_BackgroundPix.
		if(!m_SubPicDecoded && DecodeSubpic() == -1)
				return -1;
		// Change the transparency of all pixels, except for the background pixels.
		if(m_BackgroundPix != 0) m_AlphaPtr[0] = Transparency; 
		if(m_BackgroundPix != 1) m_AlphaPtr[1] = Transparency; 
		if(m_BackgroundPix != 2) m_AlphaPtr[2] = Transparency; 
		if(m_BackgroundPix != 3) m_AlphaPtr[3] = Transparency; 
	}
	else
		m_AlphaPtr[0] = m_AlphaPtr[1] = m_AlphaPtr[2] = m_AlphaPtr[3] = Transparency;
	
	// Put that in our data!
	WriteColorAlpha(Ptr+1,m_AlphaPtr);
	m_Modified = true;
	return 0;
}


int CSubPic::GetMeanTransparency()
{
	int Average = 0;
	if(m_BackgroundPix != 0) Average += m_AlphaPtr[0];
	if(m_BackgroundPix != 1) Average += m_AlphaPtr[1];
	if(m_BackgroundPix != 2) Average += m_AlphaPtr[2];
	if(m_BackgroundPix != 3) Average += m_AlphaPtr[3];
	return (int)((Average+.5f)/3);
}

int CSubPic::ChangeColor(int Color, int PixelSelect)
{
	Commands *TheCommand;
	// Only change the color if this DCSQT has a color command.
	int Offset = FindCommand(SET_COLOR,m_SelectedDCSQT,&TheCommand);
	if(Offset == -1)
		return -1;
	uchar* Ptr = m_Data+Offset;

	m_ColorPtr[PixelSelect] = Color;
	// Put that in our data!
	WriteColorAlpha(Ptr+1,m_ColorPtr);
	m_Modified = true;
	return 0;
}

int CSubPic::GetCommand(char* Command, int &DCSQT, int &CommandIndex)
{
	if(DCSQT >= m_DCSQT.size())
		return -1;
	SP_DCSQT *TheDCSQT = &m_DCSQT[DCSQT];
	if(CommandIndex == -1)
	{
		if(DCSQT == 0)
			sprintf(Command,"DCSQT: %d\n",DCSQT);
		else
			sprintf(Command,"DCSQT: %d -- Delay %.3fs\n",DCSQT,
			(float)TheDCSQT->m_Delay*1024/90000);
		goto end;
	}
	
	Commands TheCommand = TheDCSQT->m_Commands[CommandIndex];
	switch(TheCommand.m_CommandID)
	{
	case FSTA_DSP : sprintf(Command,"  Forced Start\n"); break;
	case STA_DSP : sprintf(Command,"  Start\n"); break;
	case STP_DSP : sprintf(Command,"  Stop\n"); break;
	case SET_COLOR : sprintf(Command,"  Set Color: e2=%d e1=%d p=%d b=%d\n",
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_CONTR : sprintf(Command,"  Set Contrast: e2=%d e1=%d p=%d b=%d\n",
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_DAREA : sprintf(Command,"  Set Display Area: sx=%d ex=%d sy=%d ey=%d \n", 
						 TheCommand.m_Params[3],TheCommand.m_Params[2],TheCommand.m_Params[1],TheCommand.m_Params[0]); break;
	case SET_DSPXA : sprintf(Command,"  Set field indexes\n"); break;
	case CHG_COLCON : 
		{
			// JEAN: TOTO: FIX THIS.
			int i,j;
			// Search our array for the CHG_COLCON command that's in this DCSQT.
			for(i=0;i<m_CHCOL.size();i++)
				if(m_CHCOL[i].m_DCSQT == DCSQT) break;
			// There SHOULD be one!
			_ASSERT(i != m_CHCOL.size());
			if(i == m_CHCOL.size())
				break;
			CHG_COLCON_Command *TheCom = &m_CHCOL[i];
			int ret;
			ret = sprintf(Command,"  CHG_COLCON\n\\par  Number of bands %d\n\\par ",TheCom->m_NumBands);
			for(i=0;i<TheCom->m_NumBands;i++)
			{
				// Go through all the bands, and for each one, display the relevant info.
				ret += sprintf(Command+ret,"   %d - Number of columns %d\n\\par ",i,TheCom->m_Bands[i].m_NumPX_CTLI);
				ret += sprintf(Command+ret,"    Top Line %d\n\\par ",TheCom->m_Bands[i].m_CSLN);
				ret += sprintf(Command+ret,"    Bot. Line %d\n\\par ",TheCom->m_Bands[i].m_CTLN);
				for(j=0;j<TheCom->m_Bands[i].m_NumPX_CTLI;j++)
				{
					// Same thing for all the defined columns..
					PX_CTLI* TheColumn = &TheCom->m_Bands[i].m_Columns[j];
					ret += sprintf(Command+ret,"    %d -- Start column %d\n\\par ",j,TheColumn->m_StartColNum);
					ret += sprintf(Command+ret,"     New colors: %d %d %d %d\n\\par ",
						TheColumn->m_NewColors[0],TheColumn->m_NewColors[1],
						TheColumn->m_NewColors[2],TheColumn->m_NewColors[3]);
					ret += sprintf(Command+ret,"     New alphas: %d %d %d %d\n\\par ",
						TheColumn->m_NewAlphas[0],TheColumn->m_NewAlphas[1],
						TheColumn->m_NewAlphas[2],TheColumn->m_NewAlphas[3]);
				}
			}
			break;
		}
	case CMD_END : sprintf(Command,"  End of commands\n"); break;
	}

end:
	CommandIndex++;
	if(CommandIndex >= TheDCSQT->m_Commands.size())
	{
		CommandIndex = -1;
		if(DCSQT < m_DCSQT.size()-1) DCSQT++;
		else
		{
			DCSQT = 0; return -1;
		}
	}
	return 0;
}

void CSubPic::ChangePTS(int NumFrames)
{
	// Just an experiment! Should not be used as they subpics probably won't appear in a player.
	int DeltaPTS = NumFrames * ((VideoFormat == FORMAT_PAL) ? 3600 : 3003);
	DeltaPTS = max(-(int)m_PTS,DeltaPTS);
	m_PTS += DeltaPTS;
	// Write this tack into our little "start of pack" buffer.
	BuildPTS(m_PackStartData+0x17, m_PTS, 0x20);
	m_Modified = true;
}

void CSubPic::SetPTS(int PTS)
{
	m_PTS = PTS;
	// Write this tack into our little "start of pack" buffer.
	BuildPTS(m_PackStartData+0x17, m_PTS, 0x20);
}

