#ifndef SubPic_H
#define SubPic_H

#include "includes.h"
// STL stuff for our arrays.
#include <vector>

class CCell;
class SP_DCSQT;

// Command enum
typedef enum
{
	FSTA_DSP = 0,
	STA_DSP = 1,
	STP_DSP  = 2,
	SET_COLOR = 3,
	SET_CONTR = 4,
	SET_DAREA = 5,
	SET_DSPXA = 6,
	CHG_COLCON = 7,
	CMD_END = 0xFF,
} CommandID;

// This is for the dreaded (and rare) CHG_COLCON command.
typedef struct
{
	ushort	m_StartColNum;	// Starting column number.
	ushort	m_NewColors[4];	// New color values.
	ushort	m_NewAlphas[4];	// New contrast values.
} PX_CTLI;

typedef struct
{
	ushort	m_CSLN;			// Starting (top-most) line number for this area
	uchar	m_NumPX_CTLI;	// number of PX_CTLI to follow
	ushort	m_CTLN;			// Terminating (bottom-most) line number for this area
	PX_CTLI	m_Columns[8];
} LN_CTLI;
typedef std::vector<LN_CTLI> LN_CTLI_V;

typedef struct
{
	int			m_DCSQT;	// The DCSQT index this command appears in.
	uint		m_NumBands;	// Number of bands.
	LN_CTLI_V	m_Bands;	// Data for each band.
} CHG_COLCON_Command;
typedef std::vector<CHG_COLCON_Command> CHG_COLCON_Com_V;


// Structure that contains the info about 1 command.
typedef struct
{
	CommandID	m_CommandID;	// Command ID.
	ushort		m_Offset;		// Offset to this command in m_Data.
	short		m_NumParams;	// How many parameters.
	ushort		m_Params[4];	// Command Parameters.
} Commands;
typedef std::vector<Commands> CommandsV;

// Class containing all the info about 1 CDSQT (group of commands)
#define MAX_NUM_COMMANDS 16
class SP_DCSQT
{
public:
	ushort		m_Delay;		// Delay befor execution
	ushort		m_OffsetToThis;	// Offset to start of this CDSQT
	ushort		m_OffsetToNext;	// Offset to next CDSQT. 0 if none
	CommandsV	m_Commands;		// The commands;
};

typedef std::vector<SP_DCSQT> DCSQTVect;


class CSubPic
{
public: 
	CSubPic();
	~CSubPic();

	void	Reset(bool ResetAll=true);

	// We look starting at PackIndex. If PackIndex = -1, we read the next SPU.
	int		WriteSubPacks(FILE* fd);
	// This decodes the subpic image.
	int		DecodeSubpic();
	// This outputs the subpic image as a text file, 1 index into the CLUT per pixel.
	int		OutputPicToTxtFile(const char* FileName);
	// This outputs the subpic image as a ppm file.
	int		SavePicAsPPM(const char* FileName, uchar *CLUT = 0, float FactX=1, float FactY=1,
		uchar* DefColors=0);
	// This stores the subpic image as a bitmap, rescaling at the same time. Width and Height on return
	// have the size of the scaled subpic.
	uchar*	SaveToBitmap(uchar* Bitmap, long &Width, long &Height, float FactX, float FactY, uchar *CLUT = 0, uchar* DefColors=0);
	// This one saves the subpic as an alpha (32 bit per pixel) bitmap.
	uchar*	SaveToAlphaBitmap(uchar* Bitmap, uchar *CLUT = 0, uchar *DefColors= 0);
	// This one mixes the subpic data into Bitmap, taking transparency into account.
	uchar*	MixIntoBitmap(uchar* Bitmap, long Width, long Height, uchar *CLUT = 0, uchar* DefColors=0);

	// This reads the pack in Data read from file fd at pack LBA (relative to start of file).
	// return -1 if something went wrong, 0 if it's still missing data, 1 when SPU is complete.
	// NavPackLBA is the LBA of the previous navpack, relative to start of VTS (not start of file).
	// CLBA is the cumulative (absolute) LBA.
	int		ReadThisPack(uchar* Data, FILE* fd, int LBA, int NavPackLBA, 
		ushort VOBID, uchar CELLID, uchar StreamID, int CLBA);

	// This saves the entire SPU (all its subpic packs) to their original locations.
	int		SaveSPU();

	// This reloads the entire SPU from file.
	int		ReloadSPU();

	// Retrieves the command in a human-readable form! Increments CommandIndex and DCSQT,
	// return -1 if no more...
	int		GetCommand(char* Command, int &DCSQT, int &CommandIndex);

	// This can move the subpics up or down by NumPixels, which should be even. Returns 0 if
	// it could, or the number of pixels it was able to move the subpics by. Positive numbers
	// to move up, negative to move down.
	int		VMoveSubPic(int NumPixels);

	// Here, the subpic is moved so the topmost pixel is no higher than MinTopMostPos and the bottommost
	// pixel is more than MinTopMostPos above the bottom of the picture. 
	// This is useful to move subpics inside a box, if needed. If FlushIt is non zero, the subpic is moved
	// flush with the nearest border, away from the center.
	int		VMoveSubPicInFrame(int MinTopMostPos, int FlushIt = 0);

	// This moves the subpic horizontally, but only if it can be (i.e. its background does not occupy
	// the full screen).
	int		HMoveSubPic(int NumPixels);
	// This can be called to see if the subpic can be moved horizontally.
	bool	CanHMove();

	// This changes the transparency of the selected pixel type. Transparency should be between 0 and 15.
	// PixelSelect can be 0,1,2,3 (for b,p,e1,e2) or 4 for all but background. 5 for truly all.. 
	// Returns 0 is OK, -1 if it couldn't.
	int		ChangeTransparency(int Transparency, int PixelSelect);
	// This returns the average transparency of non-background pixels...
	int		GetMeanTransparency();

	// This changes the color index of the selected pixel type. Returns 0 is OK, -1 if it couldn't.
	// Color should be between 0 and 15
	int		ChangeColor(int Color, int PixelSelect);

	// Find the command Command in the display control sequence table DCSQT. Return the
	// byte offset in m_Data to the Command itself. -1 if not found.
	int		FindCommand(CommandID Command, int DCSQT = 0, Commands** TheCommand = 0);

	// This replaces a command by another one. You need to know what you're doing here! 
	int		ChangeCommand(CommandID OldCommand, CommandID NewCommand, int DCSQT = 0);

	// This returns the timing delay (in milliseconds) of the given DCSQT
	int		GetDCSQTDelay(int DCSQT = -1);
	// Use this to determine the minimum and maximum delay for the last DCSQT.
	void	GetMinMaxDCSQTDelay(int &MinDurMs, int &MaxDurMS, int DCSQT = -1);
	// This changes the timing of the last DCSQT. Units are in ms.
	int		ChangeDCSQTDelay(int DelayMs, int DCSQT = -1);

	// Use this to select a specific DCSQT. The color and transparencies will be changed to the ones defined
	// in that DCSQT (if any).
	void	SelectDCSQT(int DCSQT);

	// This looks for Search in the OCR lines. Returns true if found. Search is not case sensitive
	bool	FindString(const char* Search);
	// This looks for the characters in Search in the OCR lines. Returns true if any is found. 
	bool	FindChars(const char* Search);
	// This determines if the subpic text if of the kind [asdfkjhasdf] or (asdflkjasdf)
	bool	IsTextBracketed();

	// This returns true if the SPU seems to contain button highlights instead of subpicture data.
	bool	ContainsHighlights();

	// Parse all commands in all DCSQT. Calls itself recursively.
	int		ParseCommands(int NumDCSQT = 0, int OffsetToDCSQT = 0);

	// This crops the background of the subpic, retaining just enough to enclose the "useful" part.
	void	CropBackground();

	int		m_CurrentBit;		// Next bit to be poped
	int		m_CurrentByte;		// Current byte we're popping from!
	int		m_PixelCount;		// Count of output m_pixels.
	int		m_CurrentPixelX;	// x-value for current output pixel.
	int		m_CurrentPixelY;	// y-value for current output pixel.
	int		m_SubPicWidth;		// Line width.
	int		m_ModSubPicWidth;	// For bitmaps, the width has to be a multiple of 4. This is it.
	int		m_SubPicHeight;		// Number of lines.

	uint	m_PTS;				// PTS for this spu.

	ushort	m_SubPicSize;		// The overall byte size of the subpic data, including commands etc.
	ushort	m_OffsetToCmd;		// Byte offset in m_Data to commands.

	ushort	m_StartXOrig;		// Start of subpic bounding rectangle in original subpic.
	ushort	m_StartYOrig;		// Start of subpic bounding rectangle in original subpic.
	ushort	m_EndXOrig;			// End of subpic bounding rectangle in original subpic.
	ushort	m_EndYOrig;			// End of subpic bounding rectangle in original subpic.
	ushort	m_StartX;			// Start of subpic bounding rectangle.
	ushort	m_StartY;			// Start of subpic bounding rectangle.
	ushort	m_EndX;				// End of subpic bounding rectangle.
	ushort	m_EndY;				// End of subpic bounding rectangle.
	ushort	m_TopFieldOffset;	// Offset to first pix data of top field in m_Data.
	ushort	m_BotFieldOffset;	// Offset to first pix data of bottom field in m_Data.
	// These will point to command parameter arrays in a DCSQT.
	ushort *m_ColorPtr;			// Array of color indexes for each pixel type b,p,e1,e2
	ushort *m_AlphaPtr;			// Array of alpha indexes for each pixel type b,p,e1,e2

	uchar*	m_CLUT;				// Just a pointer. Set it to a valid CLUT.
	
	ushort	m_NumPix[4];		// Count of how many of each pixel types were found in the subpic.

	int		m_BackgroundPix;	// Which of the four pixel types seems to be the background.
	int		m_BodyPix;			// Which of the four pixel types is used for the body of the characters.
	uchar	(*m_OCRLines)[128];	// Will contain the results of the OCR.

	static	int m_DisplayHeight;	// Vertical size of video display.
	static	int m_DisplayWidth;		// Horizontal size of video display.

	int		m_LastVMove;		// Store the amount of vertical moving last applied.
	int		m_VerticalBorders;	// Minimum space between top and bottom of pic, and subpic data.
	int		m_LastHMove;		// Store the amount of horizontal moving last applied.

	uchar	*m_Data;			// Binary data, including pixel data and commands.
static	uchar	*m_Pixels;		// Decoded pixel data (values between 0 and 15, index into KLUT).
	uchar	m_PackStartData[64];// 64 bytes of the beginning of the first pack for this SPU.
	// For each line, m_LastSetPixel indicates the last pixel whose value is set individually, as
	// opposed to "until end of line".
	ushort	*m_LastSetPixel;
	ushort	*m_StartOfLine;		// Index in m_Data of data for first pixel for this line.
	ushort	m_RightMostPixel;	// Rightmost non-background pixel;
	ushort	m_LeftMostPixel;	// Leftmost non-background pixel;
	short	m_TopMostPixel;		// First line with a pixel that's defined individually.
	short	m_BottomMostPixel;	// Last line with a pixel that's defined individually.
	short	m_TrueBottomLine;	// Last line representing the bottom of the sub (excluding the p,q,etc).
	short	m_BottomMostPixelT;	// A temporary value... 

	DCSQTVect	m_DCSQT;		// Vector of DCSQTs (sets of commands).
	int		m_SelectedDCSQT;	// Index of DCSQT currently selected for display and changes.
	CHG_COLCON_Com_V	m_CHCOL;	// A vector of CHG_COLCON commands.

	int		m_Modified;			// Indicates this SPU's data is different than what's on disk.
	ushort	m_VOBID;			// These indicate the VCID this subpic is in.
	uchar	m_CELLID;			
	uchar	m_StreamID;			// The stream ID for this spu.
	int		m_HasButtons;		// Will be non-zero if the neighboring navpack indicates buttons.

	int		m_NumPackInSPU;		// How many subpic packs form this SPU.
	FILE*	m_FilesFrom[32];	// Files the various packs that form this SPU come from.
	int		m_LBAFrom[32];		// Where in above file were the subpic packs located.
	int		m_CLBAFrom[32];		// Cumulative LBAs for the subpic packs.
	int		m_NavPackLBA;		// LBA of preceding nav pack, relative to start of VTS.
	int		m_SPUNumber;		// A number associated with this SPU.
	CSubPic*	m_NextSPUInStream;// Pointer to next spu in the same stream.

	bool	m_SPUReadOK;		// Indicates whether Subpic unit was read correctly.
	bool	m_SubPicDecoded;	// Indicates whether Subpic unit was decoded.

	// DO NOT USE THIS! JUST AN EXPERIMENT!
	void	ChangePTS(int NumFrames);
	void	SetPTS(int PTS);
	// Takes the pixel array and encodes the specified field.
	void	RLEEncode(int Field, int StartCol = 0, int EndCol = 10000, int StartLine = 0, int EndLine = 10000);	
protected:

	// Helper functions for decoding etc...
	uchar	pop2();				// Pops 2 bits from m_Data.
	uchar	pop4();				// Pops 4 bits from m_Data.
	uchar	pop8();				// Pops 8 bits from m_Data.
	int		Out(int n, int c);	// Outputs pixel data into m_Pixels;
	void	RLEDecode(int Field);	// Decodes 1 field of the subpic data.
	// Helper functions for encoding etc...
	void	push4(ushort Code);
	void	push8(ushort Code);
	void	push12(ushort Code);
	void	push16(ushort Code);
	void	push(int count, uchar c);	// Encode character c count times.

	// This find the bottom line, excluding chars like q,p,g. This is useful for moving subpics above a 
	// line, letting the part of characters such as 'q' 'p' 'g' have their bottom queues go below the line.
	int		FindTrueBottomLine();	
	// This find which pixel is used for the body of the characters.
	void	FindBodyPixel();

	// Helper function.
	int		DoWriteSubPack(FILE* fd);

	// Helper function to parse the complicated CHG_COLCON command.
	int		ParseCHG_COLCON(uchar** Ptr, CHG_COLCON_Command* Command);

	// Helper functions to write the DAERA and DSPXA parameters.
	void	WriteDAREA(ushort StartX, ushort StartY, ushort EndX, ushort EndY);
	void	WriteDSPXA(ushort TopOffset, ushort BotOffset);

	int		m_EndOfField;		// End of data area for the current field being decoded.
	uint	m_DataBytesRead;	// How many data bytes have been read so far.
	uint	m_DataBytesWritten;	// How many data bytes have been written so far.



};








#endif