/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

void DefaultLogMessage(char* AMessage, int Langage);
#include "SubPicEngine.h"
#include <stdio.h>
#include <objbase.h>
#include "Cell.h"
#include "Utils.h"
#include "SubPic.h"
#include "IFOUtils.h"
#include <algorithm>

#ifndef _ASSERT
#include <crtdbg.h>
#endif

typedef unsigned char uchar;

// Only a small part of the navpacks is kept in memory (to minimize memory usage).
#define NAV_MEM_START		0x04f1				// Offset of first byte of navpack kept in mem.
#define NAV_MEM_SIZE		(0x0600-0x04f1)		// How many bytes we're copying to mem.

#define VERS "DVDSubEdit -- Version 1.0.2 By Jeanl"

#define CHAR_SIZE 256
// This stuff is global and accessed by all the classes, for messages and errors.
// The string where messages are stored.
char Message[CHAR_SIZE];
// Our default function to handle regular and error messages.
void DefaultLogMessage(char* AMessage, int Langage)
{
	printf(AMessage);
	strcpy(Message,AMessage);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}
// Global pointers that classes use to send regular and error messages.
void (*pMessageFunction)(char* AMessage, int Langage) = DefaultLogMessage;
void (*pErrorFunction)(char* AMessage, int Langage) = DefaultLogMessage;
int	 (*FrameSelectFunction)(int VobId, int CellId, int StartLBA, int EndLBA, int Default, 
							int NumLBA, int NormallySkipped) = 0;

// I need this destructor so the CLUT arrays are automatically destroyed.
PGC_CLUT_Info::~PGC_CLUT_Info()
{
}

VobuInfo::VobuInfo()
{
	m_NavPackLBA = 0;
	m_OrigNavPackLBA = 0;
	m_PTSFirstFrame = 0;	
	m_EmptySlotCount = 0;	
	m_SizeLBA = 0;
	m_NavPtr = 0;
}



CSubPicEngine::CSubPicEngine()
{
	strcpy(m_Version,VERS);
	int i;
	for(i=0;i<10;i++)
		m_FileInfos[i].m_FileFD = 0;
	m_InputLBA = 0;	
	m_OutputLBA = 0;
	m_MemoryPtr = 0;
	Reset();
}
	
CSubPicEngine::~CSubPicEngine()
{
	Reset();
	int i;
	for(i=0;i<32;i++)
	{
		m_SubPics[i].clear();
	}
}

void CSubPicEngine::Reset()
{
	int i;
	m_AbortNow = 0;
	m_TotNumSPU = 0;
	m_PruneLength = 0;
	m_PrunedList.clear();
	m_Phase = Processing_Idle;
	m_NumVOBFiles = 0;
	m_TopStreamID = 0;
	m_SelectedSubPic = 0;
	m_PTSPerFrame = 3003;
	m_CanMove = 0;
	m_PercentDone = 0;
	m_PruneLoadWait = 0;
	m_FoundILVU = 0;
	m_ListHasOCR = 0;
	m_CanSave = true;
	m_CanReload = true;
	m_SupFileMode = false;
	m_NoVideo = true;

	m_LeftLimitLBA = 0;	
	m_RightLimitLBA = 0;
	m_MoveInc = 0;
	delete[] m_MemoryPtr;
	m_MemoryPtr = 0;
	m_MemNeedsWrite = 0;
	m_MemStartLBA = 0;
	m_MemSizeLBA = 0;
	m_MemAllocSize = 0;
	m_MemOrigVOBU = 0;
	m_SrcLBA = 0;	
	m_DestLBABM = 0;
	m_DestLBAAM = 0;

	m_VCIDLength = 0;
	m_AspectRatio = 3;
	memset(m_NumSubPics,0,32*sizeof(int));
	for(i=0;i<32;i++)
	{
		int j;
		for(j=0;j<m_SubPics[i].size();j++)
			delete m_SubPics[i][j];
		m_SubPics[i].clear();
		m_SubPics[i].push_back(0);
		m_StreamCanMove[i] = 1;
	}
	m_MasterList.clear();

	for(i=0;i<m_VobuInfo.size();i++)
		delete[] m_VobuInfo[i].m_NavPtr;
	m_VobuInfo.clear();

	for(i=0;i<m_AllCLUTS.size();i++)
		delete[] m_AllCLUTS[i];
	m_AllCLUTS.clear();
	
	for(i=0;i<10;i++)
	{
		if(m_FileInfos[i].m_FileFD)
			fclose(m_FileInfos[i].m_FileFD);
		m_FileInfos[i].m_FileFD = 0;
	}
	delete[] m_InputLBA; m_InputLBA = 0;	
	delete[] m_OutputLBA; m_OutputLBA = 0;
	for(i=0;i<m_CLUTV.size();i++)
		delete m_CLUTV[i];
	m_CLUTV.clear();
}	

void CSubPicEngine::ScanFilesForSPUs()
{
	// Scan all the vob files for subpics...
	int i,LBA,ret;
	int TotNumLBA = 0;
	int CumulLBA = 0;
	int NavPackNumButtons = 0;
	DWORD End;

	CompositeSRC LastNavPackSRC ;
	LastNavPackSRC.m_SRCInt = 0;
	LastNavPackSRC.m_SRCFrac= 0;

	for(i=0;i<m_NumVOBFiles;i++)
	{
		m_FileInfos[i].m_StartLBA = TotNumLBA;
		m_FileInfos[i].m_SizeLBA = GetFileSizekB(m_FileInfos[i].m_FileName)/2;
		TotNumLBA += m_FileInfos[i].m_SizeLBA;
		fseek(m_FileInfos[i].m_FileFD,0,SEEK_SET);
	}
	// Check whether we have the full domain.
	m_CanMove = false;
	m_SupFileMode = false;
	m_NoVideo = false;
	if(m_NumVOBFiles > 0)
	{
		int Index = -1;
		int VTS = -1;
		// Look for vts_xx_y.vob. Check whether we do have vts_xx_1.vob
		if(sscanf(m_FileInfos[0].m_FileName,"%*[^_]_%d_%d",&VTS,&Index) == 2 && Index == 1)
		{
			// We do! Let's check that we didn't skip the last file of the VTS... 
			char Temp[512];
			strcpy(Temp,m_FileInfos[m_NumVOBFiles-1].m_FileName);
			// Add one to the index and check whether this file exists.
			Temp[strlen(Temp)-5]++;
			// If the file does not exist, we have the full domain, and we can move.
			if(!PathExists(Temp))
				m_CanMove = true;
		}
	}

	// Start by scanning the IFOs.
	ScanIFOs();

	VobuInfo TheVobuInfo;
	CPack ThePack;
	int LastNavPack = -1;
	m_TotNumSPU = 0;
	m_MasterList.clear();
	DWORD Start = timeGetTime();
	m_Phase = Processing_Scanning_VOB;
	float SmoothRate = 0;
	VideoFormat = FORMAT_UNKNOWN;
	for(i=0;i<m_NumVOBFiles;i++)
	{
		for(LBA=0;;LBA++,CumulLBA++)
		{
			if(m_AbortNow)
			{
				m_Phase = Processing_Aborted;
				m_CanMove = false;
				goto end ;
			}
			m_PercentDone = 100*CumulLBA / TotNumLBA;
			if((CumulLBA+1) % 2000 == 0)
			{
				// This is just to report progress and estimated end time...
				DWORD End = timeGetTime();
				float RateMBpS = (1000*(float)CumulLBA*2/1024/(End-Start));
				if(SmoothRate == 0) SmoothRate = RateMBpS;
				else SmoothRate = .95f*SmoothRate + .05f*RateMBpS;
				m_RemainingTimeSec = (TotNumLBA - CumulLBA)*2/1024 / SmoothRate;
				MESSAGE5("%.2f MB scanned (%.2f MB/s) -- %d subpics -- %.0f secs to go",(float)CumulLBA/2/1024,
					RateMBpS,m_TotNumSPU, m_RemainingTimeSec);
			}
			// Read next pack.
			ret=(fread(m_Buffer,1,SECTOR_SIZE,m_FileInfos[i].m_FileFD) == SECTOR_SIZE);
			if(ret != 1)
				break;
			ushort VOBID,vobid;
			uchar CELLID,cellid;
			ThePack.FindPackType(m_Buffer, vobid, cellid);
			
			if(ThePack.m_PackType != PackTypeSupPic)
			{
				if(ThePack.m_PackType == PackTypeNavPack)
				{
					// Pack is a navpack. Keep track of the VCID.
					VOBID = vobid; CELLID = cellid;
					LastNavPack = CumulLBA;
					// The spu will need to know whether there are buttons.
					NavPackNumButtons = ThePack.m_NumButtons;
					// Detect ILVUs...
					m_FoundILVU |= (GetShort(m_Buffer,0x427) != 0);
					// Our info on vobus... If LBA isn't 0, now's the time to push the last
					// vobu info.
					CompositeSRC ThisSRC ;
					SrcToFrac(ThePack.m_SRCL, ThePack.m_SRCH, ThisSRC);
					if(m_VobuInfo.size())
					{
						// Count how many empty slots we have. An empty slot lasts 146.86 SRC tics.
						// Measure how many slots in the VOBU, remove the number of packs.
						m_VobuInfo.back().m_EmptySlotCount = ((150*(ThisSRC.m_SRCInt - LastNavPackSRC.m_SRCInt))/21943)
							- (CumulLBA - m_VobuInfo.back().m_NavPackLBA);
						if(m_VobuInfo.back().m_EmptySlotCount < 1)
							MESSAGE2("VOBU at LBA %d has no free slot",m_VobuInfo.back().m_NavPackLBA);
						m_VobuInfo.back().m_SizeLBA = CumulLBA - m_VobuInfo.back().m_NavPackLBA;
// Checking for discontinous PTS...
if(m_VobuInfo.back().m_PTSFirstFrame >= GetInt(m_Buffer,0x39))
	MESSAGE2("PTS discontinuous at LBA %d",CumulLBA);
					}
					// Set this.
					TheVobuInfo.m_NavPtr = new uchar[NAV_MEM_SIZE];
					memcpy(TheVobuInfo.m_NavPtr,m_Buffer+NAV_MEM_START,NAV_MEM_SIZE);
					TheVobuInfo.m_NavPackLBA = TheVobuInfo.m_OrigNavPackLBA = CumulLBA;
					// Retrieve the VOBU start presentation time.
					TheVobuInfo.m_PTSFirstFrame = GetInt(m_Buffer,0x39);
					// Put this guy into our array.
					m_VobuInfo.push_back(TheVobuInfo);
					LastNavPackSRC = ThisSRC;
				}
				else if(ThePack.m_PackType == PackTypeVideo && VideoFormat == FORMAT_UNKNOWN)
				{
					// The first video pack is scanned for the format. Maybe I should get it
					// from the nav pack.
					ThePack.FindFrameType(m_Buffer,0,&m_AspectRatio);
					if(VideoFormat != FORMAT_UNKNOWN)
					{
						CSubPic::m_DisplayHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
						m_PTSPerFrame = (VideoFormat == FORMAT_PAL) ? 3600 : 3003;
						JustFoundVideoFormat();
					}
				}
				continue;
			}
			// This is a subpic pack. Find the stream ID.
			uchar StreamID = ThePack.m_StreamID-0x20;
			// Temp might be 0 if this is the first pack of a new SPU.
			CSubPic* Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]];
			m_TopStreamID = max(m_TopStreamID,(int)StreamID);
			if(Temp == NULL)
			{
				// We just found a new SPU. Allocate the subpic instance, put that in our vector.
				Temp = new CSubPic;
				m_SubPics[StreamID][m_NumSubPics[StreamID]] = Temp;
			}
			Temp->m_HasButtons = max(NavPackNumButtons,Temp->m_HasButtons);
			// Now let the CSubPic class read the pack. It will return 1 if this is the last pack of
			// the spu.
			if(Temp->ReadThisPack(m_Buffer, m_FileInfos[i].m_FileFD, LBA, LastNavPack, VOBID, 
				CELLID,StreamID,CumulLBA) == 1)
			{
				if(Temp->m_NavPackLBA == -1)
				{
					// This can happen if there hasn't been a navpack before this subpic, for example
					// because we started scanning in the middle of a VTS. At this point, I'm simply
					// going to discard this guy!
					delete Temp;
					Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]] = 0;
				}
				else
				{
					// We finished reading the last pack for this SPU. This function does various required things.
					AfterLastPack(Temp,StreamID);
				}
			}
		}
	}
	if(CumulLBA)
	{
		m_VobuInfo.back().m_SizeLBA = CumulLBA - m_VobuInfo.back().m_NavPackLBA;
		m_VobuInfo.back().m_EmptySlotCount = 0;
	}
	m_InputLBA = new int[m_VobuInfo.size()+1];
	m_OutputLBA = new int[m_VobuInfo.size()+1];
	for(i=0;i<m_VobuInfo.size();i++)
		m_InputLBA[i] = m_OutputLBA[i] = m_VobuInfo[i].m_NavPackLBA;
	m_InputLBA[i] = m_OutputLBA[i] = CumulLBA;

	// A bit of cleanup.
	for(i=0;i<32;i++)
	{
		if(m_SubPics[i].size() && m_SubPics[i].back() == 0)
			m_SubPics[i].pop_back();
	}
	if(m_MasterList.size() && m_MasterList.back() == 0)
		m_MasterList.pop_back();

	End = timeGetTime();
	MESSAGE4("%.2f MB scanned (%.2f MB/s) -- %d subpics",(float)CumulLBA/2/1024,
		1000*(float)CumulLBA*2/1024/(End-Start),m_TotNumSPU);
	m_Phase = Processing_Done;

//	UpdateVobuSRI(false);
end:
	;
}

void CSubPicEngine::AfterLastPack(CSubPic* Temp, int StreamID)
{
	// The Last pack of our SPU was read. If this is the first SPU of this stream, notify ourselves.
	if(m_NumSubPics[StreamID] == 0)
		JustFoundStream(StreamID);
	// Wait until PruneSPUList() is done. Then set flag so PruneSPUList() won't step on us.
	while(m_PruneLoadWait)
		Sleep(1);
	m_PruneLoadWait = 1;
	// Here, I'm checking that two consecutive spus in the same stream don't fall in the same VOBU.
	if(m_StreamCanMove[StreamID] && m_NumSubPics[StreamID] >= 1 && 
		m_SubPics[StreamID].back()->m_NavPackLBA == m_SubPics[StreamID][m_NumSubPics[StreamID]-1]->m_NavPackLBA)
		m_StreamCanMove[StreamID] = 0;
	// Increment the number of subpics for this stream ID, and put a 0 one at the top.
	m_NumSubPics[StreamID]++;
	m_SubPics[StreamID].push_back(0);
	// Also increment the overall number of spus, and put this one on the master list.
	Temp->m_CLUT = FindCLUT(Temp->m_VOBID, Temp->m_CELLID);
	Temp->m_SPUNumber = m_TotNumSPU;
	m_MasterList.push_back(Temp);
	// Now find the spu just before this one in the same stream.
	for(int Prev = m_TotNumSPU-1;Prev>=0;Prev--)
	{
		if(m_MasterList[Prev]->m_StreamID == StreamID)
		{
			// Make its m_NextSPUInStream pointer point to the new guy.
			m_MasterList[Prev]->m_NextSPUInStream = Temp;
			break;
		}
	}
	m_PruneLoadWait = 0;
	m_TotNumSPU++;
	// If we haven't seen this VCID before, notify ourselves.
	// JEAN: FIX: When ILVUs, this does not work well.
	if(m_VCIDLength == 0 || 256*(uint)Temp->m_VOBID+Temp->m_CELLID != m_VCIDList[m_VCIDLength-1])
	{
		m_VCIDList[m_VCIDLength++] = 256*(uint)Temp->m_VOBID+Temp->m_CELLID;
		JustFoundVCID(256*(uint)Temp->m_VOBID+Temp->m_CELLID);
	}
	// Finally notify ourselves we added an spu.
	JustAddedSPU(StreamID,m_NumSubPics[StreamID]-1);
}


float CSubPicEngine::PTSToSecs(int PTS)
{
	return (float)PTS/90000;
}

int CSubPicEngine::SecsToPTS(float Secs)
{
	return m_PTSPerFrame*((int)(90000*Secs)/m_PTSPerFrame);
}

void CSubPicEngine::TestMove(int Index, float Seconds)
{
	// Use 3600 for PAL.
#if 0
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(DeltaPTS < 0)
		for(int i=0;i<m_TotNumSPU;i++)
			ChangeTiming(i,DeltaPTS);
	else
		for(int i=m_TotNumSPU-1;i>=0;i--)
			ChangeTiming(i,DeltaPTS);
	WritePacksToFiles();
	SortMasterLists();
#else
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	ChangeTiming(Index,DeltaPTS);
	WritePacksToFiles();
	SortMasterLists();
#endif
}

void CSubPicEngine::MoveSelection(float Seconds)
{
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(DeltaPTS < 0)
		for(int i=0;i<m_PruneLength;i++)
		{
			if(m_AbortNow) break;
			ChangeTiming(m_PrunedList[i]->m_SPUNumber,DeltaPTS);
			m_PercentDone = 100*i/m_PruneLength;
		}
	else
		for(int i=m_PruneLength-1;i>=0;i--)
		{
			if(m_AbortNow) break;
			ChangeTiming(m_PrunedList[i]->m_SPUNumber,DeltaPTS);
			m_PercentDone = 100*(m_PruneLength-i)/m_PruneLength;
		}
	WritePacksToFiles();
	SortMasterLists();
	UpdateVobuSRI();
	UpdateIFOAfterMove();
}

void CSubPicEngine::MoveThisOne(int Index, float Seconds)
{
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(ChangeTiming(Index,DeltaPTS) == 0)
	{
		WritePacksToFiles();
		SortMasterLists();
		UpdateVobuSRI();
		UpdateIFOAfterMove();
	}
}

void CSubPicEngine::ScanIFOs()
{
	int k;
	// Scan IFOs, looking for each VCID, and getting the CLUT for the PGCs which reference
	// them...
	// First, get IFO file.
	ushort SLU,LU,NumLU;
	ushort PGC,NumPGC;
	uchar CELL,NumCELL;
	CIFOLoader IFO(GetIFO(m_FileInfos[0].m_FileName));
	if(!IFO.m_LoadedOK)
		return;
	char Num = m_FileInfos[0].m_FileName[strlen(m_FileInfos[0].m_FileName)-5];
	if(Num > '0' && Num <= '9')
		SLU = 0, NumLU = 0;
	else
		SLU = 1, NumLU = IFO.GetNumLU();

	// Loop over all LUs, and all PGCs, and all Cells.
	PGC_CLUT_Info *CLUTInfo;
	for(LU=SLU;LU<=NumLU;LU++)
	{
		NumPGC = IFO.GetNumPGC(LU);
		for(PGC=1;PGC<=NumPGC;PGC++)
		{
			uchar *CurrentCLUT  = new uchar[3*16];
			IFO.GetCLUT(CurrentCLUT,LU,PGC,true);
			m_AllCLUTS.push_back(CurrentCLUT);
			NumCELL = IFO.GetNumCells(LU,PGC);
			for(CELL=1;CELL<=NumCELL;CELL++)
			{
				CELL_POS_INFO*	Info=IFO.GetCellPosInfo(LU, PGC, CELL); _ASSERT(Info);
				uint VCID = 256*(uint)FixEndian(Info->m_VOBID)+(Info->m_CELLID);
				// See if we already have this cell somewhere in our clut info.
				for(k=0;k<m_CLUTV.size();k++)
				{
					if(m_CLUTV[k]->m_VCID == VCID) break;
				}
				if(k == m_CLUTV.size())
				{
					// We don't have it. Add one item to our clut vector.
					CLUTInfo = new PGC_CLUT_Info;
					CLUTInfo->m_NumCLUT=0;
					CLUTInfo->m_VCID=VCID;
					m_CLUTV.push_back(CLUTInfo);
				}
				// Add this clut to the clut info.
				CLUTInfo = m_CLUTV.back();
				CLUTInfo->m_PGC.push_back(PGC);
				CLUTInfo->m_LU.push_back(LU);
				CLUTInfo->m_CLUT.push_back(CurrentCLUT);
				// Get the language string.
				if(CLUTInfo->m_NumCLUT == 0)
				{
					int m;
					for(m=0;m<32;m++)
						IFO.GetPgcSubpicLanguage(CLUTInfo->m_Languages+2*m,LU,PGC,0x20+m);
				}
				CLUTInfo->m_NumCLUT++;
			}
		}
	}
	
	IFO.CloseIFO();
}

int CSubPicEngine::OpenIFO(const char* FileName)
{
	// Don't load an IFO unless we're currently looking at a sup file.
	if(!m_SupFileMode)
		return -1;
	CIFOLoader IFO(FileName);
	if(!IFO.m_LoadedOK)
		return -1;	
	ushort NumPGC;
	NumPGC = IFO.GetNumPGC(0);
	if(NumPGC == 0)
		return -1;
	uchar *CurrentCLUT  = new uchar[3*16];
	// Get the clut, and assign it to all the subpics.
	IFO.GetCLUT(CurrentCLUT,0,1,true);
	m_AllCLUTS.push_back(CurrentCLUT);
	int i;
	for(i=0;i<m_TotNumSPU;i++)
	{
		m_MasterList[i]->m_CLUT = CurrentCLUT;
	}
	return 0;
}


uchar* CSubPicEngine::FindCLUT(ushort VOBID, uchar CELLID, char (*Language)[16])
{
	int i,j;
	uchar* Clut = 0;
	for(i=0;i<m_CLUTV.size();i++)
	{
		if(m_CLUTV[i]->m_VCID == (uint)VOBID*256+CELLID)
		{
			if(m_CLUTV[i]->m_NumCLUT > 0)
				Clut = m_CLUTV[i]->m_CLUT[0];
			if(Language)
			{
				for(j=0;j<32;j++)
				{
					sprintf(Language[j],"0x%x (%c%c)",j+32,m_CLUTV[i]->m_Languages[2*j],
						m_CLUTV[i]->m_Languages[2*j+1]);
				}
			}
			break;
		}
	}
	return Clut;
}

void CSubPicEngine::ReloadAllSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			Temp->ReloadSPU();
		}
	}
}

void CSubPicEngine::SaveAllModifiedSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			if(Temp->m_Modified)
				Temp->SaveSPU();
		}
	}
	for(i=0;i<m_NumVOBFiles;i++)
		fflush(m_FileInfos[i].m_FileFD);
}

void CSubPicEngine::CropAllSubpics()
{
	int i;
	for(i=0;i<m_PruneLength;i++)
	{
		m_PercentDone = i*100/m_PruneLength;
		if(m_AbortNow) break;
		m_PrunedList[i]->CropBackground();
	}
}


// Experimental code, just for testing.
int CSubPicEngine::LoadSUPFile(const char* SupFileName)
{
	Reset();
	FILE* fd = fopen(SupFileName,"rb");
	if(!fd)
		return -1;
	
	int i;
	for(i=0;;i++)
	{
		// Load "SP"
		char Head[2];
		if(fread(Head,1,2,fd) != 2) break;
		if(Head[0] != 'S' || Head[1] != 'P')
			break;
		// Now read PTS value;
		int PTS;
		if(fread(&PTS,4,1,fd) != 1) break;
		CSubPic *TheSPU = new CSubPic;
		TheSPU->m_PTS = PTS;
		TheSPU->m_StreamID = 0x20;
		if(fread(&PTS,4,1,fd) != 1) break;
		// Read data... First, Size and Offset to commands.
		ushort Size, Offset;
		if(fread(&Size,2,1,fd) != 1) break; 
		Size = FixEndian(Size);
		if(fread(&Offset,2,1,fd) != 1) break; 
		Offset = FixEndian(Offset);
		// Set a few member variables.
		TheSPU->m_SubPicSize = Size;
		TheSPU->m_OffsetToCmd = Offset;
		TheSPU->m_SPUReadOK = true;
		TheSPU->m_SPUNumber = i;
		// Allocate data, seek back to start of Size.
		TheSPU->m_Data = new uchar[Size];
		fseek(fd,-4,SEEK_CUR);
		// And read all the stuff in our bufffer.
		if(fread(TheSPU->m_Data,1,Size,fd) != Size) break;
		// Parse commands, etc.
		TheSPU->ParseCommands();
		TheSPU->SelectDCSQT(TheSPU->m_SelectedDCSQT);
		// Then add this guy to our lists. First, the stream list:
		m_SubPics[0].back() = TheSPU;
		m_NumSubPics[0]++;
		m_SubPics[0].push_back(0);
		// Then the master list.
		m_MasterList.push_back(TheSPU);
		m_TotNumSPU++;
	}
	m_TopStreamID = 0;
	// Faking this, since we don't have video data.
	VideoFormat = FORMAT_PAL;
	m_AspectRatio = 3;
	CSubPic::m_DisplayHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
	m_PTSPerFrame = (VideoFormat == FORMAT_PAL) ? 3600 : 3003;
	// Can't move and can't save, of course!
	m_CanMove = false;
	m_CanSave = false;
	m_CanReload = false;
	m_SupFileMode = true;
	m_NoVideo = true;
	// Faking faking.
	JustFoundVideoFormat();
	JustFoundStream(0);
	JustAddedSPU(0,m_NumSubPics[0]-1);
	// A bit of cleanup.
	for(i=0;i<32;i++)
	{
		if(m_SubPics[i].size() && m_SubPics[i].back() == 0)
			m_SubPics[i].pop_back();
	}
	if(m_MasterList.size() && m_MasterList.back() == 0)
		m_MasterList.pop_back();
	// This is so we can save as spu as well!
	VobuInfo AVobuInfo;
	m_VobuInfo.push_back(AVobuInfo);
	m_VobuInfo[0].m_PTSFirstFrame = 0;
	fclose(fd);
	return 0;
}

int CSubPicEngine::SaveSubpicInfo(const char* FileName)
{
	// Don't do anything if nothing was loaded.
	if(m_TotNumSPU == 0 || m_NumVOBFiles == 0)
		return -1;
	FILE* fd = fopen(FileName,"w");
	if(fd == 0)
		return -1;
	
	int i;
	// Put all the file info:
	fprintf(fd,"%d\n",m_NumVOBFiles);
	for(i=0;i<m_NumVOBFiles;i++)
		fprintf(fd,"\"%s\" %d\n",m_FileInfos[i].m_FileName,0);
	
	// Then write video format and aspect ratio and PTS of first VOBU
	fprintf(fd,"%d\n",VideoFormat);
	fprintf(fd,"%d\n",(int)m_AspectRatio);
	fprintf(fd,"%d\n",m_VobuInfo[0].m_PTSFirstFrame);
	// And go through all the spus.
	fprintf(fd,"%d\n",m_TotNumSPU);
	for(i=0;i<m_TotNumSPU;i++)
	{
		fprintf(fd,"%d %d\n",m_MasterList[i]->m_NumPackInSPU, m_MasterList[i]->m_NavPackLBA);
		int j;
		for(j=0;j<m_MasterList[i]->m_NumPackInSPU;j++)
		{
			// TODO: FIX: This is soooo ugly.
			int k;
			for(k=0;k<m_NumVOBFiles;k++)
				if(m_MasterList[i]->m_FilesFrom[j] == m_FileInfos[k].m_FileFD) break;
			_ASSERT(k!=m_NumVOBFiles);
			fprintf(fd,"%d %d %d %d %d %d\n",k,m_MasterList[i]->m_LBAFrom[j],m_MasterList[i]->m_CLBAFrom[j],
				m_MasterList[i]->m_VOBID, m_MasterList[i]->m_CELLID, m_MasterList[i]->m_StreamID);
		}
	}
	fclose(fd);
	return 0;
}

int CSubPicEngine::LoadSubpicInfo(const char* FileName)
{
	FILE* fd = fopen(FileName,"r");
	if(fd == 0)
		return -1;

	Reset();

	int i,j;
	int NumVobs;
	// Read all the file info:
	fscanf(fd,"%d\n",&NumVobs);
	if(NumVobs == 0)
	{
		fclose(fd);
		return 0;
	}
	m_CanSave = true;
	for(i=0;i<NumVobs;i++)
	{
		char Temp[1024];
		fscanf(fd,"\"%[^\"]\" %d\n",Temp,&j);
		if(AddFile(Temp) != 0)
		{
			fclose(fd);
			return -1;
		}
	}
	// Now load the IFO.
	ScanIFOs();
	// And notify ourselves that we loaded the files. 
	m_SupFileMode = false;
	m_NoVideo = false;
	OnFileLoaded(m_FileInfos[0].m_FileName);

	_ASSERT(m_NumVOBFiles = NumVobs);
	m_Phase = Processing_FastReloading;
	// Read video format and aspect ratio. Aspect ratio is a char so I have to read an int first
	// or I'll overwrite something else!
	fscanf(fd,"%d\n",&VideoFormat);
	fscanf(fd,"%d\n",&i);
	m_AspectRatio = i;

	VobuInfo AVobuInfo;
	fscanf(fd,"%d\n",&AVobuInfo.m_PTSFirstFrame);
	m_VobuInfo.push_back(AVobuInfo);
	// Now that we have the format...
	CSubPic::m_DisplayHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
	m_PTSPerFrame = (VideoFormat == FORMAT_PAL) ? 3600 : 3003;
	JustFoundVideoFormat();

	// And load all the SPU data.
	int TotNumSPU, NumPacks, NavPackLBA;
	fscanf(fd,"%d\n",&TotNumSPU);
	for(i=m_TotNumSPU=0;i<TotNumSPU;i++)
	{
		m_PercentDone = i*100/TotNumSPU;
		CSubPic* Temp = new CSubPic;
		fscanf(fd,"%d %d\n",&NumPacks,&NavPackLBA);
		int k;
		int LBAFrom, CLBAFrom,VOBID,CELLID,StreamID;
		for(j=0;j<NumPacks;j++)
		{
			// Read location of pack, seek to it, read data and tell CSubPic to load pack.
			fscanf(fd,"%d %d %d %d %d %d\n",&k,&LBAFrom,&CLBAFrom,
				&VOBID, &CELLID, &StreamID);
			fseek(m_FileInfos[k].m_FileFD,LBAFrom*SECTOR_SIZE,SEEK_SET);
			fread(m_Buffer,SECTOR_SIZE,1,m_FileInfos[k].m_FileFD);
			Temp->ReadThisPack(m_Buffer, m_FileInfos[k].m_FileFD, LBAFrom, NavPackLBA, VOBID, 
				CELLID,StreamID,CLBAFrom);
		}
		// This better be true!
		_ASSERT(Temp->m_SPUReadOK == true);
		// m_SubPics[StreamID] already as a 0 pointer at the back().
		m_SubPics[StreamID].back() = Temp;
		m_TopStreamID = max(m_TopStreamID,StreamID);
		AfterLastPack(Temp, StreamID);
	}
	// A bit of cleanup.
	for(i=0;i<32;i++)
	{
		if(m_SubPics[i].size() && m_SubPics[i].back() == 0)
			m_SubPics[i].pop_back();
	}
	if(m_MasterList.size() && m_MasterList.back() == 0)
		m_MasterList.pop_back();
	// Can't move because I don't have the VOBU info stuff.
	m_CanMove = false;
	fclose(fd);
	return 0;
}


int CSubPicEngine::SaveSelectedAsSupFile(const char* SupFileName)
{
	int i;
	FILE* fd = fopen(SupFileName,"wb");
	if(fd == 0)
		return -1;
	
	for(i=0;i<m_PruneLength;i++)
	{
		char Head[2]={'S','P'};
		// Write "SP" header.
		fwrite(Head,1,2,fd);
		CSubPic *TheSPU = m_PrunedList[i];
		int Zero = 0;
		// FIX: Apparently, the SPU format requires the PTS to be relative to the first PTS of the vob?
		// But is it the first PTS of the cell?
		int PTS = TheSPU->m_PTS - m_VobuInfo[0].m_PTSFirstFrame;
		fwrite(&PTS,4,1,fd);
		fwrite(&Zero,4,1,fd);
		fwrite(TheSPU->m_Data,1,TheSPU->m_SubPicSize,fd);
	}
	fclose(fd);
	return 0;
}


int CSubPicEngine::AddFile(const char* FileName)
{
	strcpy(m_FileInfos[m_NumVOBFiles].m_FileName,FileName);
	m_FileInfos[m_NumVOBFiles].m_FileFD = fopen(FileName,"rb+");
	if(m_FileInfos[m_NumVOBFiles].m_FileFD == 0)
	{
		m_FileInfos[m_NumVOBFiles].m_FileFD = fopen(FileName,"rb");
		if(m_FileInfos[m_NumVOBFiles].m_FileFD == 0)
			return -1;
		if(m_CanSave)
			ERROR2("File %s is read-only, you won't be able to save your changes!",FileName);
		m_CanSave = m_CanMove = false;
	}
	m_NumVOBFiles++;
	return 0;
}

int CSubPicEngine::PruneSPUList(int CurrentSelected, int StreamID, int StartLBA, int EndLBA, 
								ushort VOBID, uchar CELLID, int Modified, char* Search)
{
	int i;
	int N;
	int NavPackLBA = -1;
	int PrevSPUNumber = -1;
	int RetIndex = -1;
	m_ListHasOCR = 0;
	CSubPicPV Ptr;

	// Wait until the master list is updated.
	while(m_PruneLoadWait == 1)
		Sleep(1);
	m_PruneLoadWait = 1;
	if(CurrentSelected < m_PruneLength)
	{
		NavPackLBA = m_PrunedList[CurrentSelected]->m_NavPackLBA;
		PrevSPUNumber = m_PrunedList[CurrentSelected]->m_SPUNumber;
	}
	m_PruneLength = 0;
	m_PrunedList.clear();
	if(StreamID >= 0)
	{
		// A streamID was selected. Select the pointer that corresponds to that stream ID.
		if(StreamID > m_TopStreamID)
			goto end;
		Ptr = m_SubPics[StreamID];
		N = m_NumSubPics[StreamID];
	}
	else
	{
		// Otherwise, select the master pointer to get to all the SPUs.
		Ptr = m_MasterList;
		N = m_TotNumSPU;
	}
	for(i=0;i<N;i++)
	{
		CSubPic* Temp = Ptr[i];
		if(StartLBA != -1 && Temp->m_LBAFrom[0] < StartLBA)
			continue;
		if(EndLBA != -1 && Temp->m_LBAFrom[0] > EndLBA)
			continue;
		if(VOBID != 0 && Temp->m_VOBID != VOBID)
			continue;
		if(CELLID != 0 && Temp->m_CELLID != CELLID)
			continue;
		if(Modified != 0 && Temp->m_Modified == 0)
			continue;
		if(Search != 0 && strcmp(Search,"") !=0 && Temp->FindChars(Search) == false)
			continue;
		// The new list might not have the previously selected SPU. Let's find one that corresponds to
		// the same image (navpack).
		if(RetIndex == -1 && NavPackLBA <= Temp->m_NavPackLBA)
			RetIndex = m_PruneLength;
		// But if the previously selected SPU is in the new list, select this guy!
		if(PrevSPUNumber == Temp->m_SPUNumber)
			RetIndex = m_PruneLength;
		m_PrunedList.push_back(Temp);m_PruneLength++;
		if(!m_ListHasOCR && Temp->m_OCRLines[0][0] != '\0')
			m_ListHasOCR = true;
	}
	m_PruneLoadWait = 0;
	end: 
	return RetIndex;
}

int CSubPicEngine::RetreiveNextCommand(char* Command)
{
	if(m_SelectedSubPic == 0)
		return -1;
	static int CurrentDCSQT = 0;
	static int CurrentCommand = -1;
	return m_SelectedSubPic->GetCommand(Command, CurrentDCSQT, CurrentCommand);
}

int CSubPicEngine::CumulToLocalLBA(int CumulLBA, int &FileIndex)
{
	int i;
	for(i=0;i<m_NumVOBFiles;i++)
	{
		if(CumulLBA >= m_FileInfos[i].m_StartLBA + m_FileInfos[i].m_SizeLBA)
			continue;
		if(CumulLBA >= m_FileInfos[i].m_StartLBA)
		{
			FileIndex = i;
			return CumulLBA - m_FileInfos[i].m_StartLBA;
		}
	}
	FileIndex = -1;
	return -1;
}

int CSubPicEngine::LocalToCumulLBA(int LocalLBA, int FileIndex)
{
	_ASSERT(FileIndex < m_NumVOBFiles);
	return LocalLBA + m_FileInfos[FileIndex].m_StartLBA;
}

int CSubPicEngine::ReadPacksIntoMemory(int StartLBA, int SizeLBA)
{
	m_MemNeedsWrite = 0;
	m_MemStartLBA = StartLBA;
	m_MemSizeLBA = SizeLBA;
	// Find the file in which m_MemStartLBA resides.
	int FileIndex;
	int LStartLBA = CumulToLocalLBA(m_MemStartLBA,FileIndex);
	_ASSERT(FileIndex != -1);
	if(FileIndex == -1)
		return -1;
	uchar* Ptr = m_MemoryPtr;
	// Now read as much as needed,
	fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE,SEEK_SET);
	int ToRead = m_MemSizeLBA;
	while(1)
	{
		// Read as much as possible.
		int Read = fread(Ptr,1,SECTOR_SIZE*ToRead,m_FileInfos[FileIndex].m_FileFD);
		ToRead -= Read/SECTOR_SIZE;
		// We might be done here.
		if(ToRead == 0) return 0;
		// If not, increment m_MemoryPtr, go to the next file, seek back to 0, and go again!
		Ptr += Read;
		FileIndex++;
		_ASSERT(FileIndex < m_NumVOBFiles);
		if(FileIndex >= m_NumVOBFiles) return -1;
		fseek(m_FileInfos[FileIndex].m_FileFD,0,SEEK_SET);
	}
	return -1;
}

int CSubPicEngine::WritePacksToFiles()
{
	// Don't write if it's not needed.
	if(!m_MemNeedsWrite || m_MemSizeLBA == 0)
		return 0;
	// Find the file in which m_MemStartLBA resides.
	int FileIndex;
	int LStartLBA = CumulToLocalLBA(m_MemStartLBA,FileIndex);
	_ASSERT(FileIndex != -1);
	if(FileIndex == -1)
		return -1;
	uchar* Ptr = m_MemoryPtr;
	// Now write as much as needed,
	fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE,SEEK_SET);
	int ToWrite = m_MemSizeLBA;
	while(1)
	{
		// Write as much as possible.
		int Write = min(ToWrite,m_FileInfos[FileIndex].m_SizeLBA-LStartLBA);
		int ret = fwrite(Ptr,1,SECTOR_SIZE*Write,m_FileInfos[FileIndex].m_FileFD);
		_ASSERT(ret == SECTOR_SIZE*Write);
		ToWrite -= Write;
		// We might be done here.
		if(ToWrite == 0) {m_MemNeedsWrite = 0; return 0;}
		// If not, increment m_MemoryPtr, go to the next file, seek back to 0, and go again!
		Ptr += Write*SECTOR_SIZE;
		FileIndex++;
		LStartLBA = 0;
		_ASSERT(FileIndex < m_NumVOBFiles);
		if(FileIndex >= m_NumVOBFiles) return -1;
		fseek(m_FileInfos[FileIndex].m_FileFD,0,SEEK_SET);
	}
	return -1;
}

int CSubPicEngine::MovePackInMem()
{
	// Convert to local LBA.
	int LLeftLBA = m_LeftLimitLBA - m_MemStartLBA; _ASSERT(LLeftLBA >= 0);
	int LRightLBA = m_RightLimitLBA - m_MemStartLBA; _ASSERT(LRightLBA >= 0);
	int	DeltaLBA = (m_RightLimitLBA - m_LeftLimitLBA) + 1;

	uchar TempBuf[SECTOR_SIZE];
	if(m_MoveInc > 0)
	{
		// Moving memory to the right.
		memcpy(TempBuf,m_MemoryPtr+(LRightLBA+1)*SECTOR_SIZE,SECTOR_SIZE);
		memmove(m_MemoryPtr+(LLeftLBA+1)*SECTOR_SIZE,m_MemoryPtr+LLeftLBA*SECTOR_SIZE,
			DeltaLBA*SECTOR_SIZE);
		memcpy(m_MemoryPtr+LLeftLBA*SECTOR_SIZE,TempBuf,SECTOR_SIZE);
	}
	else if(m_MoveInc < 0)
	{
		// Moving memory to the left.
		memcpy(TempBuf,m_MemoryPtr+(LLeftLBA-1)*SECTOR_SIZE,SECTOR_SIZE);
		memmove(m_MemoryPtr+(LLeftLBA-1)*SECTOR_SIZE,m_MemoryPtr+LLeftLBA*SECTOR_SIZE,
			DeltaLBA*SECTOR_SIZE);
		memcpy(m_MemoryPtr+LRightLBA*SECTOR_SIZE,TempBuf,SECTOR_SIZE);
	}
	return 0;
}



// VobuStartLBA is the lba of the vobu start (absolute), PackLBA is the pack LBA, and
// Inc is +1 to add a pack, or -1 to remove it.
void CSubPicEngine::UpdateNavPackPtrs(VobuInfo* Vobu, int PackLBA, int Inc)
{
	// Increment or decrement forward pointer.
	int VobuStartLBA = Vobu->m_NavPackLBA;
	uchar* Ptr = m_MemoryPtr + (VobuStartLBA-m_MemStartLBA)*SECTOR_SIZE;
	AddInt(Ptr,0x040f,Inc);
//	AddInt(Vobu->m_NavPtr,0x040f,Inc);
	// Adjust first, second and third reference pointers if needed.
	PackLBA -= VobuStartLBA;
	if(PackLBA <= GetInt(Ptr,0x413)) 
	{
		AddInt(Ptr,0x413,Inc); 
//		AddInt(Vobu->m_NavPtr,0x413,Inc);
	}
	if(PackLBA <= GetInt(Ptr,0x417)) 
	{
		AddInt(Ptr,0x417,Inc);
//		AddInt(Vobu->m_NavPtr,0x417,Inc);
	}
	if(PackLBA <= GetInt(Ptr,0x41b)) 
	{
		AddInt(Ptr,0x41b,Inc);
//		AddInt(Vobu->m_NavPtr,0x41b,Inc);
	}
	// JEAN: TODO: adjust offsets to audio, sub packs...
}


int CSubPicEngine::MoveOnePack(int OrigVOBU)
{
	int i;
	_ASSERT(m_MemStartLBA == m_VobuInfo[OrigVOBU].m_NavPackLBA);
	
	// Find the vobu the Dest falls in. 
	int Dest = 0;
	int Src = 0;
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA + m_VobuInfo[i].m_SizeLBA > m_DestLBABM)
		{
			Dest = i;
			// Update the navpack pointers for an insertion (+1);
			UpdateNavPackPtrs(&m_VobuInfo[i], m_DestLBABM, 1);
			break;
		}
	}
	_ASSERT(i<m_VobuInfo.size());
	// Find the vobu the Src falls in.
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA + m_VobuInfo[i].m_SizeLBA > m_SrcLBA)
		{
			Src = i;
			// Update the navpack pointers for a deletion (-1)
			UpdateNavPackPtrs(&m_VobuInfo[i], m_SrcLBA, -1);
			break;
		}
	}
	_ASSERT(i<m_VobuInfo.size());
	// The sizes also changed, and the slot counts too.
	m_VobuInfo[Dest].m_SizeLBA ++;
	m_VobuInfo[Dest].m_EmptySlotCount --;
	m_VobuInfo[Src].m_SizeLBA --;
	m_VobuInfo[Src].m_EmptySlotCount ++;

	// Finally, correct the this pointers and the LBA pointers in our vobu infos.
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA  > m_RightLimitLBA)
			break;
		if(m_VobuInfo[i].m_NavPackLBA >= m_LeftLimitLBA)
		{
			// Update the this->lba pointers
			AddInt(m_MemoryPtr+(m_VobuInfo[i].m_NavPackLBA-m_MemStartLBA)*SECTOR_SIZE,0x2d,m_MoveInc);
			AddInt(m_MemoryPtr+(m_VobuInfo[i].m_NavPackLBA-m_MemStartLBA)*SECTOR_SIZE,0x40b,m_MoveInc);
//			AddInt(m_VobuInfo[i].m_NavPtr,0x2d,m_MoveInc);
//			AddInt(m_VobuInfo[i].m_NavPtr,0x40b,m_MoveInc);
			m_VobuInfo[i].m_NavPackLBA += m_MoveInc;
		}
	}

	// Now move the pack.
	MovePackInMem();
	// And adjust its SRC. Retrieve the SRC of the preceding pack and increment it.
	CompositeSRC CSRC = ParseSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA-1)*SECTOR_SIZE);
	IncSRC(CSRC);
	WriteSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA)*SECTOR_SIZE,CSRC);
#ifdef _DEBUG
	// Check that the next SRC is good too!
	IncSRC(CSRC);
	CompositeSRC Next = ParseSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA+1)*SECTOR_SIZE);
	_ASSERT(SRC1GESRC2(Next, CSRC));
#endif
	return 0;
}

int CSubPicEngine::FindEmptySlot(VobuInfo* DestVobu, int MinLBA)
{
	// Start at StartLBA
	MinLBA = max(MinLBA,DestVobu->m_NavPackLBA);
	uchar* Ptr = m_MemoryPtr + (MinLBA-m_MemStartLBA)*SECTOR_SIZE;
	CompositeSRC ThisSRC = ParseSRC(Ptr);

	int i;
	CompositeSRC NextSRC;
	for(i=1;i<DestVobu->m_SizeLBA;i++)
	{
		// Get SRC.
		NextSRC = ParseSRC(Ptr+i*SECTOR_SIZE);

		// Now increment the previous SRC twice, and see if the current one is larger or equal. 
		IncSRC(ThisSRC);
		IncSRC(ThisSRC);
		if(SRC1GESRC2(NextSRC, ThisSRC))
		{
			// We're good!
			return MinLBA+i;
		}
		ThisSRC = NextSRC;
	}
	return -1;	
}

int CSubPicEngine::FindEmptySlotBck(VobuInfo* DestVobu, int MinLBA)
{
	// Start at StartLBA
	MinLBA = min(MinLBA,DestVobu->m_NavPackLBA+DestVobu->m_SizeLBA-1);
	uchar* Ptr = m_MemoryPtr + (MinLBA-m_MemStartLBA)*SECTOR_SIZE;
	ushort SRCH = GetShort(Ptr,0x4);
	uint SRCL = GetInt(Ptr,0x6);
	CompositeSRC ThisSRC;
	SrcToFrac(SRCL, SRCH, ThisSRC);

	int i;
	CompositeSRC PrevSRC;
	for(i=1;i<MinLBA-DestVobu->m_NavPackLBA;i++)
	{
		// Get SRC.
		SRCH = GetShort(Ptr-i*SECTOR_SIZE,0x4);
		SRCL = GetInt(Ptr-i*SECTOR_SIZE,0x6);
		SrcToFrac(SRCL, SRCH, PrevSRC);
		CompositeSRC TempSRC = PrevSRC;
		// Now increment the previous SRC twice, and see if the current one is larger or equal. 
		IncSRC(PrevSRC);
		IncSRC(PrevSRC);
		if(PrevSRC.m_SRCInt < ThisSRC.m_SRCInt || 
			(PrevSRC.m_SRCInt == ThisSRC.m_SRCInt && PrevSRC.m_SRCFrac <= ThisSRC.m_SRCFrac))
		{
			// We're good!
			return MinLBA-i+1;
		}
		ThisSRC = TempSRC;
	}
	return -1;	
}

int CSubPicEngine::FindDestVobu(int SrcVobu, int PTS, int DeltaPTS)
{
	int i;
	int DestPTS = PTS + DeltaPTS;
	if(DeltaPTS >= 0)
	{
		for(i=SrcVobu;i<m_VobuInfo.size();i++)
		{
			if(m_VobuInfo[i].m_PTSFirstFrame >= DestPTS)
				break;
		}
		_ASSERT(i>0);
		i=max(0,i-1);
	}
	else
	{
		for(i=SrcVobu;i>=0;i--)
		{
			if(m_VobuInfo[i].m_PTSFirstFrame <= DestPTS)
				break;
		}
		_ASSERT(i>=0);
		i=max(0,i);
	}
	return i;
}

#define OVERALLOC 1
int CSubPicEngine::ChangeTiming(int SubpicIndex, int DeltaPTS)
{
	int i;
	CSubPic* ThisSubpic = m_MasterList[SubpicIndex];
	// Not moving button highlights!!! 
	if(ThisSubpic->ContainsHighlights())
	{
		MESSAGE2("Subpic %d seems to contain button highlights, not moving",SubpicIndex);
		return -1;
	}
	if(m_StreamCanMove[ThisSubpic->m_StreamID] == 0)
	{
		MESSAGE2("Subpic %d belongs to a stream that can't be moved ",SubpicIndex);
		return -1;
	}

	// First find the source vobu (the vobu the first spu pack is in).
	int SrcVobu;
	int SrcLBA = ThisSubpic->m_CLBAFrom[0];
	for(SrcVobu=0;SrcVobu<m_VobuInfo.size();SrcVobu++)
	{
		if(m_VobuInfo[SrcVobu].m_NavPackLBA > SrcLBA)
			break;
	}
	SrcVobu --;
	_ASSERT(SrcVobu >= 0 && SrcVobu<m_VobuInfo.size()-1);

	// Then find destination VOBU.
	int DestPTS = ThisSubpic->m_PTS + DeltaPTS;
	int DestVobu = FindDestVobu(SrcVobu, ThisSubpic->m_PTS, DeltaPTS);

	// Check whether the VOBU has enough empty slots for our packs and if we run a risk of colliding.
	// I'm trying one, two vobus before, then 1 after then 3 before...
	int Tests[] = {0,-1,-2,1};
	int TestDest;
	for(i=0;i<4;i++)
	{
		TestDest = DestVobu+Tests[i];
		if(TestDest == SrcVobu) break; // No problem with leaving this guy here.
		// See whether there's enough empty slots for our packs, and if there's no collision with another subpic.
		if(TestDest >= 0 && TestDest < m_VobuInfo.size() && m_VobuInfo[TestDest].m_EmptySlotCount >= ThisSubpic->m_NumPackInSPU 
			&& CheckForCollision(ThisSubpic, TestDest) == -1)
			break;
	}
	if(i==4)
	{
		ERROR2("Can't find an empty slot for subpic %d! It won't be moved....\n",SubpicIndex);
		return -1;
	}
	DestVobu = TestDest;

	// If we're staying in the same VOBU, we don't need to move anything, right?
	if(SrcVobu == DestVobu)
	{
		// Change the PTS: write it in the subpic, and on the file.
		MESSAGE2("No moving required for spu %d, setting PTS",SubpicIndex);
		ThisSubpic->SetPTS(DestPTS);
		// Save it, since we're always saving when we move.
		WritePacksToFiles();
		ThisSubpic->SaveSPU();
		return -1;
	}

	if(m_FoundILVU)
	{
		ERROR2("This timing change for spu %d would require moving the subpic pack.\nInterleaved video units (ILVUs)\
 were found in the VOB file. I can't move the pack",SubpicIndex);
		return -1;
	}

	MESSAGE5("Moving spu %d from VOBU %d to VOBU %d -- %d",SubpicIndex,SrcVobu,DestVobu,i);

	// Because the end of the spu can be beyond the end of the VOBU in which the beginning appears,
	// we'll load more memory than we need.
	int OrigVOBU = min(SrcVobu,DestVobu)-OVERALLOC; OrigVOBU = max(OrigVOBU,0);
	int EndVOBU = max(SrcVobu,DestVobu)+OVERALLOC; EndVOBU = min(EndVOBU,m_VobuInfo.size()-1);
	// Now we have our original and end lbas.
	int OrigLBA = m_VobuInfo[OrigVOBU].m_NavPackLBA;
	int EndLBA = m_VobuInfo[EndVOBU].m_NavPackLBA + m_VobuInfo[EndVOBU].m_SizeLBA;
	int SizeLBA = EndLBA - OrigLBA;

	// Load the memory between original and end lbas...
	if(OrigLBA < m_MemStartLBA || OrigLBA+SizeLBA > m_MemStartLBA+m_MemSizeLBA)
	{
		// We need more memory. Let's write what we have and alloc more.
		WritePacksToFiles();
		if(m_MemAllocSize < SizeLBA)
		{
			delete[] m_MemoryPtr;
			m_MemoryPtr = new uchar[SizeLBA * SECTOR_SIZE];
			m_MemAllocSize = SizeLBA;
		}
		if(ReadPacksIntoMemory(OrigLBA, SizeLBA) != 0)
		{
			ERROR1("Couldn't load memory");
			return -1;
		}
		m_MemOrigVOBU = OrigVOBU; 
	}

	int DestLBA = 0;
	// Move the packs!
	for(i=0;i<ThisSubpic->m_NumPackInSPU;i++)
	{
		m_MemNeedsWrite = 1;
		SrcLBA = ThisSubpic->m_CLBAFrom[i];
		// Find the next empty slot in destination vobu. Make sure we place subsequent packs after
		// in the right order!
		DestLBA = FindEmptySlot(&m_VobuInfo[DestVobu],DestLBA+1);
		if(DestLBA == -1)
		{
			// This should never happen, since we checked we had one.
			ERROR1("Couldn't locate an empty slot!");
			return -1;
		}

		// This sets the lba area that will move, as well as the src and dest lbas.
		SetMoveAreaLimit(SrcLBA, DestLBA);

		// Then move the pack there. 
		if(MoveOnePack(m_MemOrigVOBU) != 0)
		{
			ERROR1("Problem moving the pack!");
			return -1;
		}
		// Update all the subpics!
		UpdateAllSubpics();
		// Update the spu's navpack, fd and LBA info.
		if(i==0)
			ThisSubpic->m_NavPackLBA = m_VobuInfo[DestVobu].m_NavPackLBA;
		int NewFileIndex;
		int NewLocalLBA = CumulToLocalLBA(m_DestLBAAM, NewFileIndex);
		ThisSubpic->m_FilesFrom[i] = m_FileInfos[NewFileIndex].m_FileFD;
		ThisSubpic->m_LBAFrom[i] = NewLocalLBA;
		ThisSubpic->m_CLBAFrom[i] = m_DestLBAAM;
	}
#ifdef _DEBUG
	// Checking that we didn't mess up the order in which subpic packs appear!
	for(i=1;i<ThisSubpic->m_NumPackInSPU;i++)
		_ASSERT(ThisSubpic->m_CLBAFrom[i] > ThisSubpic->m_CLBAFrom[i-1]);
#endif
	// Change the PTS: write it in the subpic, and on the file.
	ThisSubpic->SetPTS(DestPTS);
	BuildPTS(m_MemoryPtr+(ThisSubpic->m_CLBAFrom[0]-m_MemStartLBA)*SECTOR_SIZE+0x17, DestPTS, 0x20);
	return 0;
}


bool CSubPicEngine::DidThisPackMove(int PackLBA)
{
	return (PackLBA >= m_LeftLimitLBA && PackLBA <= m_RightLimitLBA);
}

int CSubPicEngine::UpdateAllSubpics()
{
	int i,j;
	int NewFileIndex;
	int NewLocalLBA;
	for(i=0;i<m_TotNumSPU;i++)
	{
		CSubPic* TheSubpic = m_MasterList[i];
		// If the last pack is left of LeftLBA, nothing to do.
		// CAN'T do that! During processing, it's possible for m_CLBAFrom[TheSubpic->m_NumPackInSPU-1]
		// to have an "old" value even though m_CLBAFrom[0] needs updating.
//		if(TheSubpic->m_CLBAFrom[TheSubpic->m_NumPackInSPU-1] < m_LeftLimitLBA)
//			continue;
		// If the navpack pack is right of RightLBA, we're done.
		if(TheSubpic->m_NavPackLBA > m_RightLimitLBA)
			break;
		// Check if the navpack moved, update if needed.
		if(DidThisPackMove(TheSubpic->m_NavPackLBA))
			TheSubpic->m_NavPackLBA += m_MoveInc;
		// Check if any of the subpic packs have moved.
		for(j=0;j<TheSubpic->m_NumPackInSPU;j++)
		{
			if(DidThisPackMove(TheSubpic->m_CLBAFrom[j]))
			{
				TheSubpic->m_CLBAFrom[j] += m_MoveInc;
				NewLocalLBA = CumulToLocalLBA(TheSubpic->m_CLBAFrom[j], NewFileIndex);
				TheSubpic->m_LBAFrom[j] = NewLocalLBA;
				TheSubpic->m_FilesFrom[j] = m_FileInfos[NewFileIndex].m_FileFD;
			}
		}
	}
	return 0;
}

int CSubPicEngine::ReorderMasterList(int SrcSPUIndex)
{
	int i;
	// Find destination index. To do that, find the first SPU whose first pack is past
	// ours.
	int FirstPackLBA = m_MasterList[SrcSPUIndex]->m_CLBAFrom[0];
	for(i=0;i<m_TotNumSPU;i++)
	{
		if(m_MasterList[i]->m_CLBAFrom[0] > FirstPackLBA)
			break;
	}
	int DestIndex = i;

	CSubPic* ThisSPU = m_MasterList[SrcSPUIndex];
	if(DestIndex < SrcSPUIndex)
	{
		// Moving left.
		_ASSERT(DestIndex < m_TotNumSPU);
		for(i=SrcSPUIndex;i>DestIndex;i--)
		{
			m_MasterList[i] = m_MasterList[i-1];
			m_MasterList[i]->m_SPUNumber = i;
		}
	}
	else
	{
		// Moving right. The destination should not move, the one immediately left of it should.
		DestIndex--;
		_ASSERT(DestIndex >= 0);
		for(i=SrcSPUIndex;i<DestIndex;i++)
		{
			m_MasterList[i] = m_MasterList[i+1];
			m_MasterList[i]->m_SPUNumber = i;
		}
	}
	m_MasterList[DestIndex] = ThisSPU;
	m_MasterList[DestIndex]->m_SPUNumber = DestIndex;
	return 0;
}

void CSubPicEngine::SetMoveAreaLimit(int SrcLBA, int DestLBA)
{
	// SrcLBA is the original LBA of the pack to be moved.
	// DestLBA is the target LBA where the pack is to be moved measured *before the move!*.
	// When moving the pack to the right, the memory will be moved 1 pack to the left, so the
	// "after the move" destination LBA will be DestLBA-1.
	if(DestLBA < SrcLBA)
	{
		// The pack is moving left, the area is moving right.
		m_LeftLimitLBA = DestLBA;	// The destination pack is moving right.
		m_RightLimitLBA = SrcLBA-1;	// The last pack to move is the one right before Src.
		m_MoveInc = 1;
		m_DestLBAAM = DestLBA;
	}
	else if(DestLBA > SrcLBA)
	{
		// The pack is moving right, the area is moving left.
		m_LeftLimitLBA = SrcLBA+1;	// The first pack to move is the one rigth of Src.
		m_RightLimitLBA = DestLBA-1;// The last pack to move is the one right before DestLBA.
		m_MoveInc = -1;
		m_DestLBAAM = DestLBA-1;
	}
	else
	{
		m_LeftLimitLBA = m_RightLimitLBA = SrcLBA;	
		m_DestLBAAM = DestLBA;
		m_MoveInc = 0;
	}
	m_SrcLBA = SrcLBA;
	m_DestLBABM = DestLBA;
}

int CSubPicEngine::SortMasterLists()
{
	// Make sure the SPU number is updated in each spu.
	MESSAGE1("Sorting Master Lists");
	SortThisList(&m_MasterList,true);
	// Now do the stream lists.
	int StreamID;
	for(StreamID=0;StreamID<32;StreamID++)
	{
		SortThisList(&m_SubPics[StreamID]);
	}
	return 0;
}

// Function used to sort subpics.
bool foo(CSubPic* A, CSubPic* B) 
{
	return A->m_CLBAFrom[0] < B->m_CLBAFrom[0];
}

int CSubPicEngine::SortThisList(CSubPicPV* List, bool UpdateSPUNumber)
{
	// Sorting from scratch. Using STL sort for speed (much much faster).
	int i,k;
	// Just a precaution, because of the way I'm creating these lists.
	if(List->size() && List->back() == 0)
		List->pop_back();
	int N=List->size();

	// Sort this guy.
	std::sort(List->begin(),List->end(),foo);
	if(UpdateSPUNumber)
	{
		for(i=0;i<N;i++)
		{
			(*List)[i]->m_SPUNumber = i;
			// Also update the "forward" pointer.
			for(k=i-1;k>0;k--)
			{
				if((*List)[k]->m_StreamID == (*List)[i]->m_StreamID)
				{
					// Make its m_NextSPUInStream pointer point to the new guy.
					// TODO FIX: Adjust the spu duration so it does not collide?
					(*List)[k]->m_NextSPUInStream = (*List)[i];
					break;
				}
			}
		}
	}
	return 0;
}


void CSubPicEngine::UpdateIFOAfterMove()
{
	MESSAGE1("Updating IFO");
	if(m_VobuInfo.size() == 0 || m_OutputLBA == 0)
		return;
	int i,j;
	int TableSize = m_VobuInfo.size()+1;
	// Build the output LBA array based on our vobu info vector.
	for(i=0;i<m_VobuInfo.size();i++)
		m_OutputLBA[i] = m_VobuInfo[i].m_NavPackLBA;
	m_OutputLBA[i] = m_InputLBA[i];

	// See if anything changed.
	for(i=0;i<TableSize;i++)
	{
		if(m_OutputLBA[i] != m_InputLBA[i]) 
			break;
	}
	// No need to updated.
	if(i==TableSize)
		return;

	// Now that we have both the input and output LBA array, update IFO...
	// First load it.
	CIFOLoader IFO(GetIFO(m_FileInfos[0].m_FileName));
	char Num = m_FileInfos[0].m_FileName[strlen(m_FileInfos[0].m_FileName)-5];
	// Not updating menu domain.
	if(Num <= '0' || Num > '9' || !IFO.m_LoadedOK)
		return;
	// This updates the VOBU_ADMAP tables, and the PGC pointers.
	IFO.UpdateNavPackInfo(m_InputLBA, m_OutputLBA, TableSize,false);

	// Now cleanup VTS_C_ADT. I'm doing it by hand because the CIFOLoader class does not have
	// a function to do that from an input/output vobu map.
	// Get the table.
	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) IFO.FindTable(VTS_C_ADT);
	if(Table != 0)
	{
		ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
		VTSM_C_ADTItem* ThisItem;
		// Loop over all the cells in the VTS_C_ADT table.
		for(i=0;i<(int)NumCells;i++)
		{
			// Update the start and end lba of each cell according to our input/output map.
			ThisItem = &Table->m_Items[i];
			uint StartLBA = FixEndian((uint)ThisItem->m_StartLBA);
			uint EndLBA = FixEndian((uint)ThisItem->m_EndLBA);
			for(j=0;j<TableSize;j++)
			{
				// Find where the startlba was moved
				if(StartLBA == m_InputLBA[j])
				{
					ThisItem->m_StartLBA = FixEndian((uint)m_OutputLBA[j]);
					break;
				}
			}
//			_ASSERT(j<TableSize);
			for(j=0;j<TableSize;j++)
			{
				// Find where the endlba was moved. The end lba points 1 pack before the start of the
				// next vobu. 
				if(EndLBA == m_InputLBA[j]-1)
				{
					ThisItem->m_EndLBA = FixEndian((uint)m_OutputLBA[j]-1);
					break;
				}
			}
//			_ASSERT(j<TableSize);
		}
	}
	// Save IFO.
	IFO.WriteIFO();
	
	// Now everything is in synch.
	for(i=0;i<TableSize;i++)
		m_InputLBA[i] = m_OutputLBA[i];
}


int CSubPicEngine::CheckForCollision(CSubPic* TheSubpic, int DestVobu)
{
	int i;
	for(i=0;i<m_TotNumSPU;i++)
	{
		if(m_MasterList[i] == TheSubpic ||
			m_MasterList[i]->m_StreamID != TheSubpic->m_StreamID)
			continue;
		// Same stream ID, let's see whether the subpic packs will intersect.
		if(m_MasterList[i]->m_NavPackLBA == m_VobuInfo[DestVobu].m_NavPackLBA)
		{
			// Collision will occur. The subpic in question has the same navpack...
			return i;
		}
	}
	return -1;
}

// This will update Offset depending on whether the target lba moved relative to the start
// of the vobu. Dir indicates whether Offset is a forward or backward pointer.
int CSubPicEngine::UpdateThisPointer(int VobuIndex, int Offset)
{
	// Calculate the distance between the current VobuStart and the current location of the
	// target. 
	int OrigTargetLBA = m_VobuInfo[VobuIndex].m_OrigNavPackLBA + Offset;
	int NewTargetLBA;
	// Now, find the new location of OrigTargetLBA;
	int Inc = Offset > 0 ? 1 : -1;
	int i = VobuIndex;
#if 1
	// This version will not fix corrupted SRI (trick-play) pointers.
	for(i=VobuIndex;i>=0 && i< m_VobuInfo.size();i+=Inc)
		if(OrigTargetLBA == m_VobuInfo[i].m_OrigNavPackLBA)
			break;
	if(i == -1 ||  i == m_VobuInfo.size())
	{
		// We didn't find the navpack. Could be two things: either the pointers were corrupted or we only
		// loaded 1 of several vob files from a multi-vob VTS.
		MESSAGE2("VOBU SRI offset corrupted at LBA %d",m_VobuInfo[VobuIndex].m_NavPackLBA);
		return Offset;
	}
#else
	// FIX: TODO: return the closest pointers if you can't find an exact match!
	if(Offset > 0)
	{
		for(i=VobuIndex;i< m_VobuInfo.size();i++)
			if(m_VobuInfo[i].m_OrigNavPackLBA >= OrigTargetLBA)
				break;
		int Diff = m_VobuInfo[i].m_OrigNavPackLBA - OrigTargetLBA;
		if(Diff != 0 && i>0 && OrigTargetLBA - m_VobuInfo[i-1].m_OrigNavPackLBA < Diff)
		{
			MESSAGE2("VOBU SRI offset corrupted at LBA %d... Fixing",m_VobuInfo[VobuIndex].m_NavPackLBA);
			i--;
		}
	}
	else
	{
		for(i=VobuIndex;i>=0;i--)
			if(m_VobuInfo[i].m_OrigNavPackLBA <= OrigTargetLBA)
				break;
		int Diff = OrigTargetLBA - m_VobuInfo[i].m_OrigNavPackLBA;
		if(Diff != 0 && i<m_VobuInfo.size()-1 && m_VobuInfo[i+1].m_OrigNavPackLBA - OrigTargetLBA < Diff)
		{
			MESSAGE2("VOBU SRI offset corrupted at LBA %d... Fixing",m_VobuInfo[VobuIndex].m_NavPackLBA);
			i++;
		}
	}
#endif
	NewTargetLBA = m_VobuInfo[i].m_NavPackLBA;
	return NewTargetLBA-m_VobuInfo[VobuIndex].m_NavPackLBA;

}

// Parses an offset in the navpack data in Ptr at offset Loc. Dir indicate whether the offset
// is forward (+1) or backward (-1). VobuIndex is the index of the current navpack we're updating.
// If the offset needs to be updated, it is written and the function returns true. Otherwise false.
bool CSubPicEngine::ParseAndUpdatePointer1(uchar* Ptr, int Loc, int Dir, int VobuIndex)
{
	int Offset;
	Offset = GetInt(Ptr,Loc-NAV_MEM_START);
	if(Offset & 0x80000000 && Offset != 0xbfffffff)
	{
		int NewOffset = UpdateThisPointer(VobuIndex,Dir*(int)(Offset & 0x7FFFFFFF));
		NewOffset = abs(NewOffset);
		NewOffset |= 0x80000000;
		PutInt(Ptr,Loc-NAV_MEM_START,NewOffset);
		return NewOffset!=Offset;
	}
	return false;
}

// Parses an offset in the navpack data in Ptr at offset Loc. Dir indicate whether the offset
// is forward (+1) or backward (-1). VobuIndex is the index of the current navpack we're updating.
// If the offset needs to be updated, it is written and the function returns true. Otherwise false.
bool CSubPicEngine::ParseAndUpdatePointer2(uchar* Ptr, int Loc, int Dir, int VobuIndex)
{
	int Offset;
	Offset = GetInt(Ptr,Loc-NAV_MEM_START);
	if(Offset & 0x80000000 && Offset != 0xbfffffff)
	{
		int Bits = Offset & 0xc0000000;
		int NewOffset = UpdateThisPointer(VobuIndex,Dir*(int)(Offset & 0x3FFFFFFF));
		NewOffset = abs(NewOffset);
		NewOffset |= Bits;
		PutInt(Ptr,Loc-NAV_MEM_START,NewOffset);
		return NewOffset!=Offset;
	}
	return false;
}

int CSubPicEngine::UpdateVobuSRI(bool JustCheck)
{
	int VobuIndex;
	MESSAGE1("Updating VOBU SRI pointers");
	m_Phase = Processing_UpdatingSRI;
	for(VobuIndex=0;VobuIndex<m_VobuInfo.size();VobuIndex++)
	{
		m_PercentDone = 100*VobuIndex/m_VobuInfo.size();
		// Read all the pointers, see if they need updating, update and write if necessary.
		bool NeedsWrite = 0;
		uchar* Ptr = m_VobuInfo[VobuIndex].m_NavPtr;
		// Parse and update first vobu_sri offset.
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x4f1,1,VobuIndex);
		int i;
		for(i=0x4f5;i<=0x53d;i+=4)
		{	// Forward pointers.
			NeedsWrite |= ParseAndUpdatePointer2(Ptr,i,1,VobuIndex);
		}
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x541,1,VobuIndex);
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x545,-1,VobuIndex);
		for(i=0x549;i<=0x591;i+=4)
		{	// Backward pointers
			NeedsWrite |= ParseAndUpdatePointer2(Ptr,i,-1,VobuIndex);
		}
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x595,-1,VobuIndex);
		if(NeedsWrite && !JustCheck)
		{
			// At least one of the offset was updated. Write the pack..
			int FileIndex;
			int LStartLBA = CumulToLocalLBA(m_VobuInfo[VobuIndex].m_NavPackLBA,FileIndex);
			_ASSERT(FileIndex != -1);
			if(FileIndex == -1)
				return -1;
			fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE+NAV_MEM_START,SEEK_SET);
			int ret = fwrite(Ptr,1,NAV_MEM_SIZE,m_FileInfos[FileIndex].m_FileFD);
			_ASSERT(ret == NAV_MEM_SIZE);		
		}
	}
	for(VobuIndex=0;VobuIndex<m_VobuInfo.size();VobuIndex++)
		m_VobuInfo[VobuIndex].m_OrigNavPackLBA = m_VobuInfo[VobuIndex].m_NavPackLBA;
	return 0;
}
