#include "SubPicEngine.h"
#include <stdio.h>
#include <objbase.h>
#include "Cell.h"
#include "Utils.h"
#include "SubPic.h"
#include "IFOUtils.h"
#include <algorithm>

typedef unsigned char uchar;

// Only a small part of the navpacks is kept in memory (to minimize memory usage).
#define NAV_MEM_START		0x04f1				// Offset of first byte of navpack kept in mem.
#define NAV_MEM_SIZE		(0x0600-0x04f1)		// How many bytes we're copying to mem.

#define VERS "DVDSubEdit -- Beta 0.79 By Jeanl"

#define CHAR_SIZE 256
// This stuff is global and accessed by all the classes, for messages and errors.
// The string where messages are stored.
char Message[CHAR_SIZE];
// Our default function to handle regular and error messages.
void DefaultLogMessage(char* AMessage, int Langage)
{
	printf(AMessage);
	strcpy(Message,AMessage);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}
// Global pointers that classes use to send regular and error messages.
void (*pMessageFunction)(char* AMessage, int Langage) = DefaultLogMessage;
void (*pErrorFunction)(char* AMessage, int Langage) = DefaultLogMessage;
int	 (*FrameSelectFunction)(int VobId, int CellId, int StartLBA, int EndLBA, int Default, 
							int NumLBA, int NormallySkipped) = 0;

// I need this destructor so the CLUT arrays are automatically destroyed.
PGC_CLUT_Info::~PGC_CLUT_Info()
{
	int i;
	for(i=0;i<m_CLUT.size();i++)
		delete[] m_CLUT[i];
}

VobuInfo::VobuInfo()
{
	m_NavPackLBA = 0;
	m_OrigNavPackLBA = 0;
	m_PTSFirstFrame = 0;	
	m_EmptySlotCount = 0;	
	m_SizeLBA = 0;
	m_NavPtr = 0;
}



CSubPicEngine::CSubPicEngine()
{
	m_CLUTs = 0;
	strcpy(m_Version,VERS);
	int i;
	for(i=0;i<10;i++)
		m_FileInfos[i].m_FileFD = 0;
	m_InputLBA = 0;	
	m_OutputLBA = 0;
	m_MemoryPtr = 0;
	Reset();
}
	
CSubPicEngine::~CSubPicEngine()
{
	Reset();
	delete[] m_CLUTs;
	int i;
	for(i=0;i<32;i++)
	{
		m_SubPics[i].clear();
	}
}

void CSubPicEngine::Reset()
{
	int i;
	m_AbortNow = 0;
	m_TotNumSPU = 0;
	m_PruneLength = 0;
	m_PrunedList.clear();
	m_PercentScanned = 0;
	m_Phase = Processing_Idle;
	m_NumVOBFiles = 0;
	m_TopStreamID = 0;
	m_SelectedSubPic = 0;
	m_PTSPerFrame = 3003;
	m_CanMove = 0;
	m_PercentDone = 0;
	m_Wait = 0;
	m_FoundILVU = 0;

	m_LeftLimitLBA = 0;	
	m_RightLimitLBA = 0;
	m_MoveInc = 0;
	delete[] m_MemoryPtr;
	m_MemoryPtr = 0;
	m_MemNeedsWrite = 0;
	m_MemStartLBA = 0;
	m_MemSizeLBA = 0;
	m_MemAllocSize = 0;
	m_MemOrigVOBU = 0;
	m_SrcLBA = 0;	
	m_DestLBABM = 0;
	m_DestLBAAM = 0;

	m_VCIDLength = 0;
	m_CLUTLength = 0;
	m_AspectRatio = 0;
	memset(m_NumSubPics,0,32*sizeof(int));
	for(i=0;i<32;i++)
	{
		int j;
		for(j=0;j<m_SubPics[i].size();j++)
			delete m_SubPics[i][j];
		m_SubPics[i].clear();
		m_SubPics[i].push_back(0);
	}
	m_MasterList.clear();

	for(i=0;i<m_VobuInfo.size();i++)
		delete[] m_VobuInfo[i].m_NavPtr;
	m_VobuInfo.clear();
	
	for(i=0;i<10;i++)
	{
		if(m_FileInfos[i].m_FileFD)
			fclose(m_FileInfos[i].m_FileFD);
		m_FileInfos[i].m_FileFD = 0;
	}
	delete[] m_InputLBA; m_InputLBA = 0;	
	delete[] m_OutputLBA; m_OutputLBA = 0;
}	

void CSubPicEngine::ScanFilesForSPUs()
{
	// Scan all the vob files for subpics...
	int i,LBA,ret;
	int TotNumLBA = 0;
	int CumulLBA = 0;
	int NavPackNumButtons = 0;
	DWORD End;

	CompositeSRC LastNavPackSRC ;
	LastNavPackSRC.m_SRCInt = 0;
	LastNavPackSRC.m_SRCFrac= 0;

	for(i=0;i<m_NumVOBFiles;i++)
	{
		m_FileInfos[i].m_StartLBA = TotNumLBA;
		m_FileInfos[i].m_SizeLBA = GetFileSizekB(m_FileInfos[i].m_FileName)/2;
		TotNumLBA += m_FileInfos[i].m_SizeLBA;
		fseek(m_FileInfos[i].m_FileFD,0,SEEK_SET);
	}

	VobuInfo TheVobuInfo;

	CPack ThePack;
	int LastNavPack = -1;
	m_TotNumSPU = 0;
	m_MasterList.clear();
	DWORD Start = timeGetTime();
	m_Phase = Processing_Scanning_VOB;
	float SmoothRate = 0;
	VideoFormat = FORMAT_UNKNOWN;
	for(i=0;i<m_NumVOBFiles;i++)
	{
		for(LBA=0;;LBA++,CumulLBA++)
		{
			if(m_AbortNow)
			{
				m_Phase = Processing_Aborted;
				goto end ;
			}
			m_PercentScanned = 100*(float)CumulLBA / TotNumLBA;
			if((CumulLBA+1) % 2000 == 0)
			{
				// This is just to report progress and estimated end time...
				DWORD End = timeGetTime();
				float RateMBpS = (1000*(float)CumulLBA*2/1024/(End-Start));
				if(SmoothRate == 0) SmoothRate = RateMBpS;
				else SmoothRate = .95f*SmoothRate + .05f*RateMBpS;
				m_RemainingTimeSec = (TotNumLBA - CumulLBA)*2/1024 / SmoothRate;
				MESSAGE5("%.2f MB scanned (%.2f MB/s) -- %d subpics -- %.0f secs to go",(float)CumulLBA/2/1024,
					RateMBpS,m_TotNumSPU, m_RemainingTimeSec);
			}
			// Read next pack.
			ret=(fread(m_Buffer,1,SECTOR_SIZE,m_FileInfos[i].m_FileFD) == SECTOR_SIZE);
			if(ret != 1)
				break;
			ushort VOBID,vobid;
			uchar CELLID,cellid;
			ThePack.FindPackType(m_Buffer, vobid, cellid);
			
			if(ThePack.m_PackType != PackTypeSupPic)
			{
				if(ThePack.m_PackType == PackTypeNavPack)
				{
					// Pack is a navpack. Keep track of the VCID.
					VOBID = vobid; CELLID = cellid;
					LastNavPack = CumulLBA;
					// The spu will need to know whether there are buttons.
					NavPackNumButtons = ThePack.m_NumButtons;
					// Detect ILVUs...
					m_FoundILVU |= (GetShort(m_Buffer,0x427) != 0);
					// Our info on vobus... If LBA isn't 0, now's the time to push the last
					// vobu info.
					CompositeSRC ThisSRC ;
					SrcToFrac(ThePack.m_SRCL, ThePack.m_SRCH, ThisSRC);
					if(m_VobuInfo.size())
					{
						// Count how many empty slots we have. An empty slot lasts 146.86 SRC tics.
						// Measure how many slots in the VOBU, remove the number of packs.
						m_VobuInfo.back().m_EmptySlotCount = ((150*(ThisSRC.m_SRCInt - LastNavPackSRC.m_SRCInt))/21943)
							- (CumulLBA - m_VobuInfo.back().m_NavPackLBA);
						if(m_VobuInfo.back().m_EmptySlotCount < 1)
							MESSAGE2("VOBU at LBA %d has no free slot",m_VobuInfo.back().m_NavPackLBA);
						m_VobuInfo.back().m_SizeLBA = CumulLBA - m_VobuInfo.back().m_NavPackLBA;
// Checking for discontinous PTS...
if(m_VobuInfo.back().m_PTSFirstFrame >= GetInt(m_Buffer,0x39))
	MESSAGE2("PTS discontinuous at LBA %d",CumulLBA);
					}
					// Set this.
					TheVobuInfo.m_NavPtr = new uchar[NAV_MEM_SIZE];
					memcpy(TheVobuInfo.m_NavPtr,m_Buffer+NAV_MEM_START,NAV_MEM_SIZE);
					TheVobuInfo.m_NavPackLBA = TheVobuInfo.m_OrigNavPackLBA = CumulLBA;
					// Retrieve the VOBU start presentation time.
					TheVobuInfo.m_PTSFirstFrame = GetInt(m_Buffer,0x39);
					// Put this guy into our array.
					m_VobuInfo.push_back(TheVobuInfo);
					LastNavPackSRC = ThisSRC;
				}
				else if(ThePack.m_PackType == PackTypeVideo && VideoFormat == FORMAT_UNKNOWN)
				{
					// The first video pack is scanned for the format. Maybe I should get it
					// from the nav pack.
					ThePack.FindFrameType(m_Buffer,0,&m_AspectRatio);
					if(VideoFormat != FORMAT_UNKNOWN)
					{
						CSubPic::m_DisplayHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
						m_PTSPerFrame = (VideoFormat == FORMAT_PAL) ? 3600 : 3003;
						JustFoundVideoFormat();
					}
				}
				continue;
			}
			// This is a subpic pack. Find the stream ID.
			uchar StreamID = ThePack.m_StreamID-0x20;
			// Temp might be 0 if this is the first pack of a new SPU.
			CSubPic* Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]];
			m_TopStreamID = max(m_TopStreamID,(int)StreamID);
			if(Temp == NULL)
			{
				// We just found a new SPU. Allocate the subpic instance, put that in our vector.
				Temp = new CSubPic;
				m_SubPics[StreamID][m_NumSubPics[StreamID]] = Temp;
			}
			Temp->m_HasButtons = max(NavPackNumButtons,Temp->m_HasButtons);
			// Now let the CSubPic class read the pack. It will return 1 if this is the last pack of
			// the spu.
			if(Temp->ReadThisPack(m_Buffer, m_FileInfos[i].m_FileFD, LBA, LastNavPack, VOBID, 
				CELLID,StreamID,CumulLBA) == 1)
			{
				if(Temp->m_NavPackLBA == -1)
				{
					// This can happen if there hasn't been a navpack before this subpic, for example
					// because we started scanning in the middle of a VTS. At this point, I'm simply
					// going to discard this guy!
					delete Temp;
					Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]] = 0;
				}
				else
				{
					// Last pack of our SPU. If this is the first SPU of this stream, notify ourselves.
					if(m_NumSubPics[StreamID] == 0)
						JustFoundStream(StreamID);
					// Wait until PruneSPUList() is done.
					while(m_Wait)
						Sleep(1);
					m_Wait = 1;
					// Increment the number of subpics for this stream ID, and put a 0 one at the top.
					m_NumSubPics[StreamID]++;
					m_SubPics[StreamID].push_back(0);
					// Also increment the overall number of spus, and put this one on the master list.
					Temp->m_SPUNumber = m_TotNumSPU;
					m_MasterList.push_back(Temp);
					// Now find the spu just before this one in the same stream.
					for(int Prev = m_TotNumSPU-1;Prev>=0;Prev--)
					{
						if(m_MasterList[Prev]->m_StreamID == StreamID)
						{
							// Make its m_NextSPUInStream pointer point to the new guy.
							m_MasterList[Prev]->m_NextSPUInStream = Temp;
							break;
						}
					}
					m_Wait = 0;
					m_TotNumSPU++;
					// If we haven't seen this VCID before, notify ourselves.
					// JEAN: FIX: When ILVUs, this does not work well.
					if(m_VCIDLength == 0 || 256*(uint)VOBID+CELLID != m_VCIDList[m_VCIDLength-1])
					{
						m_VCIDList[m_VCIDLength++] = 256*(uint)VOBID+CELLID;
						JustFoundVCID(256*(uint)VOBID+CELLID);
					}
					// Finally notify ourselves we added an spu.
					JustAddedSPU(StreamID,m_NumSubPics[StreamID]-1);
				}
			}
		}
	}
	if(CumulLBA)
	{
		m_VobuInfo.back().m_SizeLBA = CumulLBA - m_VobuInfo.back().m_NavPackLBA;
		m_VobuInfo.back().m_EmptySlotCount = 0;
	}
	m_InputLBA = new int[m_VobuInfo.size()+1];
	m_OutputLBA = new int[m_VobuInfo.size()+1];
	for(i=0;i<m_VobuInfo.size();i++)
		m_InputLBA[i] = m_OutputLBA[i] = m_VobuInfo[i].m_NavPackLBA;
	m_InputLBA[i] = m_OutputLBA[i] = CumulLBA;

	// A bit of cleanup.
	for(i=0;i<32;i++)
	{
		if(m_SubPics[i].size() && m_SubPics[i].back() == 0)
			m_SubPics[i].pop_back();
	}
	if(m_MasterList.size() && m_MasterList.back() == 0)
		m_MasterList.pop_back();

	End = timeGetTime();
	MESSAGE4("%.2f MB scanned (%.2f MB/s) -- %d subpics",(float)CumulLBA/2/1024,
		1000*(float)CumulLBA*2/1024/(End-Start),m_TotNumSPU);
	m_Phase = Processing_Done;

//	UpdateVobuSRI(false);
end:
	ScanIFOs();
}

float CSubPicEngine::PTSToSecs(int PTS)
{
	return (float)PTS/90000;
}

int CSubPicEngine::SecsToPTS(float Secs)
{
	return m_PTSPerFrame*((int)(90000*Secs)/m_PTSPerFrame);
}

void CSubPicEngine::TestMove(int Index, float Seconds)
{
	// Use 3600 for PAL.
#if 0
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(DeltaPTS < 0)
		for(int i=0;i<m_TotNumSPU;i++)
			ChangeTiming(i,DeltaPTS);
	else
		for(int i=m_TotNumSPU-1;i>=0;i--)
			ChangeTiming(i,DeltaPTS);
	WritePacksToFiles();
	SortMasterLists();
#else
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	ChangeTiming(Index,DeltaPTS);
	WritePacksToFiles();
	SortMasterLists();
#endif
}

void CSubPicEngine::MoveSelection(float Seconds)
{
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(DeltaPTS < 0)
		for(int i=0;i<m_PruneLength;i++)
		{
			if(m_AbortNow) break;
			ChangeTiming(m_PrunedList[i]->m_SPUNumber,DeltaPTS);
			m_PercentDone = 100*i/m_PruneLength;
		}
	else
		for(int i=m_PruneLength-1;i>=0;i--)
		{
			if(m_AbortNow) break;
			ChangeTiming(m_PrunedList[i]->m_SPUNumber,DeltaPTS);
			m_PercentDone = 100*(m_PruneLength-i)/m_PruneLength;
		}
	WritePacksToFiles();
	SortMasterLists();
	UpdateVobuSRI();
	UpdateIFOAfterMove();
}

void CSubPicEngine::MoveThisOne(int Index, float Seconds)
{
	int DeltaPTS = m_PTSPerFrame*((int)(90000*Seconds)/m_PTSPerFrame);
	if(ChangeTiming(Index,DeltaPTS) == 0)
	{
		WritePacksToFiles();
		SortMasterLists();
		UpdateVobuSRI();
		UpdateIFOAfterMove();
	}
}

void CSubPicEngine::ScanIFOs()
{
	int i,j, NumFound;
	// Scan IFOs, looking for each VCID, and getting the CLUT for the PGCs which reference
	// them...
	// First, get IFO file.
	ushort *LU;
	ushort ALU[1024];
	CIFOLoader IFO(GetIFO(m_FileInfos[0].m_FileName));
	char Num = m_FileInfos[0].m_FileName[strlen(m_FileInfos[0].m_FileName)-5];
	if(Num > '0' && Num <= '9')
		LU = 0;
	else
		LU = ALU;
	if(!IFO.m_LoadedOK)
		return;

	// Allocate CLUTs.
	delete[] m_CLUTs; m_CLUTs = 0;
	m_CLUTs = new PGC_CLUT_Info[m_VCIDLength];

	// Now go through all the VCID, and find which PGCs referenced it..
	ushort PGC[1024]; 
	ushort CELLOut[1024];
	for(i=0;i<m_VCIDLength;i++)
	{
		NumFound = IFO.FindThisCell(m_VCIDList[i] / 256, m_VCIDList[i] % 256, LU, PGC, CELLOut);
		m_CLUTs[i].m_NumCLUT = NumFound;
		// For each PGC, store the clut in our table...
		for(j=0;j<NumFound;j++)
		{
			m_CLUTs[i].m_PGC.push_back(PGC[j]);
			m_CLUTs[i].m_LU.push_back((LU)?LU[j]:0);
			uchar *Foo  = new uchar[3*16];
			m_CLUTs[i].m_CLUT.push_back(Foo);
			IFO.GetCLUT(m_CLUTs[i].m_CLUT[j],(LU)?LU[j]:0,PGC[j],true);
		}
		// Get the language for the first PGC only...
		if(j>0)
		{
			int m;
			for(m=0;m<32;m++)
			{
				IFO.GetPgcSubpicLanguage(m_CLUTs[i].m_Languages+2*m,(LU)?LU[0]:0,PGC[0],0x20+m);
			}
		}
	}
	m_CLUTLength = m_VCIDLength;
	IFO.CloseIFO();
}

void CSubPicEngine::ReloadAllSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			Temp->ReloadSPU();
		}
	}
}

void CSubPicEngine::SaveAllModifiedSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			if(Temp->m_Modified)
				Temp->SaveSPU();
		}
	}
	for(i=0;i<m_NumVOBFiles;i++)
		fflush(m_FileInfos[i].m_FileFD);
}

int CSubPicEngine::AddFile(const char* FileName)
{
	strcpy(m_FileInfos[m_NumVOBFiles].m_FileName,FileName);
	m_FileInfos[m_NumVOBFiles].m_FileFD = fopen(FileName,"rb+");
	if(m_FileInfos[m_NumVOBFiles].m_FileFD == 0)
		return -1;
	m_NumVOBFiles++;
	return 0;
}

int CSubPicEngine::PruneSPUList(int CurrentSelected, int StreamID, int StartLBA, int EndLBA, ushort VOBID, uchar CELLID, int Modified)
{
	int i;
	int N;
	int NavPackLBA = -1;
	int PrevSPUNumber = -1;
	int RetIndex = -1;
	CSubPicPV Ptr;

	// Wait until the master list is updated.
	while(m_Wait == 1)
		Sleep(1);
	m_Wait = 1;
	if(CurrentSelected < m_PruneLength)
	{
		NavPackLBA = m_PrunedList[CurrentSelected]->m_NavPackLBA;
		PrevSPUNumber = m_PrunedList[CurrentSelected]->m_SPUNumber;
	}
	m_PruneLength = 0;
	m_PrunedList.clear();
	if(StreamID >= 0)
	{
		// A streamID was selected. Select the pointer that corresponds to that stream ID.
		if(StreamID > m_TopStreamID)
			goto end;
		Ptr = m_SubPics[StreamID];
		N = m_NumSubPics[StreamID];
	}
	else
	{
		// Otherwise, select the master pointer to get to all the SPUs.
		Ptr = m_MasterList;
		N = m_TotNumSPU;
	}
	for(i=0;i<N;i++)
	{
		CSubPic* Temp = Ptr[i];
		if(StartLBA != -1 && Temp->m_LBAFrom[0] < StartLBA)
			continue;
		if(EndLBA != -1 && Temp->m_LBAFrom[0] > EndLBA)
			continue;
		if(VOBID != 0 && Temp->m_VOBID != VOBID)
			continue;
		if(CELLID != 0 && Temp->m_CELLID != CELLID)
			continue;
		if(Modified != 0 && Temp->m_Modified == 0)
			continue;
		// The new list might not have the previously selected SPU. Let's find one that corresponds to
		// the same image (navpack).
		if(RetIndex == -1 && NavPackLBA <= Temp->m_NavPackLBA)
			RetIndex = m_PruneLength;
		// But if the previously selected SPU is in the new list, select this guy!
		if(PrevSPUNumber == Temp->m_SPUNumber)
			RetIndex = m_PruneLength;
		m_PrunedList.push_back(Temp);m_PruneLength++;
	}
	m_Wait = 0;
	end: 
	return RetIndex;
}

int CSubPicEngine::RetreiveNextCommand(char* Command)
{
	if(m_SelectedSubPic == 0)
		return -1;
	static int CurrentDCSQT = 0;
	static int CurrentCommand = -1;
	return m_SelectedSubPic->GetCommand(Command, CurrentDCSQT, CurrentCommand);
}

int CSubPicEngine::CumulToLocalLBA(int CumulLBA, int &FileIndex)
{
	int i;
	for(i=0;i<m_NumVOBFiles;i++)
	{
		if(CumulLBA >= m_FileInfos[i].m_StartLBA + m_FileInfos[i].m_SizeLBA)
			continue;
		if(CumulLBA >= m_FileInfos[i].m_StartLBA)
		{
			FileIndex = i;
			return CumulLBA - m_FileInfos[i].m_StartLBA;
		}
	}
	FileIndex = -1;
	return -1;
}

int CSubPicEngine::LocalToCumulLBA(int LocalLBA, int FileIndex)
{
	assert(FileIndex < m_NumVOBFiles);
	return LocalLBA + m_FileInfos[FileIndex].m_StartLBA;
}

int CSubPicEngine::ReadPacksIntoMemory(int StartLBA, int SizeLBA)
{
	m_MemNeedsWrite = 0;
	m_MemStartLBA = StartLBA;
	m_MemSizeLBA = SizeLBA;
	// Find the file in which m_MemStartLBA resides.
	int FileIndex;
	int LStartLBA = CumulToLocalLBA(m_MemStartLBA,FileIndex);
	assert(FileIndex != -1);
	if(FileIndex == -1)
		return -1;
	uchar* Ptr = m_MemoryPtr;
	// Now read as much as needed,
	fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE,SEEK_SET);
	int ToRead = m_MemSizeLBA;
	while(1)
	{
		// Read as much as possible.
		int Read = fread(Ptr,1,SECTOR_SIZE*ToRead,m_FileInfos[FileIndex].m_FileFD);
		ToRead -= Read/SECTOR_SIZE;
		// We might be done here.
		if(ToRead == 0) return 0;
		// If not, increment m_MemoryPtr, go to the next file, seek back to 0, and go again!
		Ptr += Read;
		FileIndex++;
		assert(FileIndex < m_NumVOBFiles);
		if(FileIndex >= m_NumVOBFiles) return -1;
		fseek(m_FileInfos[FileIndex].m_FileFD,0,SEEK_SET);
	}
	return -1;
}

int CSubPicEngine::WritePacksToFiles()
{
	// Don't write if it's not needed.
	if(!m_MemNeedsWrite || m_MemSizeLBA == 0)
		return 0;
	// Find the file in which m_MemStartLBA resides.
	int FileIndex;
	int LStartLBA = CumulToLocalLBA(m_MemStartLBA,FileIndex);
	assert(FileIndex != -1);
	if(FileIndex == -1)
		return -1;
	uchar* Ptr = m_MemoryPtr;
	// Now write as much as needed,
	fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE,SEEK_SET);
	int ToWrite = m_MemSizeLBA;
	while(1)
	{
		// Write as much as possible.
		int Write = min(ToWrite,m_FileInfos[FileIndex].m_SizeLBA-LStartLBA);
		int ret = fwrite(Ptr,1,SECTOR_SIZE*Write,m_FileInfos[FileIndex].m_FileFD);
		assert(ret == SECTOR_SIZE*Write);
		ToWrite -= Write;
		// We might be done here.
		if(ToWrite == 0) {m_MemNeedsWrite = 0; return 0;}
		// If not, increment m_MemoryPtr, go to the next file, seek back to 0, and go again!
		Ptr += Write*SECTOR_SIZE;
		FileIndex++;
		LStartLBA = 0;
		assert(FileIndex < m_NumVOBFiles);
		if(FileIndex >= m_NumVOBFiles) return -1;
		fseek(m_FileInfos[FileIndex].m_FileFD,0,SEEK_SET);
	}
	return -1;
}

int CSubPicEngine::MovePackInMem()
{
	// Convert to local LBA.
	int LLeftLBA = m_LeftLimitLBA - m_MemStartLBA; assert(LLeftLBA >= 0);
	int LRightLBA = m_RightLimitLBA - m_MemStartLBA; assert(LRightLBA >= 0);
	int	DeltaLBA = (m_RightLimitLBA - m_LeftLimitLBA) + 1;

	uchar TempBuf[SECTOR_SIZE];
	if(m_MoveInc > 0)
	{
		// Moving memory to the right.
		memcpy(TempBuf,m_MemoryPtr+(LRightLBA+1)*SECTOR_SIZE,SECTOR_SIZE);
		memmove(m_MemoryPtr+(LLeftLBA+1)*SECTOR_SIZE,m_MemoryPtr+LLeftLBA*SECTOR_SIZE,
			DeltaLBA*SECTOR_SIZE);
		memcpy(m_MemoryPtr+LLeftLBA*SECTOR_SIZE,TempBuf,SECTOR_SIZE);
	}
	else if(m_MoveInc < 0)
	{
		// Moving memory to the left.
		memcpy(TempBuf,m_MemoryPtr+(LLeftLBA-1)*SECTOR_SIZE,SECTOR_SIZE);
		memmove(m_MemoryPtr+(LLeftLBA-1)*SECTOR_SIZE,m_MemoryPtr+LLeftLBA*SECTOR_SIZE,
			DeltaLBA*SECTOR_SIZE);
		memcpy(m_MemoryPtr+LRightLBA*SECTOR_SIZE,TempBuf,SECTOR_SIZE);
	}
	return 0;
}



// VobuStartLBA is the lba of the vobu start (absolute), PackLBA is the pack LBA, and
// Inc is +1 to add a pack, or -1 to remove it.
void CSubPicEngine::UpdateNavPackPtrs(VobuInfo* Vobu, int PackLBA, int Inc)
{
	// Increment or decrement forward pointer.
	int VobuStartLBA = Vobu->m_NavPackLBA;
	uchar* Ptr = m_MemoryPtr + (VobuStartLBA-m_MemStartLBA)*SECTOR_SIZE;
	AddInt(Ptr,0x040f,Inc);
//	AddInt(Vobu->m_NavPtr,0x040f,Inc);
	// Adjust first, second and third reference pointers if needed.
	PackLBA -= VobuStartLBA;
	if(PackLBA <= GetInt(Ptr,0x413)) 
	{
		AddInt(Ptr,0x413,Inc); 
//		AddInt(Vobu->m_NavPtr,0x413,Inc);
	}
	if(PackLBA <= GetInt(Ptr,0x417)) 
	{
		AddInt(Ptr,0x417,Inc);
//		AddInt(Vobu->m_NavPtr,0x417,Inc);
	}
	if(PackLBA <= GetInt(Ptr,0x41b)) 
	{
		AddInt(Ptr,0x41b,Inc);
//		AddInt(Vobu->m_NavPtr,0x41b,Inc);
	}
	// JEAN: TODO: adjust offsets to audio, sub packs...
}


int CSubPicEngine::MoveOnePack(int OrigVOBU)
{
	int i;
	assert(m_MemStartLBA == m_VobuInfo[OrigVOBU].m_NavPackLBA);
	
	// Find the vobu the Dest falls in. 
	int Dest = 0;
	int Src = 0;
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA + m_VobuInfo[i].m_SizeLBA > m_DestLBABM)
		{
			Dest = i;
			// Update the navpack pointers for an insertion (+1);
			UpdateNavPackPtrs(&m_VobuInfo[i], m_DestLBABM, 1);
			break;
		}
	}
	assert(i<m_VobuInfo.size());
	// Find the vobu the Src falls in.
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA + m_VobuInfo[i].m_SizeLBA > m_SrcLBA)
		{
			Src = i;
			// Update the navpack pointers for a deletion (-1)
			UpdateNavPackPtrs(&m_VobuInfo[i], m_SrcLBA, -1);
			break;
		}
	}
	assert(i<m_VobuInfo.size());
	// The sizes also changed, and the slot counts too.
	m_VobuInfo[Dest].m_SizeLBA ++;
	m_VobuInfo[Dest].m_EmptySlotCount --;
	m_VobuInfo[Src].m_SizeLBA --;
	m_VobuInfo[Src].m_EmptySlotCount ++;

	// Finally, correct the this pointers and the LBA pointers in our vobu infos.
	for(i=OrigVOBU;i<m_VobuInfo.size();i++)
	{
		if(m_VobuInfo[i].m_NavPackLBA  > m_RightLimitLBA)
			break;
		if(m_VobuInfo[i].m_NavPackLBA >= m_LeftLimitLBA)
		{
			// Update the this->lba pointers
			AddInt(m_MemoryPtr+(m_VobuInfo[i].m_NavPackLBA-m_MemStartLBA)*SECTOR_SIZE,0x2d,m_MoveInc);
			AddInt(m_MemoryPtr+(m_VobuInfo[i].m_NavPackLBA-m_MemStartLBA)*SECTOR_SIZE,0x40b,m_MoveInc);
//			AddInt(m_VobuInfo[i].m_NavPtr,0x2d,m_MoveInc);
//			AddInt(m_VobuInfo[i].m_NavPtr,0x40b,m_MoveInc);
			m_VobuInfo[i].m_NavPackLBA += m_MoveInc;
		}
	}

	// Now move the pack.
	MovePackInMem();

	// And adjust its SRC. Retrieve the SRC of the preceding pack and increment it.
	ushort SRCH = GetShort(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA-1)*SECTOR_SIZE,0x4);
	uint SRCL = GetInt(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA-1)*SECTOR_SIZE,0x6);
	CompositeSRC CSRC;
	SrcToFrac(SRCL, SRCH, CSRC);
CSRC = ParseSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA-1)*SECTOR_SIZE);
	IncSRC(CSRC);
	FracToSrc(CSRC, SRCL, SRCH);
	PutShort(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA)*SECTOR_SIZE,0x4,SRCH);
	PutInt(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA)*SECTOR_SIZE,0x6,SRCL);
WriteSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA)*SECTOR_SIZE,CSRC);
#ifdef _DEBUG
	// Check that the next SRC is good too!
	IncSRC(CSRC);
	CompositeSRC Next = ParseSRC(m_MemoryPtr+(m_DestLBAAM-m_MemStartLBA+1)*SECTOR_SIZE);
	assert(Next.m_SRCInt > CSRC.m_SRCInt || 
		(Next.m_SRCInt == CSRC.m_SRCInt && Next.m_SRCFrac >= CSRC.m_SRCFrac));
#endif
	return 0;
}

int CSubPicEngine::FindEmptySlot(VobuInfo* DestVobu, int MinLBA)
{
	// Start at StartLBA
	MinLBA = max(MinLBA,DestVobu->m_NavPackLBA);
	uchar* Ptr = m_MemoryPtr + (MinLBA-m_MemStartLBA)*SECTOR_SIZE;
	ushort SRCH = GetShort(Ptr,0x4);
	uint SRCL = GetInt(Ptr,0x6);
	CompositeSRC ThisSRC;
	SrcToFrac(SRCL, SRCH, ThisSRC);

	int i;
	CompositeSRC NextSRC;
	for(i=1;i<DestVobu->m_SizeLBA;i++)
	{
		// Get SRC.
		SRCH = GetShort(Ptr+i*SECTOR_SIZE,0x4);
		SRCL = GetInt(Ptr+i*SECTOR_SIZE,0x6);
		SrcToFrac(SRCL, SRCH, NextSRC);
		// Now increment the previous SRC twice, and see if the current one is larger or equal. 
		IncSRC(ThisSRC);
		IncSRC(ThisSRC);
		if(NextSRC.m_SRCInt > ThisSRC.m_SRCInt || 
			(NextSRC.m_SRCInt == ThisSRC.m_SRCInt && NextSRC.m_SRCFrac >= ThisSRC.m_SRCFrac))
		{
			// We're good!
			return MinLBA+i;
		}
		ThisSRC = NextSRC;
	}
	return -1;	
}

int CSubPicEngine::FindEmptySlotBck(VobuInfo* DestVobu, int MinLBA)
{
	// Start at StartLBA
	MinLBA = min(MinLBA,DestVobu->m_NavPackLBA+DestVobu->m_SizeLBA-1);
	uchar* Ptr = m_MemoryPtr + (MinLBA-m_MemStartLBA)*SECTOR_SIZE;
	ushort SRCH = GetShort(Ptr,0x4);
	uint SRCL = GetInt(Ptr,0x6);
	CompositeSRC ThisSRC;
	SrcToFrac(SRCL, SRCH, ThisSRC);

	int i;
	CompositeSRC PrevSRC;
	for(i=1;i<MinLBA-DestVobu->m_NavPackLBA;i++)
	{
		// Get SRC.
		SRCH = GetShort(Ptr-i*SECTOR_SIZE,0x4);
		SRCL = GetInt(Ptr-i*SECTOR_SIZE,0x6);
		SrcToFrac(SRCL, SRCH, PrevSRC);
		CompositeSRC TempSRC = PrevSRC;
		// Now increment the previous SRC twice, and see if the current one is larger or equal. 
		IncSRC(PrevSRC);
		IncSRC(PrevSRC);
		if(PrevSRC.m_SRCInt < ThisSRC.m_SRCInt || 
			(PrevSRC.m_SRCInt == ThisSRC.m_SRCInt && PrevSRC.m_SRCFrac <= ThisSRC.m_SRCFrac))
		{
			// We're good!
			return MinLBA-i+1;
		}
		ThisSRC = TempSRC;
	}
	return -1;	
}

int CSubPicEngine::FindDestVobu(int SrcVobu, int PTS, int DeltaPTS)
{
	int i;
	int DestPTS = PTS + DeltaPTS;
	if(DeltaPTS >= 0)
	{
		for(i=SrcVobu;i<m_VobuInfo.size();i++)
		{
			if(m_VobuInfo[i].m_PTSFirstFrame >= DestPTS)
				break;
		}
		assert(i>0);
		i=max(0,i-1);
	}
	else
	{
		for(i=SrcVobu;i>=0;i--)
		{
			if(m_VobuInfo[i].m_PTSFirstFrame <= DestPTS)
				break;
		}
		assert(i>=0);
		i=max(0,i);
	}
	return i;
}

#define OVERALLOC 1
int CSubPicEngine::ChangeTiming(int SubpicIndex, int DeltaPTS)
{
	int i;
	CSubPic* ThisSubpic = m_MasterList[SubpicIndex];
	// Not moving button highlights!!! I'm testing whether buttons are defined, and a forced start is
	// used. It's still not 100% foolproof! In practice, I think highlights would be in a separate stream...
	if(ThisSubpic->m_HasButtons && ThisSubpic->FindCommand(FSTA_DSP,0) != -1)
	{
		MESSAGE2("Subpic %d seems to contain button highlights, not moving",SubpicIndex);
		return -1;
	}

	// First find the source vobu (the vobu the first spu pack is in).
	int SrcVobu;
	int SrcLBA = ThisSubpic->m_CLBAFrom[0];
	for(SrcVobu=0;SrcVobu<m_VobuInfo.size();SrcVobu++)
	{
		if(m_VobuInfo[SrcVobu].m_NavPackLBA > SrcLBA)
			break;
	}
	SrcVobu --;
	assert(SrcVobu >= 0 && SrcVobu<m_VobuInfo.size()-1);

	// Then find destination VOBU.
	int DestPTS = ThisSubpic->m_PTS + DeltaPTS;
	int DestVobu = FindDestVobu(SrcVobu, ThisSubpic->m_PTS, DeltaPTS);

	// If we're staying in the same VOBU, we don't need to move anything, right?
	if(SrcVobu == DestVobu)
	{
		// Change the PTS: write it in the subpic, and on the file.
		MESSAGE2("No moving required for spu %d, setting PTS",SubpicIndex);
		ThisSubpic->SetPTS(DestPTS);
		// Save it, since we're always saving when we move.
		ThisSubpic->SaveSPU();
		return -1;
	}

	if(m_FoundILVU)
	{
		ERROR2("This timing change for spu %d would require moving the subpic pack.\nInterleaved video units (ILVUs)\
 were found in the VOB file. I can't move the pack",SubpicIndex);
		return -1;
	}

	// Check whether the VOBU has enough empty slots for our packs and if we run a risk of colliding.
	int Tests[] = {-1,-2,1,-3};
	int TestDest;
	for(i=0;i<4;i++)
	{
		TestDest = DestVobu+Tests[i];
VobuInfo Foo = m_VobuInfo[TestDest];
		if(TestDest >= 0 && TestDest < m_VobuInfo.size() && m_VobuInfo[TestDest].m_EmptySlotCount >= ThisSubpic->m_NumPackInSPU 
			&& CheckForCollision(ThisSubpic, TestDest) == -1)
			break;
	}
	if(i==4)
	{
		ERROR2("Can't find an empty slot for subpic %d! It won't be moved....\n",SubpicIndex);
		return -1;
	}
	DestVobu = TestDest;

	MESSAGE5("Moving spu %d from VOBU %d to VOBU %d -- %d",SubpicIndex,SrcVobu,DestVobu,i);

	// Because the end of the spu can be beyond the end of the VOBU in which the beginning appears,
	// we'll load more memory than we need.
	int OrigVOBU = min(SrcVobu,DestVobu)-OVERALLOC; OrigVOBU = max(OrigVOBU,0);
	int EndVOBU = max(SrcVobu,DestVobu)+OVERALLOC; EndVOBU = min(EndVOBU,m_VobuInfo.size()-1);
	// Now we have our original and end lbas.
	int OrigLBA = m_VobuInfo[OrigVOBU].m_NavPackLBA;
	int EndLBA = m_VobuInfo[EndVOBU].m_NavPackLBA + m_VobuInfo[EndVOBU].m_SizeLBA;
	int SizeLBA = EndLBA - OrigLBA;


	// Load the memory between original and end lbas...
	if(OrigLBA < m_MemStartLBA || OrigLBA+SizeLBA > m_MemStartLBA+m_MemSizeLBA)
	{
		// We need more memory. Let's write what we have.
		WritePacksToFiles();
		if(m_MemAllocSize < SizeLBA)
		{
			delete[] m_MemoryPtr;
			m_MemoryPtr = new uchar[SizeLBA * SECTOR_SIZE];
			m_MemAllocSize = SizeLBA;
		}
		if(ReadPacksIntoMemory(OrigLBA, SizeLBA) != 0)
		{
			ERROR1("Couldn't load memory"); assert(0);
			return -1;
		}
		m_MemOrigVOBU = OrigVOBU; 
	}

	int DestLBA = (DestVobu > SrcVobu)? 0 : m_VobuInfo[DestVobu].m_NavPackLBA+m_VobuInfo[DestVobu].m_SizeLBA-1;
	int Inc = (DestVobu > SrcVobu)? 1 : -1;
	// Move the packs!
	for(i=0;i<ThisSubpic->m_NumPackInSPU;i++)
	{
		m_MemNeedsWrite = 1;
		SrcLBA = ThisSubpic->m_CLBAFrom[i];
		// Find the next empty slot in destination vobu. Make sure we place subsequent packs after
		// in the right order!
		if(Inc > 0)
			DestLBA = FindEmptySlot(&m_VobuInfo[DestVobu],DestLBA+1);
		else
			DestLBA = FindEmptySlotBck(&m_VobuInfo[DestVobu],DestLBA-1);
		if(DestLBA == -1)
		{
			ERROR1("Couldn't locate an empty slot!");
			assert(0); // No empty slots???
			return -1;
		}

		// This sets the lba area that will move, as well as the src and dest lbas.
		SetMoveAreaLimit(SrcLBA, DestLBA);

		// Then move the pack there. 
		if(MoveOnePack(m_MemOrigVOBU) != 0)
		{
			ERROR1("Problem moving the pack!");
			assert(0);
			return -1;
		}
		// Update all the subpics!
		UpdateAllSubpics();
		// Update the spu's navpack, fd and LBA info.
		if(i==0)
			ThisSubpic->m_NavPackLBA = m_VobuInfo[DestVobu].m_NavPackLBA;
		int NewFileIndex;
		int NewLocalLBA = CumulToLocalLBA(m_DestLBAAM, NewFileIndex);
		ThisSubpic->m_FilesFrom[i] = m_FileInfos[NewFileIndex].m_FileFD;
		ThisSubpic->m_LBAFrom[i] = NewLocalLBA;
		ThisSubpic->m_CLBAFrom[i] = m_DestLBAAM;
	}
	// Change the PTS: write it in the subpic, and on the file.
	ThisSubpic->SetPTS(DestPTS);
	BuildPTS(m_MemoryPtr+(ThisSubpic->m_CLBAFrom[0]-m_MemStartLBA)*SECTOR_SIZE+0x17, DestPTS, 0x20);
	return 0;
}


bool CSubPicEngine::DidThisPackMove(int PackLBA)
{
	return (PackLBA >= m_LeftLimitLBA && PackLBA <= m_RightLimitLBA);
}

int CSubPicEngine::UpdateAllSubpics()
{
	int i,j;
	int NewFileIndex;
	int NewLocalLBA;
	for(i=0;i<m_TotNumSPU;i++)
	{
		CSubPic* TheSubpic = m_MasterList[i];
		// If the last pack is left of LeftLBA, nothing to do.
		// CAN'T do that! During processing, it's possible for m_CLBAFrom[TheSubpic->m_NumPackInSPU-1]
		// to have an "old" value even though m_CLBAFrom[0] needs updating.
//		if(TheSubpic->m_CLBAFrom[TheSubpic->m_NumPackInSPU-1] < m_LeftLimitLBA)
//			continue;
		// If the navpack pack is right of RightLBA, we're done.
		if(TheSubpic->m_NavPackLBA > m_RightLimitLBA)
			break;
		// Check if the navpack moved, update if needed.
		if(DidThisPackMove(TheSubpic->m_NavPackLBA))
			TheSubpic->m_NavPackLBA += m_MoveInc;
		// Check if any of the subpic packs have moved.
		for(j=0;j<TheSubpic->m_NumPackInSPU;j++)
		{
			if(DidThisPackMove(TheSubpic->m_CLBAFrom[j]))
			{
				TheSubpic->m_CLBAFrom[j] += m_MoveInc;
				NewLocalLBA = CumulToLocalLBA(TheSubpic->m_CLBAFrom[j], NewFileIndex);
				TheSubpic->m_LBAFrom[j] = NewLocalLBA;
				TheSubpic->m_FilesFrom[j] = m_FileInfos[NewFileIndex].m_FileFD;
			}
		}
	}
	return 0;
}

int CSubPicEngine::ReorderMasterList(int SrcSPUIndex)
{
	int i;
	// Find destination index. To do that, find the first SPU whose first pack is past
	// ours.
	int FirstPackLBA = m_MasterList[SrcSPUIndex]->m_CLBAFrom[0];
	for(i=0;i<m_TotNumSPU;i++)
	{
		if(m_MasterList[i]->m_CLBAFrom[0] > FirstPackLBA)
			break;
	}
	int DestIndex = i;

	CSubPic* ThisSPU = m_MasterList[SrcSPUIndex];
	if(DestIndex < SrcSPUIndex)
	{
		// Moving left.
		assert(DestIndex < m_TotNumSPU);
		for(i=SrcSPUIndex;i>DestIndex;i--)
		{
			m_MasterList[i] = m_MasterList[i-1];
			m_MasterList[i]->m_SPUNumber = i;
		}
	}
	else
	{
		// Moving right. The destination should not move, the one immediately left of it should.
		DestIndex--;
		assert(DestIndex >= 0);
		for(i=SrcSPUIndex;i<DestIndex;i++)
		{
			m_MasterList[i] = m_MasterList[i+1];
			m_MasterList[i]->m_SPUNumber = i;
		}
	}
	m_MasterList[DestIndex] = ThisSPU;
	m_MasterList[DestIndex]->m_SPUNumber = DestIndex;
	return 0;
}

void CSubPicEngine::SetMoveAreaLimit(int SrcLBA, int DestLBA)
{
	// SrcLBA is the original LBA of the pack to be moved.
	// DestLBA is the target LBA where the pack is to be moved measured *before the move!*.
	// When moving the pack to the right, the memory will be moved 1 pack to the left, so the
	// "after the move" destination LBA will be DestLBA-1.
	if(DestLBA < SrcLBA)
	{
		// The pack is moving left, the area is moving right.
		m_LeftLimitLBA = DestLBA;	// The destination pack is moving right.
		m_RightLimitLBA = SrcLBA-1;	// The last pack to move is the one right before Src.
		m_MoveInc = 1;
		m_DestLBAAM = DestLBA;
	}
	else if(DestLBA > SrcLBA)
	{
		// The pack is moving right, the area is moving left.
		m_LeftLimitLBA = SrcLBA+1;	// The first pack to move is the one rigth of Src.
		m_RightLimitLBA = DestLBA-1;// The last pack to move is the one right before DestLBA.
		m_MoveInc = -1;
		m_DestLBAAM = DestLBA-1;
	}
	else
	{
		m_LeftLimitLBA = m_RightLimitLBA = SrcLBA;	
		m_MoveInc = 0;
	}
	m_SrcLBA = SrcLBA;
	m_DestLBABM = DestLBA;
}

int CSubPicEngine::SortMasterLists()
{
	// Make sure the SPU number is updated in each spu.
	MESSAGE1("Sorting Master Lists");
	SortThisList(&m_MasterList,true);
	// Now do the stream lists.
	int StreamID;
	for(StreamID=0;StreamID<32;StreamID++)
	{
		SortThisList(&m_SubPics[StreamID]);
	}
	return 0;
}

// Function used to sort subpics.
bool foo(CSubPic* A, CSubPic* B) 
{
	return A->m_CLBAFrom[0] < B->m_CLBAFrom[0];
}

int CSubPicEngine::SortThisList(CSubPicPV* List, bool UpdateSPUNumber)
{
	// Sorting from scratch. Using STL sort for speed (much much faster).
	int i,k;
	// Just a precaution, because of the way I'm creating these lists.
	if(List->size() && List->back() == 0)
		List->pop_back();
	int N=List->size();

	// Sort this guy.
	std::sort(List->begin(),List->end(),foo);
	if(UpdateSPUNumber)
	{
		for(i=0;i<N;i++)
		{
			(*List)[i]->m_SPUNumber = i;
			// Also update the "forward" pointer.
			for(k=i-1;k>0;k--)
			{
				if((*List)[k]->m_StreamID == (*List)[i]->m_StreamID)
				{
					// Make its m_NextSPUInStream pointer point to the new guy.
					// TODO FIX: Adjust the spu duration so it does not collide?
					(*List)[k]->m_NextSPUInStream = (*List)[i];
					break;
				}
			}
		}
	}
	return 0;
}


void CSubPicEngine::UpdateIFOAfterMove()
{
	MESSAGE1("Updating IFO");
	if(m_VobuInfo.size() == 0 || m_OutputLBA == 0)
		return;
	int i,j;
	int TableSize = m_VobuInfo.size()+1;
	// Build the output LBA array based on our vobu info vector.
	for(i=0;i<m_VobuInfo.size();i++)
		m_OutputLBA[i] = m_VobuInfo[i].m_NavPackLBA;
	m_OutputLBA[i] = m_InputLBA[i];

	// See if anything changed.
	for(i=0;i<TableSize;i++)
	{
		if(m_OutputLBA[i] != m_InputLBA[i]) 
			break;
	}
	// No need to updated.
	if(i==TableSize)
		return;

	// Now that we have both the input and output LBA array, update IFO...
	// First load it.
	CIFOLoader IFO(GetIFO(m_FileInfos[0].m_FileName));
	char Num = m_FileInfos[0].m_FileName[strlen(m_FileInfos[0].m_FileName)-5];
	// Not updating menu domain.
	if(Num <= '0' || Num > '9' || !IFO.m_LoadedOK)
		return;
	// This updates the VOBU_ADMAP tables, and the PGC pointers.
	IFO.UpdateNavPackInfo(m_InputLBA, m_OutputLBA, TableSize,false);

	// Now cleanup VTS_C_ADT. I'm doing it by hand because the CIFOLoader class does not have
	// a function to do that from an input/output vobu map.
	// Get the table.
	VTSM_C_ADTTable* Table = (VTSM_C_ADTTable*) IFO.FindTable(VTS_C_ADT);
	if(Table != 0)
	{
		ushort NumCells = (FixEndian(Table->m_LastByte) - 8 + 1) / sizeof(VTSM_C_ADTItem);
		VTSM_C_ADTItem* ThisItem;
		// Loop over all the cells in the VTS_C_ADT table.
		for(i=0;i<(int)NumCells;i++)
		{
			// Update the start and end lba of each cell according to our input/output map.
			ThisItem = &Table->m_Items[i];
			uint StartLBA = FixEndian((uint)ThisItem->m_StartLBA);
			uint EndLBA = FixEndian((uint)ThisItem->m_EndLBA);
			for(j=0;j<TableSize;j++)
			{
				// Find where the startlba was moved
				if(StartLBA == m_InputLBA[j])
				{
					ThisItem->m_StartLBA = FixEndian((uint)m_OutputLBA[j]);
					break;
				}
			}
			assert(j<TableSize);
			for(j=0;j<TableSize;j++)
			{
				// Find where the endlba was moved. The end lba points 1 pack before the start of the
				// next vobu. 
				if(EndLBA == m_InputLBA[j]-1)
				{
					ThisItem->m_EndLBA = FixEndian((uint)m_OutputLBA[j]-1);
					break;
				}
			}
			assert(j<TableSize);
		}
	}
	// Save IFO.
	IFO.WriteIFO();
	
	// Now everything is in synch.
	for(i=0;i<TableSize;i++)
		m_InputLBA[i] = m_OutputLBA[i];
}


int CSubPicEngine::CheckForCollision(CSubPic* TheSubpic, int DestVobu)
{
	int i;
	for(i=0;i<m_TotNumSPU;i++)
	{
		if(m_MasterList[i] == TheSubpic ||
			m_MasterList[i]->m_StreamID != TheSubpic->m_StreamID)
			continue;
		// Same stream ID, let's see whether the subpic packs will intersect.
		if(m_MasterList[i]->m_NavPackLBA == m_VobuInfo[DestVobu].m_NavPackLBA)
		{
			// Collision will occur. The subpic in question has the same navpack...
			return i;
		}
	}
	return -1;
}

// This will update Offset depending on whether the target lba moved relative to the start
// of the vobu. Dir indicates whether Offset is a forward or backward pointer.
int CSubPicEngine::UpdateThisPointer(int VobuIndex, int Offset)
{
	// Calculate the distance between the current VobuStart and the current location of the
	// target. 
	int OrigTargetLBA = m_VobuInfo[VobuIndex].m_OrigNavPackLBA + Offset;
	int NewTargetLBA;
	// Now, find the new location of OrigTargetLBA;
	int Inc = Offset > 0 ? 1 : -1;
	int i = VobuIndex;
#if 1
	// This version will not fix corrupted SRI (trick-play) pointers.
	for(i=VobuIndex;i>=0 && i< m_VobuInfo.size();i+=Inc)
		if(OrigTargetLBA == m_VobuInfo[i].m_OrigNavPackLBA)
			break;
	if(i == -1 ||  i == m_VobuInfo.size())
	{
		// We didn't find the navpack. Could be two things: either the pointers were corrupted or we only
		// loaded 1 of several vob files from a multi-vob VTS.
		MESSAGE2("VOBU SRI offset corrupted at LBA %d",m_VobuInfo[VobuIndex].m_NavPackLBA);
		return Offset;
	}
#else
	// FIX: TODO: return the closest pointers if you can't find an exact match!
	if(Offset > 0)
	{
		for(i=VobuIndex;i< m_VobuInfo.size();i++)
			if(m_VobuInfo[i].m_OrigNavPackLBA >= OrigTargetLBA)
				break;
		int Diff = m_VobuInfo[i].m_OrigNavPackLBA - OrigTargetLBA;
		if(Diff != 0 && i>0 && OrigTargetLBA - m_VobuInfo[i-1].m_OrigNavPackLBA < Diff)
		{
			MESSAGE2("VOBU SRI offset corrupted at LBA %d... Fixing",m_VobuInfo[VobuIndex].m_NavPackLBA);
			i--;
		}
	}
	else
	{
		for(i=VobuIndex;i>=0;i--)
			if(m_VobuInfo[i].m_OrigNavPackLBA <= OrigTargetLBA)
				break;
		int Diff = OrigTargetLBA - m_VobuInfo[i].m_OrigNavPackLBA;
		if(Diff != 0 && i<m_VobuInfo.size()-1 && m_VobuInfo[i+1].m_OrigNavPackLBA - OrigTargetLBA < Diff)
		{
			MESSAGE2("VOBU SRI offset corrupted at LBA %d... Fixing",m_VobuInfo[VobuIndex].m_NavPackLBA);
			i++;
		}
	}
#endif
	NewTargetLBA = m_VobuInfo[i].m_NavPackLBA;
	return NewTargetLBA-m_VobuInfo[VobuIndex].m_NavPackLBA;

}

// Parses an offset in the navpack data in Ptr at offset Loc. Dir indicate whether the offset
// is forward (+1) or backward (-1). VobuIndex is the index of the current navpack we're updating.
// If the offset needs to be updated, it is written and the function returns true. Otherwise false.
bool CSubPicEngine::ParseAndUpdatePointer1(uchar* Ptr, int Loc, int Dir, int VobuIndex)
{
	int Offset;
	Offset = GetInt(Ptr,Loc-NAV_MEM_START);
	if(Offset & 0x80000000 && Offset != 0xbfffffff)
	{
		int NewOffset = UpdateThisPointer(VobuIndex,Dir*(int)(Offset & 0x7FFFFFFF));
		NewOffset = abs(NewOffset);
		NewOffset |= 0x80000000;
		PutInt(Ptr,Loc-NAV_MEM_START,NewOffset);
		return NewOffset!=Offset;
	}
	return false;
}

// Parses an offset in the navpack data in Ptr at offset Loc. Dir indicate whether the offset
// is forward (+1) or backward (-1). VobuIndex is the index of the current navpack we're updating.
// If the offset needs to be updated, it is written and the function returns true. Otherwise false.
bool CSubPicEngine::ParseAndUpdatePointer2(uchar* Ptr, int Loc, int Dir, int VobuIndex)
{
	int Offset;
	Offset = GetInt(Ptr,Loc-NAV_MEM_START);
	if(Offset & 0x80000000 && Offset != 0xbfffffff)
	{
		int Bits = Offset & 0xc0000000;
		int NewOffset = UpdateThisPointer(VobuIndex,Dir*(int)(Offset & 0x3FFFFFFF));
		NewOffset = abs(NewOffset);
		NewOffset |= Bits;
		PutInt(Ptr,Loc-NAV_MEM_START,NewOffset);
		return NewOffset!=Offset;
	}
	return false;
}

int CSubPicEngine::UpdateVobuSRI(bool JustCheck)
{
	int VobuIndex;
	MESSAGE1("Updating VOBU SRI pointers");
	m_Phase = Processing_UpdatingSRI;
	for(VobuIndex=0;VobuIndex<m_VobuInfo.size();VobuIndex++)
	{
		m_PercentDone = 100*VobuIndex/m_VobuInfo.size();
		// Read all the pointers, see if they need updating, update and write if necessary.
		bool NeedsWrite = 0;
		uchar* Ptr = m_VobuInfo[VobuIndex].m_NavPtr;
		// Parse and update first vobu_sri offset.
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x4f1,1,VobuIndex);
		int i;
		for(i=0x4f5;i<=0x53d;i+=4)
		{	// Forward pointers.
			NeedsWrite |= ParseAndUpdatePointer2(Ptr,i,1,VobuIndex);
		}
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x541,1,VobuIndex);
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x545,-1,VobuIndex);
		for(i=0x549;i<=0x591;i+=4)
		{	// Backward pointers
			NeedsWrite |= ParseAndUpdatePointer2(Ptr,i,-1,VobuIndex);
		}
		NeedsWrite |= ParseAndUpdatePointer1(Ptr,0x595,-1,VobuIndex);
		if(NeedsWrite && !JustCheck)
		{
			// At least one of the offset was updated. Write the pack..
			int FileIndex;
			int LStartLBA = CumulToLocalLBA(m_VobuInfo[VobuIndex].m_NavPackLBA,FileIndex);
			assert(FileIndex != -1);
			if(FileIndex == -1)
				return -1;
			fseek(m_FileInfos[FileIndex].m_FileFD,LStartLBA*SECTOR_SIZE+NAV_MEM_START,SEEK_SET);
			int ret = fwrite(Ptr,1,NAV_MEM_SIZE,m_FileInfos[FileIndex].m_FileFD);
			assert(ret == NAV_MEM_SIZE);		
		}
	}
	for(VobuIndex=0;VobuIndex<m_VobuInfo.size();VobuIndex++)
		m_VobuInfo[VobuIndex].m_OrigNavPackLBA = m_VobuInfo[VobuIndex].m_NavPackLBA;
	return 0;
}
